Package: automatedRecLin
Title: Record Linkage Based on an Entropy-Maximizing Classifier
Version: 1.1.1
Authors@R: 
  c(person(given = "Adam",
           family = "Struzik",
           role = c("aut", "cre"),
           email = "adastr5@st.amu.edu.pl",
           comment = c(ORCID = "0009-0002-2547-482X")),
    person(given = "Maciej",
           family = "Beręsewicz",
           role = c("aut", "ctb"),
           email = "maciej.beresewicz@ue.poznan.pl",
           comment = c(ORCID = "0000-0002-8281-4301")))
Description: The goal of 'automatedRecLin' is to perform record linkage (also known as entity resolution) in unsupervised or supervised settings. It compares pairs of records from two datasets using selected comparison functions to estimate the probability or density ratio between matched and non-matched records. Based on these estimates, it predicts a set of matches that maximizes entropy. For details see: Lee et al. (2022) <https://www150.statcan.gc.ca/n1/pub/12-001-x/2022001/article/00007-eng.htm>, Vo et al. (2023) <https://ideas.repec.org/a/eee/csdana/v179y2023ics0167947322002365.html>, Sugiyama et al. (2008) <doi:10.1007/s10463-008-0197-x>.
License: GPL-3
Encoding: UTF-8
URL: https://github.com/ncn-foreigners/automatedRecLin
BugReports: https://github.com/ncn-foreigners/automatedRecLin/issues
Imports: blocking, data.table, densityratio, FixedPoint, methods,
        nleqslv, purrr, reclin2, stats, utils
Suggests: knitr, rmarkdown, tinytest, xgboost
Depends: R (>= 4.1.0)
LazyData: true
VignetteBuilder: knitr
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-21 06:30:50 UTC; adamstruzik
Author: Adam Struzik [aut, cre] (ORCID:
    <https://orcid.org/0009-0002-2547-482X>),
  Maciej Beręsewicz [aut, ctb] (ORCID:
    <https://orcid.org/0000-0002-8281-4301>)
Maintainer: Adam Struzik <adastr5@st.amu.edu.pl>
Repository: CRAN
Date/Publication: 2026-05-21 07:10:02 UTC
Built: R 4.6.0; ; 2026-05-22 00:25:19 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(predict,rec_lin_model)
S3method(print,comparison_vectors)
S3method(print,mec_blocking)
S3method(print,mec_rec_lin)
S3method(print,rec_lin_model)
S3method(print,rec_lin_predictions)
export(abs_distance)
export(comparison_vectors)
export(control_kliep)
export(custom_rec_lin_model)
export(jarowinkler_complement)
export(mec)
export(mec_blocking)
export(train_rec_lin)
import(data.table)
import(stats)
importFrom(FixedPoint,FixedPoint)
importFrom(blocking,blocking)
importFrom(densityratio,kliep)
importFrom(methods,is)
importFrom(nleqslv,nleqslv)
importFrom(purrr,partial)
importFrom(reclin2,cmp_identical)
importFrom(reclin2,cmp_jarowinkler)
importFrom(stats,dbinom)
importFrom(stats,dgamma)
importFrom(stats,predict)
importFrom(stats,runif)
importFrom(utils,head)

The `automatedRecLin` Package
================

<!-- badges: start -->

[![R-CMD-check](https://github.com/ncn-foreigners/automatedRecLin/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/ncn-foreigners/automatedRecLin/actions/workflows/R-CMD-check.yaml)
[![test-coverage](https://github.com/ncn-foreigners/automatedRecLin/actions/workflows/test-coverage.yaml/badge.svg)](https://github.com/ncn-foreigners/automatedRecLin/actions/workflows/test-coverage.yaml)
[![CRAN
status](https://www.r-pkg.org/badges/version/automatedRecLin)](https://cran.r-project.org/package=automatedRecLin)
[![CRAN
downloads](https://cranlogs.r-pkg.org/badges/grand-total/automatedRecLin)](https://cran.r-project.org/package=automatedRecLin)
[![CRAN
downloads](https://cranlogs.r-pkg.org/badges/automatedRecLin)](https://cran.r-project.org/package=automatedRecLin)
<!-- badges: end -->

## Description

This R package is designed to perform record linkage (also known as
entity resolution) in unsupervised or supervised settings. It compares
pairs of records from two datasets using selected comparison functions
to estimate the probability or density ratio between matched and
non-matched records. Based on these estimates, it predicts a set of
matches that maximizes entropy.

## Installation

Install the stable version from CRAN.

``` r
install.packages("automatedRecLin")
```

To install the development version from GitHub you can use the `pak`
package.

``` r
# install.packages("pak") # uncomment if needed
pak::pkg_install("ncn-foreigners/automatedRecLin")
```

## Basic usage

Load the package for the examples.

``` r
library(automatedRecLin)
```

### Unsupervised maximum entropy classifier for record linkage

Generate two simple datasets that contain some common records, with
typos in some cases.

``` r
df_1 <- data.frame(
  name = c("Emma", "Liam", "Olivia", "Noah", "Ava",
           "Ethan", "Sophia", "Mason", "Isabella", "James"),
  surname = c("Smith", "Johnson", "Williams", "Brown", "Jones",
              "Garcia", "Miller", "Davis", "Rodriguez", "Wilson"),
  city = c("New York", "Los Angeles", "Chicago", "Houston", "Phoenix",
           "Philadelphia", "San Antonio", "San Diego", "Dallas", "San Jose")
)
df_2 <- data.frame(
  name = c(
    "Emma", "Liam", "Olivia", "Noah",
    "Ava", "Ehtan", "Sopia", "Mson",
    "Charlotte", "Benjamin", "Amelia", "Lucas"
  ),
  surname = c(
    "Smith", "Johnson", "Williams", "Brown",
    "Jnes", "Garca", "Miler", "Dvis",
    "Martinez", "Lee", "Hernandez", "Clark"
  ),
  city = c(
    "New York", "Los Angeles", "Chicago", "Houston",
    "Phonix", "Philadelpia", "San Antnio", "San Dieg",
    "Seattle", "Miami", "Boston", "Denver"
  )
)
df_1
#>        name   surname         city
#> 1      Emma     Smith     New York
#> 2      Liam   Johnson  Los Angeles
#> 3    Olivia  Williams      Chicago
#> 4      Noah     Brown      Houston
#> 5       Ava     Jones      Phoenix
#> 6     Ethan    Garcia Philadelphia
#> 7    Sophia    Miller  San Antonio
#> 8     Mason     Davis    San Diego
#> 9  Isabella Rodriguez       Dallas
#> 10    James    Wilson     San Jose
df_2
#>         name   surname        city
#> 1       Emma     Smith    New York
#> 2       Liam   Johnson Los Angeles
#> 3     Olivia  Williams     Chicago
#> 4       Noah     Brown     Houston
#> 5        Ava      Jnes      Phonix
#> 6      Ehtan     Garca Philadelpia
#> 7      Sopia     Miler  San Antnio
#> 8       Mson      Dvis    San Dieg
#> 9  Charlotte  Martinez     Seattle
#> 10  Benjamin       Lee       Miami
#> 11    Amelia Hernandez      Boston
#> 12     Lucas     Clark      Denver
```

Specify the key variables used for record linkage. Select a comparison
function (i.e., a function to compare pairs of records) for each
variable. For example, use `jarowinkler_complement()` from the
`automatedRecLin` package (the Jaro-Winkler distance, i.e.,
`1 - Jaro-Winkler similarity`). Choose a method for estimating the
probability or density ratio for each variable. The available methods
are: `"binary"`, `"continuous_parametric"`,
`"continuous_nonparametric"`, and `"hit_miss"` (only for unsupervised
learning).

``` r
variables <- c("name", "surname", "city")
comparators <- list(
  "name" = jarowinkler_complement(),
  "surname" = jarowinkler_complement(),
  "city" = jarowinkler_complement()
)
methods <- list(
  "name" = "continuous_parametric",
  "surname" = "continuous_parametric",
  "city" = "continuous_parametric"
)
```

Perform record linkage using `mec()`. The output contains the following
information:

- the names of key variables,
- the number of predicted matches,
- the first 6 predicted matches (with their estimated probability or
  density ratio),
- the method for constructing the predicted set of matches (default:
  `"size"`),
- estimated false link rate (FLR),
- estimated missing match rate (MMR),
- estimated parameters for the variables using the `"binary"`,
  `"continuous_parametric"` or `"hit_miss"` methods.

``` r
set.seed(1)
unsup_result <- mec(A = df_1, B = df_2,
                    variables = variables,
                    comparators = comparators,
                    methods = methods)
unsup_result
#> Record linkage based on the following variables: name, surname, city.
#> ========================================================
#> The algorithm predicted 8 matches.
#> The first 6 predicted matches are:
#>        a     b ratio / 1000
#>    <int> <int>        <num>
#> 1:     6     6 1.433031e+08
#> 2:     8     8 3.198692e+07
#> 3:     7     7 9.673745e+05
#> 4:     5     5 2.813745e+04
#> 5:     1     1 3.375000e+00
#> 6:     2     2 3.375000e+00
#> ========================================================
#> The construction of the classification set was based on estimates of its size.
#> Estimated false link rate (FLR): 0.2066 %.
#> Estimated missing match rate (MMR): 0.2066 %.
#> ========================================================
#> Variables selected for the continuous parametric method: name, surname, city.
#> Estimated parameters for the continuous parametric method:
#>         variable p_0_M    alpha_M   beta_M      p_0_U  alpha_U    beta_U
#>           <char> <num>      <num>    <num>      <num>    <num>     <num>
#> 1:    gamma_name 0.625 138.462279 2199.107 0.04166667 6.516736 11.173089
#> 2: gamma_surname 0.500 120.665706 1974.530 0.03333333 4.622775  7.167261
#> 3:    gamma_city 0.500   6.512723  135.163 0.03333333 5.233194  9.313035
```

### Supervised maximum entropy classifier for record linkage

Generate two simple training datasets that contain some common records,
with typos in some cases.

``` r
df_1_train <- data.frame(
        name = c(
          "James", "Emma", "William", "Olivia", "Thomas",
          "Sophie", "Harry", "Amelia", "George", "Isabella"
          ),
        surname = c(
          "Smith", "Johnson", "Brown", "Taylor", "Wilson",
          "Davis", "Clark", "Harris", "Lewis", "Walker"
          )
)
df_2_train <- data.frame(
        name = c(
          "James", "Ema", "Wimliam", "Olivia", "Charlotte",
          "Henry", "Lucy", "Edward", "Alice", "Jack"
          ),
        surname = c(
          "Smith", "Johnson", "Bron", "Tailor", "Moore",
          "Evans", "Hall", "Wright", "Green", "King"
          )
)
df_1_train
#>        name surname
#> 1     James   Smith
#> 2      Emma Johnson
#> 3   William   Brown
#> 4    Olivia  Taylor
#> 5    Thomas  Wilson
#> 6    Sophie   Davis
#> 7     Harry   Clark
#> 8    Amelia  Harris
#> 9    George   Lewis
#> 10 Isabella  Walker
df_2_train
#>         name surname
#> 1      James   Smith
#> 2        Ema Johnson
#> 3    Wimliam    Bron
#> 4     Olivia  Tailor
#> 5  Charlotte   Moore
#> 6      Henry   Evans
#> 7       Lucy    Hall
#> 8     Edward  Wright
#> 9      Alice   Green
#> 10      Jack    King
```

Specify the key variables, select comparison functions and choose
methods for estimating the probability or density ratio. Additionally,
provide a `data.frame` or `data.table` indicating known matches.

``` r
variables_train <- c("name", "surname")
comparators_train <- list("name" = jarowinkler_complement(),
                          "surname" = jarowinkler_complement())
methods_train <- list("name" = "continuous_nonparametric",
                      "surname" = "continuous_nonparametric")
matches_train <- data.frame("a" = 1:4, "b" = 1:4)
```

Train a record linkage model using `train_rec_lin()`.

``` r
model <- train_rec_lin(A = df_1_train, B = df_2_train,
                       matches = matches_train,
                       variables = variables_train,
                       comparators = comparators_train,
                       methods = methods_train)
#> Warning in check.sigma(nsigma, sigma_quantile, sigma, dist_nu): There are duplicate values in 'sigma', only the unique values are used.
#> Warning in check.sigma(nsigma, sigma_quantile, sigma, dist_nu): There are duplicate values in 'sigma', only the unique values are used.
model
#> Record linkage model based on the following variables: name, surname.
#> The prior probability of matching is 0.04.
#> ========================================================
#> Variables selected for the continuous nonparametric method: name, surname.
#> ========================================================
#> Probability/density ratio type: 1.
```

Generate two new datasets for record linkage prediction.

``` r
df_1_new <- data.frame(
  "name" = c("John", "Emily", "Mark", "Anna", "David"),
  "surname" = c("Smith", "Johnson", "Taylor", "Williams", "Brown")
)
df_2_new <- data.frame(
  "name" = c("John", "Emely", "Mark", "Michael"),
  "surname" = c("Smitth", "Johnson", "Tailor", "Henders")
)
df_1_new
#>    name  surname
#> 1  John    Smith
#> 2 Emily  Johnson
#> 3  Mark   Taylor
#> 4  Anna Williams
#> 5 David    Brown
df_2_new
#>      name surname
#> 1    John  Smitth
#> 2   Emely Johnson
#> 3    Mark  Tailor
#> 4 Michael Henders
```

Predict matches using the `predict()` method for `rec_lin_model`
objects. The output has a similar structure to that of `mec()`.

``` r
result_sup <- predict(model, df_1_new, df_2_new)
result_sup
#> The algorithm predicted 3 matches.
#> The first 3 predicted matches are:
#>        a     b ratio / 1000
#>    <int> <int>        <num>
#> 1:     2     2   0.04761403
#> 2:     3     3   0.02929836
#> 3:     1     1   0.02665718
#> ========================================================
#> The construction of the classification set was based on estimates of its size.
#> Estimated false link rate (FLR): 0.0000 %.
#> Estimated missing match rate (MMR): 0.0000 %.
```

## Integration with a custom machine learning model

The `automatedRecLin` package supports supervised record linkage using a
custom machine learning (ML) model that predicts the probability of
matching based on comparison vectors (e.g., XGBoost, logistic
regression). For example, install and load the `xgboost` package.

``` r
# install.packages("xgboost") # uncomment if needed
library(xgboost)
```

Use the same data, variables, and comparators as in the previous
example. First, use `comparison_vectors()` to create comparison vectors
that the model will be trained on.

``` r
vectors <- comparison_vectors(A = df_1_train, B = df_2_train,
                              variables = variables_train,
                              comparators = comparators_train,
                              matches = matches_train)
vectors
#> Comparison based on the following variables: name, surname.
#> ========================================================
#>        a     b gamma_name gamma_surname match
#>    <int> <int>      <num>         <num> <num>
#> 1:     1     1  0.0000000     0.0000000     1
#> 2:     1     2  0.4777778     0.5523810     0
#> 3:     1     3  0.5523810     1.0000000     0
#> 4:     1     4  1.0000000     0.5444444     0
#> 5:     1     5  0.5629630     1.0000000     0
#> 6:     1     6  1.0000000     1.0000000     0
```

Train the XGBoost model.

``` r
model_xgb <- xgboost(x = as.matrix(vectors$Omega[, c("gamma_name", "gamma_surname")]),
                     y = factor(vectors$Omega$match),
                     objective = "binary:logistic", eval_metric = "logloss",
                     nrounds = 100, verbosity = 0, nthread = 1)
```

Create the XGBoost-based record linkage model.

``` r
custom_xgb_model <- custom_rec_lin_model(model_xgb, vectors)
custom_xgb_model
#> Record linkage model based on the following variables: name, surname.
#> A custom ML model was used.
#> The prior probability of matching is 0.04.
#> ========================================================
#> Probability/density ratio type: 2.
```

Use the model for predictions. Note that the `xgboost` package requires
a matrix as input for the `predict()` method for `rec_lin_model` objects
and that needs to be specified in the `data_type` argument. Set
`type = "response"` to ensure the XGBoost model predicts the probability
of matching (this argument may vary depending on the model or library
used).

``` r
result_xgb <- predict(custom_xgb_model, df_1_new, df_2_new,
                      data_type = "matrix", type = "response")
result_xgb
#> The algorithm predicted 3 matches.
#> The first 3 predicted matches are:
#>        a     b ratio / 1000
#>    <int> <int>        <num>
#> 1:     1     1    0.0299477
#> 2:     2     2    0.0299477
#> 3:     3     3    0.0299477
#> ========================================================
#> The construction of the classification set was based on estimates of its size.
#> Estimated false link rate (FLR): 15.9112 %.
#> Estimated missing match rate (MMR): 15.9112 %.
```

## Integration with the `blocking` package

The `automatedRecLin` R package can be integrated with the
[`blocking`](https://cran.r-project.org/package=blocking) package. For
details, see the vignette entitled *MEC with blocking*.

## Funding

Work on this package is supported by the National Science Centre, OPUS
20 grant no. 2020/39/B/HS4/00941 (Towards census-like statistics for
foreign-born populations – quality, data integration and estimation).

## References

Lee, D., Zhang, L.-C. and Kim, J. K. (2022). [Maximum entropy
classification for record
linkage.](https://www150.statcan.gc.ca/n1/pub/12-001-x/2022001/article/00007-eng.htm)
Survey Methodology, Statistics Canada, Catalogue No. 12-001-X, Vol. 48,
No. 1.

Vo, T. H., Chauvet, G., Happe, A., Oger, E., Paquelet, S., and Garès, V.
(2023). [Extending the Fellegi-Sunter record linkage model for
mixed-type data with application to the French national health data
system.](https://ideas.repec.org/a/eee/csdana/v179y2023ics0167947322002365.html)
Computational Statistics & Data Analysis, 179, 107656.

Sugiyama, M., Suzuki, T., Nakajima, S. et al. [Direct importance
estimation for covariate shift
adaptation.](https://doi.org/10.1007/s10463-008-0197-x) Ann Inst Stat
Math 60, 699–746 (2008).

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, message = FALSE---------------------------------------------------
library(automatedRecLin)
library(data.table)

options("text2vec.mc.cores" = 1L)

## ----data---------------------------------------------------------------------
data("census", package = "automatedRecLin")
data("cis", package = "automatedRecLin")
setDT(census)
setDT(cis)

NROW(cis)
NROW(census)

## ----true-matches-------------------------------------------------------------
cis[is.na(cis)] <- ""
census[is.na(census)] <- ""

cis[, pername1 := gsub("-", "", pername1)]
census[, pername1 := gsub("-", "", pername1)]

true_matches <- merge(
  x = cis[, .(a = .I, person_id)],
  y = census[, .(b = .I, person_id)],
  by = "person_id"
)[, .(a, b)]

NROW(true_matches)

## ----model-specification------------------------------------------------------
variables <- c(
  "pername1", "pername2", "sex",
  "dob_day", "dob_mon", "dob_year"
)

comparators <- list(
  "pername1" = jarowinkler_complement(),
  "pername2" = jarowinkler_complement()
)

methods <- list(
  "pername1" = "continuous_parametric",
  "pername2" = "continuous_parametric"
)

blocking_variables <- c(variables, "enumcap", "enumpc")

## ----mec-blocking-------------------------------------------------------------
set.seed(1)

result <- mec_blocking(
  A = cis,
  B = census,
  variables = variables,
  comparators = comparators,
  methods = methods,
  blocking_variables = blocking_variables,
  blocking_sep = " ",
  controls_blocking = list(seed = 1, n_threads = 1),
  alpha = 0.5,
  true_matches = true_matches
)

result

## ----results, echo = FALSE----------------------------------------------------
data.table(
  step = c("Training", "Blocking", "Linkage"),
  result = c(
    paste0(
      "all_candidate_pairs on ",
      format(NROW(result$block_summary), big.mark = ","),
      " blocks"
    ),
    paste0(
      format(result$blocking_eval[["preserved_matches"]], big.mark = ","),
      " of ",
      format(result$blocking_eval[["true_matches"]], big.mark = ","),
      " known links retained"
    ),
    paste0(
      "FLR = ", sprintf("%.2f%%", 100 * result$eval_metrics[["FLR"]]),
      "; MMR = ", sprintf("%.2f%%", 100 * result$eval_metrics[["MMR"]])
    )
  )
)

#' @import data.table
#' @importFrom stats dbinom
#' @importFrom stats predict
#' @importFrom FixedPoint FixedPoint
#' @importFrom utils head
#'
#' @title Predict Matches Based on a Given Record Linkage Model
#'
#' @author Adam Struzik
#'
#' @description
#' Predicts matches between records in two datasets based on a given record linkage model,
#' using the maximum entropy classification (MEC) algorithm
#' (see [Lee et al. (2022)](https://www150.statcan.gc.ca/n1/pub/12-001-x/2022001/article/00007-eng.htm)).
#'
#' @param object A `rec_lin_model` object from [train_rec_lin()] or [custom_rec_lin_model()].
#' @param newdata_A A duplicate-free `data.frame` or `data.table`.
#' @param newdata_B A duplicate-free `data.frame` or `data.table`.
#' @param duplicates_in_A Logical indicating whether to allow `A` to have duplicate records.
#' @param set_construction A method for constructing the predicted set of matches (`"size"`, `"flr"` or `"mmr"`).
#' @param fixed_method A method for solving fixed-point equations using the \link[FixedPoint:FixedPoint]{FixedPoint()} function.
#' @param target_rate A target false link rate (FLR) or missing match rate (MMR)
#' (used only if `set_construction == "flr"` or `set_construction == "mmr"`).
#' @param tol Error tolerance in the bisection procedure
#' (used only if `set_construction == "flr"` or `set_construction == "mmr"`).
#' @param max_iter A maximum number of iterations for the bisection procedure
#' (used only if `set_construction == "flr"` or `set_construction == "mmr"`).
#' @param data_type Data type for predictions with a custom ML model (`"data.frame"`, `"data.table"` or `"matrix"`;
#' used only if `object` is from [custom_rec_lin_model()]).
#' @param true_matches A `data.frame` or `data.table` indicating true matches.
#' @param verbose Logical indicating whether to print progress messages.
#' @param ... Additional controls passed to [predict.rec_lin_model()] for custom ML model
#' (used only if the `object` is from [custom_rec_lin_model()]).
#'
#' @details
#' The [predict.rec_lin_model()] method estimates the probability/density ratio
#' between matches and non-matches for pairs in given
#' datasets, based on a model obtained using the
#' [train_rec_lin()] or [custom_rec_lin_model()].
#' Then, it estimates the number of matches and
#' returns the predicted matches, using the maximum
#' entropy classification (MEC) algorithm
#' (see [Lee et al. (2022)](https://www150.statcan.gc.ca/n1/pub/12-001-x/2022001/article/00007-eng.htm)).
#'
#' The [predict.rec_lin_model()] method allows the construction of the predicted set
#' of matches using its estimated size or the bisection procedure,
#' described in [Lee et al. (2022)](https://www150.statcan.gc.ca/n1/pub/12-001-x/2022001/article/00007-eng.htm),
#' based on a target false link rate (FLR)
#' or missing match rate (MMR). To use the second option, set `set_construction = "flr"`
#' or `set_construction = "mmr"` and
#' specify a target error rate using the `target_rate` argument.
#'
#' By default, the function assumes that the datasets `newdata_A` and `newdata_B`
#' contain no duplicate records. This assumption
#' might be relaxed by allowing `newdata_A` to have duplicates. To do so,
#' set `duplicates_in_A = TRUE`.
#'
#' @return
#' Returns a list containing:\cr
#' \itemize{
#' \item{`M_est` -- a `data.table` with predicted matches,}
#' \item{`set_construction` -- a method for constructing the predicted set of matches,}
#' \item{`n_M_est` -- estimated classification set size,}
#' \item{`flr_est` -- estimated false link rate (FLR),}
#' \item{`mmr_est` -- estimated missing match rate (MMR),}
#' \item{`iter` -- the number of iterations in the bisection procedure,}
#' \item{`eval_metrics` -- standard metrics for quality assessment, if `true_matches` is provided,}
#' \item{`confusion` -- confusion matrix, if `true_matches` is provided.}
#' }
#'
#' @references
#' Lee, D., Zhang, L.-C. and Kim, J. K. (2022). Maximum entropy classification for record linkage.
#' Survey Methodology, Statistics Canada, Catalogue No. 12-001-X, Vol. 48, No. 1.
#'
#' Vo, T. H., Chauvet, G., Happe, A., Oger, E., Paquelet, S., and Garès, V. (2023).
#' Extending the Fellegi-Sunter record linkage model for mixed-type data with application to the French national health data system.
#' Computational Statistics & Data Analysis, 179, 107656.
#'
#' Sugiyama, M., Suzuki, T., Nakajima, S. et al. Direct importance estimation for covariate shift adaptation.
#' Ann Inst Stat Math 60, 699–746 (2008). \doi{10.1007/s10463-008-0197-x}
#'
#' @examples
#' df_1 <- data.frame(
#'   "name" = c("James", "Emma", "William", "Olivia", "Thomas",
#'   "Sophie", "Harry", "Amelia", "George", "Isabella"),
#'   "surname" = c("Smith", "Johnson", "Brown", "Taylor", "Wilson",
#'   "Davis", "Clark", "Harris", "Lewis", "Walker")
#' )
#'  df_2 <- data.frame(
#'   "name" = c("James", "Ema", "Wimliam", "Olivia", "Charlotte",
#'   "Henry", "Lucy", "Edward", "Alice", "Jack"),
#'   "surname" = c("Smith", "Johnson", "Bron", "Tailor", "Moore",
#'   "Evans", "Hall", "Wright", "Green", "King")
#' )
#' comparators <- list("name" = jarowinkler_complement(),
#'                     "surname" = jarowinkler_complement())
#' matches <- data.frame("a" = 1:4, "b" = 1:4)
#' methods <- list("name" = "continuous_nonparametric",
#'                 "surname" = "continuous_nonparametric")
#' model <- train_rec_lin(A = df_1, B = df_2, matches = matches,
#'                        variables = c("name", "surname"),
#'                        comparators = comparators,
#'                        methods = methods)
#'
#' df_new_1 <- data.frame(
#'   "name" = c("John", "Emily", "Mark", "Anna", "David"),
#'   "surname" = c("Smith", "Johnson", "Taylor", "Williams", "Brown")
#' )
#' df_new_2 <- data.frame(
#'   "name" = c("John", "Emely", "Mark", "Michael"),
#'   "surname" = c("Smitth", "Johnson", "Tailor", "Henders")
#' )
#' predict(model, df_new_1, df_new_2)
#' @export
predict.rec_lin_model <- function(object,
                                  newdata_A,
                                  newdata_B,
                                  duplicates_in_A = FALSE,
                                  set_construction = c("size", "flr", "mmr"),
                                  fixed_method = "Newton",
                                  target_rate = 0.03,
                                  tol = 0.005,
                                  max_iter = 50,
                                  data_type = c("data.frame", "data.table", "matrix"),
                                  true_matches = NULL,
                                  verbose = FALSE,
                                  ...) {

  stopifnot("`newdata_A` is required for predictions." =
              !missing(newdata_A))

  stopifnot("`newdata_B` is required for predictions." =
              !missing(newdata_B))

  if (missing(set_construction)) set_construction <- "size"
  if (missing(data_type)) data_type <- "data.frame"
  validate_choice(set_construction, c("size", "flr", "mmr"), "set_construction")
  validate_choice(data_type, c("data.frame", "data.table", "matrix"), "data_type")
  true_matches <- sanitize_true_matches(
    true_matches = true_matches,
    n_A = nrow(newdata_A),
    n_B = nrow(newdata_B),
    arg_name = "true_matches"
  )

  vectors <- comparison_vectors(A = newdata_A,
                                B = newdata_B,
                                variables = object$variables,
                                comparators = object$comparators)
  Omega <- vectors$Omega

  n <- NROW(Omega)
  if (verbose) {
    message(sprintf(
      "Created %d prediction comparison vector(s) for %d records in A and %d records in B.",
      n,
      NROW(newdata_A),
      NROW(newdata_B)
    ))
  }
  prob_est <- object$match_prop / max(NROW(newdata_A), NROW(newdata_B))

  if (!is.null(object$ml_model)) {

    ml_model <- object$ml_model
    gamma_variables <- paste0("gamma_", vectors$variables)

    if (data_type == "data.frame") {

      Omega_temp <- as.data.frame(Omega[, gamma_variables, with = FALSE])

    } else if (data_type == "matrix") {

      Omega_temp <- as.matrix(Omega[, gamma_variables, with = FALSE])

    } else {

      Omega_temp <- Omega

    }

    controls_ml_predictions <- list(...)

    predicted_probs <- do.call(
      predict,
      c(list(
        ml_model,
        Omega_temp
      ),
      controls_ml_predictions)
    )

    predicted_ratio <- predicted_probs * (1 - prob_est) / ((1 - predicted_probs) * prob_est)
    data.table::set(Omega, j = "ratio", value = predicted_ratio)

  } else {

    Omega <- score_mec_ratio(Omega, object)

  }

  n_M_start <- min(NROW(newdata_A), NROW(newdata_B))
  if (object$prob_ratio == "1" && is.null(object$ml_model)) {
    # Under ratio-I, r_q = m / q and the posterior match probability is pi * r_q.
    g_est <- pmin(prob_est * Omega[["ratio"]], 1)
    n_M_est <- min(sum(g_est), n_M_start)
  } else {
    fun_n_M <- fixed_n_M(n = n, ratio_gamma = Omega[["ratio"]])
    n_M_original <- FixedPoint::FixedPoint(Function = fun_n_M,
                                           Inputs = n_M_start,
                                           Method = fixed_method)$FixedPoint
    n_M_est <- min(n_M_original, n_M_start)
    n_M_est <- max(n_M_est, 0)
    n_M_est <- round(n_M_est)

    g_est <- pmin(n_M_est * Omega$ratio / (n_M_est * (Omega$ratio - 1) + n), 1)
  }
  if (verbose) {
    message(sprintf(
      "Scored candidate pairs and estimated %s match(es).",
      format(n_M_est, trim = TRUE)
    ))
  }
  selection_summary <- summarize_mec_selection(
    a = Omega[["a"]],
    b = Omega[["b"]],
    ratio = Omega[["ratio"]],
    g_est = g_est,
    n_M_est = n_M_est,
    duplicates_in_A = duplicates_in_A,
    set_construction = set_construction,
    target_rate = target_rate,
    tol = tol,
    max_iter = max_iter
  )
  M_est <- Omega[selection_summary$selected_idx, c("a", "b", "ratio")]
  flr_est <- selection_summary$flr_est
  mmr_est <- selection_summary$mmr_est
  iter <- selection_summary$iter
  if (verbose) {
    message(sprintf("Finished prediction with %d selected match(es).", NROW(M_est)))
  }

  if (!is.null(true_matches)) {
    eval <- evaluation(M_est, true_matches, n)
    eval_metrics <- unlist(get_metrics(
      TP = eval$TP,
      FP = eval$FP,
      FN = eval$FN,
      TN = eval$TN
    ))
    confusion <- get_confusion(
      TP = eval$TP,
      FP = eval$FP,
      FN = eval$FN,
      TN = eval$TN
    )

  }

  structure(
    list(
      M_est = M_est,
      set_construction = set_construction,
      n_M_est = n_M_est,
      flr_est = flr_est,
      mmr_est = mmr_est,
      iter = if (is.null(iter)) NULL else iter,
      eval_metrics = if (is.null(true_matches)) NULL else eval_metrics,
      confusion = if (is.null(true_matches)) NULL else confusion
    ),
    class = "rec_lin_predictions"
  )

}

#' @import data.table
#' @importFrom reclin2 cmp_identical
#'
#' @title Create Comparison Vectors for Record Linkage
#'
#' @author Adam Struzik
#'
#' @description
#' Creates comparison vectors between records in two datasets based on specified variables
#' and comparison functions.
#'
#' @param A A duplicate-free `data.frame` or `data.table`.
#' @param B A duplicate-free `data.frame` or `data.table`.
#' @param variables A character vector of key variables used to create comparison vectors.
#' @param comparators A named list of functions for comparing pairs of records.
#' @param pairs Optional. A `data.frame` or `data.table` with columns `a` and `b`
#' indicating pairs for which comparison vectors should be created.
#' @param matches Optional. A `data.frame` or `data.table` indicating known matches.
#'
#' @details
#' Consider two datasets: \eqn{A} and \eqn{B}.
#' For each pair of records \eqn{(a,b) \in \Omega},
#' the function creates a comparison vector
#' \eqn{\pmb{\gamma}_{ab} = (\gamma_{ab}^1,\gamma_{ab}^2,\ldots,\gamma_{ab}^K)'}
#' based on specified \eqn{K} variables and comparison functions.
#'
#' @note
#' Each comparison function must return another function,
#' which serves as the actual comparator.
#'
#' @return
#' Returns a list containing:\cr
#' \itemize{
#' \item{`Omega` -- a `data.table` with comparison vectors between records from both datasets,
#' including optional match information,}
#' \item{`variables` -- a character vector of key variables used for comparison,}
#' \item{`comparators` -- a list of functions used to compare pairs of records,}
#' \item{`match_prop` -- proportion of matches in the smaller dataset.}
#' }
#'
#' @examples
#' df_1 <- data.frame(
#' "name" = c("John", "Emily", "Mark", "Anna", "David"),
#' "surname" = c("Smith", "Johnson", "Taylor", "Williams", "Brown")
#' )
#' df_2 <- data.frame(
#'   "name" = c("Jon", "Emely", "Marc", "Michael"),
#'   "surname" = c("Smitth", "Jonson", "Tailor", "Henderson")
#' )
#' comparators <- list("name" = jarowinkler_complement(),
#'                     "surname" = jarowinkler_complement())
#' matches <- data.frame("a" = 1:3, "b" = 1:3)
#' result <- comparison_vectors(A = df_1, B = df_2, variables = c("name", "surname"),
#'                              comparators = comparators, matches = matches)
#' result
#' @export
comparison_vectors <- function(
    A,
    B,
    variables,
    comparators = NULL,
    pairs = NULL,
    matches = NULL) {

  stopifnot("`A` should be a data.frame or a data.table." =
              is.data.frame(A) | is.data.table(A))
  stopifnot("`B` should be a data.frame or a data.table." =
              is.data.frame(B) | is.data.table(B))
  stopifnot("`variables` should be a character vector." =
              is.character(variables))
  stopifnot("`variables` should contain at least one variable." =
              length(variables) > 0L)
  stopifnot("Not all variables are present in A." =
              all(variables %in% names(A)))
  stopifnot("Not all variables are present in B." =
              all(variables %in% names(B)))

  if (!is.null(comparators)) {
    stopifnot("`comparators` should be a list." =
                is.list(comparators))
  }

  K <- length(variables)

  missing_variables <- variables[!(variables %in% names(comparators))]
  comparators[missing_variables] <- rep(list(reclin2::cmp_identical()),
                                        length(missing_variables))
  comparators <- comparators[variables]

  data.table::setDT(A)
  data.table::setDT(B)
  n_A <- nrow(A)
  n_B <- nrow(B)

  if (!is.null(pairs)) {
    if (!(is.data.frame(pairs) || is.data.table(pairs))) {
      stop("`pairs` should be a data.frame or a data.table.")
    }

    pairs <- data.table::copy(data.table::as.data.table(pairs))

    if (!all(c("a", "b") %in% names(pairs))) {
      stop("`pairs` should contain columns: a, b.")
    }

    if (anyNA(pairs[["a"]]) || anyNA(pairs[["b"]])) {
      stop("`pairs` cannot contain missing values.")
    }

    if (!is.numeric(pairs[["a"]]) || !is.numeric(pairs[["b"]])) {
      stop("`pairs` should contain numeric row indices in columns a and b.")
    }

    if (any(pairs[["a"]] != as.integer(pairs[["a"]])) ||
        any(pairs[["b"]] != as.integer(pairs[["b"]]))) {
      stop("`pairs` should contain integer row indices in columns a and b.")
    }

    if (any(pairs[["a"]] < 1) || any(pairs[["b"]] < 1)) {
      stop("`pairs` should contain positive row indices in columns a and b.")
    }

    if (any(pairs[["a"]] > n_A) || any(pairs[["b"]] > n_B)) {
      stop("`pairs` contains row indices outside the input datasets.")
    }

    if (anyDuplicated(pairs[, c("a", "b"), with = FALSE]) > 0L) {
      stop("`pairs` should not contain duplicate record pairs.")
    }
  }

  if (!is.null(matches)) {
    validate_match_pairs(matches, n_A, n_B)
  }

  if (is.null(pairs)) {
    Omega <- data.table::CJ(a = seq_len(n_A), b = seq_len(n_B))
    data.table::setkey(Omega, NULL)
  } else {
    Omega <- pairs
  }

  gamma_names <- paste0("gamma_", variables)
  omega_a <- Omega[["a"]]
  omega_b <- Omega[["b"]]

  # Compute one comparison column at a time to avoid copying the full key-variable blocks.
  gamma_list <- lapply(1:K, function(x) {
    variable <- variables[x]
    as.numeric(comparators[[x]](A[[variable]][omega_a], B[[variable]][omega_b]))
  })

  invalid_counts <- vapply(gamma_list, function(gamma) {
    sum(!is.finite(gamma))
  }, numeric(1))

  if (any(invalid_counts > 0)) {
    invalid_vars <- variables[invalid_counts > 0]
    invalid_details <- sprintf(
      "%s (%d invalid value%s)",
      invalid_vars,
      invalid_counts[invalid_counts > 0],
      ifelse(invalid_counts[invalid_counts > 0] == 1, "", "s")
    )
    stop(
      sprintf(
        "Comparison variables produced missing or non-finite values: %s. Please handle missing key values or adjust comparators before running record linkage.",
        paste(invalid_details, collapse = ", ")
      ),
      call. = FALSE
    )
  }

  Omega[, (gamma_names) := gamma_list]

  if (!is.null(matches)) {
    data.table::setDT(matches)
    match_idx <- data.table(
      omega_idx = seq_len(NROW(Omega)),
      a = Omega[["a"]],
      b = Omega[["b"]]
    )[matches[, c("a", "b"), with = FALSE], on = c("a", "b"), nomatch = 0L][["omega_idx"]]
    Omega[, match := 0]
    Omega[match_idx, match := 1]
  }

  structure(
    list(
      Omega = data.table(Omega),
      variables = variables,
      comparators = comparators,
      match_prop = if (is.null(matches)) NULL else NROW(matches) / NROW(Omega) * max(n_A, n_B)
    ),
    class = "comparison_vectors"
  )

}

library(automatedRecLin)
library(data.table)

df_1 <- data.frame(
  "name" = c("James", "Emma", "William", "Olivia", "Thomas",
             "Sophie", "Harry", "Amelia", "George", "Isabella"),
  "surname" = c("Smith", "Johnson", "Brown", "Taylor", "Wilson",
                "Davis", "Clark", "Harris", "Lewis", "Walker")
)
df_2 <- data.frame(
  "name" = c("James", "Ema", "Wimliam", "Olivia", "Charlotte",
             "Henry", "Lucy", "Edward", "Alice", "Jack"),
  "surname" = c("Smith", "Johnson", "Bron", "Tailor", "Moore",
                "Evans", "Hall", "Wright", "Green", "King")
)
comparators <- list("name" = jarowinkler_complement(),
                    "surname" = jarowinkler_complement())
matches <- data.frame("a" = 1:4, "b" = 1:4)

expect_equal(
  head(comparison_vectors(A = df_1, B = df_2, variables = c("name", "surname"), matches = matches)$Omega),
  data.table("a" = rep(1, 6), "b" = 1:6, "gamma_name" = c(1, rep(0, 5)),
             "gamma_surname" = c(1, rep(0, 5)), match = c(1, rep(0, 5)))
)

expect_equal(
  head(comparison_vectors(A = df_1, B = df_2, variables = c("name", "surname"),
                          matches = matches, comparators = comparators)$Omega),
  data.table("a" = rep(1, 6), "b" = 1:6,
             "gamma_name" = c(0, 0.4777777777777778567270, 0.5523809523809524169025,
                              1, 0.5629629629629629983256, 1),
             "gamma_surname" = c(0, 0.5523809523809524169025, 1,
                                 0.5444444444444445085907, 1, 1),
             "match" = c(1, rep(0, 5)))
)

#' @importFrom utils head
#' @method print comparison_vectors
#' @exportS3Method
print.comparison_vectors <- function(x, ...) {

  cat("Comparison based on the following variables: ", paste(x$variables, collapse = ", "), ".\n", sep = "")
  if (!("match" %in% colnames(x$Omega))) {
    cat("True matches are unknown.\n")
  }
  cat("========================================================\n")
  print(head(x$Omega))

}

#' @importFrom utils head
#' @method print rec_lin_model
#' @exportS3Method
print.rec_lin_model <- function(x, ...) {

  cat("Record linkage model based on the following variables: ", paste(x$variables, collapse = ", "), ".\n", sep = "")

  if (!is.null(x$ml_model)) {
    cat("A custom ML model was used.\n")
  }

  cat("The prior probability of matching is ", x$pi_est, ".\n", sep = "")

  if (!is.null(x$b_vars)) {
    cat("========================================================\n")
    cat("Variables selected for the binary method: ", paste(substring(x$b_vars, 7), collapse = ", "), ".\n", sep = "")
    cat("Estimated parameters for the binary method:\n")
    print(x$b_params)
  }

  if (!is.null(x$cpar_vars)) {
    cat("========================================================\n")
    cat("Variables selected for the continuous parametric method: ", paste(substring(x$cpar_vars, 7), collapse = ", "), ".\n", sep = "")
    cat("Estimated parameters for the continuous parametric method:\n")
    print(x$cpar_params)
  }

  if (!is.null(x$cnonpar_vars)) {
    cat("========================================================\n")
    cat("Variables selected for the continuous nonparametric method: ", paste(substring(x$cnonpar_vars, 7), collapse = ", "), ".\n", sep = "")
  }

  cat("========================================================\n")
  cat("Probability/density ratio type: ", x$prob_ratio, ".\n", sep = "")

}

# Keep printed evaluation errors on the same percentage scale as estimated rates.
format_linkage_error_rates <- function(x) {
  formatted_rates <- sprintf("%.4f", x * 100)
  names(formatted_rates) <- paste0(names(x), " (%)")
  noquote(formatted_rates)
}

#' @import data.table
#' @importFrom utils head
#' @method print rec_lin_predictions
#' @exportS3Method
print.rec_lin_predictions <- function(x, ...) {

  if (NROW(x$M_est) == 0) {
    cat("No matches were predicted.\n")
  } else {
    M_est <- data.table::copy(x$M_est)
    data.table::set(M_est, j = "ratio / 1000", value = M_est[["ratio"]] / 1000)
    data.table::set(M_est, j = "ratio", value = NULL)
    M_est_head <- head(M_est, 6)
    cat("The algorithm predicted", NROW(M_est), "matches.\n")
    cat("The first", NROW(M_est_head), "predicted matches are:\n")
    print(M_est_head)
    cat("========================================================\n")
    if (x$set_construction == "size") {
      cat("The construction of the classification set was based on estimates of its size.\n")
    } else if (x$set_construction == "flr") {
      cat("The construction of the classification set was based on the target false link rate (FLR).\n")
      cat("The bisection procedure ended after", x$iter, "iterations.\n")
    } else if (x$set_construction == "mmr") {
      cat("The construction of the classification set was based on the target missing match rate (MMR).\n")
      cat("The bisection procedure ended after", x$iter, "iterations.\n")
    }
    cat("Estimated false link rate (FLR): ", sprintf("%.4f", x$flr_est * 100), " %.\n", sep = "")
    if (x$n_M_est != 0) {
      cat("Estimated missing match rate (MMR): ", sprintf("%.4f", x$mmr_est * 100), " %.\n", sep = "")
    } else {
      cat("Missing match rate (MMR) cannot be estimated because the estimated classification set size is equal to 0.")
    }
  }

  if (!is.null(x$eval_metrics)) {
    cat("========================================================\n")
    cat("Evaluation metrics:\n")
    print(format_linkage_error_rates(x$eval_metrics))
  }

}

#' @import data.table
#' @importFrom utils head
#' @method print mec_rec_lin
#' @exportS3Method
print.mec_rec_lin <- function(x, ...) {

  cat("Record linkage based on the following variables: ", paste(x$variables, collapse = ", "), ".\n", sep = "")
  cat("========================================================\n")
  if (NROW(x$M_est) == 0) {
    cat("No matches were predicted.\n")
  } else {
    M_est <- data.table::copy(x$M_est)
    data.table::set(M_est, j = "ratio / 1000", value = M_est[["ratio"]] / 1000)
    data.table::set(M_est, j = "ratio", value = NULL)
    M_est_head <- head(M_est, 6)
    cat("The algorithm predicted", NROW(M_est), "matches.\n")
    cat("The first", NROW(M_est_head), "predicted matches are:\n")
    print(M_est_head)
    cat("========================================================\n")
  }
  if (x$set_construction == "size") {
    cat("The construction of the classification set was based on estimates of its size.\n")
  } else if (x$set_construction == "flr") {
    cat("The construction of the classification set was based on the target false link rate (FLR).\n")
    cat("The bisection procedure ended after", x$iter_bisection, "iterations.\n")
  } else if (x$set_construction == "mmr") {
    cat("The construction of the classification set was based on the target missing match rate (MMR).\n")
    cat("The bisection procedure ended after", x$iter_bisection, "iterations.\n")
  }
  cat("Estimated false link rate (FLR): ", sprintf("%.4f", x$flr_est * 100), " %.\n", sep = "")
  if (!is.null(x$mmr_est)) {
    cat("Estimated missing match rate (MMR): ", sprintf("%.4f", x$mmr_est * 100), " %.\n", sep = "")
  }

  if (!is.null(x$b_vars)) {
    cat("========================================================\n")
    cat("Variables selected for the binary method: ", paste(substring(x$b_vars, 7), collapse = ", "), ".\n", sep = "")
    cat("Estimated parameters for the binary method:\n")
    print(x$b_params)
  }

  if (!is.null(x$cpar_vars)) {
    cat("========================================================\n")
    cat("Variables selected for the continuous parametric method: ", paste(substring(x$cpar_vars, 7), collapse = ", "), ".\n", sep = "")
    cat("Estimated parameters for the continuous parametric method:\n")
    print(x$cpar_params)
  }

  if (!is.null(x$cnonpar_vars)) {
    cat("========================================================\n")
    cat("Variables selected for the continuous nonparametric method: ", paste(substring(x$cnonpar_vars, 7), collapse = ", "), ".\n", sep = "")
  }

  if (!is.null(x$hm_params)) {
    cat("========================================================\n")
    cat("Variables selected for the hit-miss method: ", paste(substring(x$hm_vars, 7), collapse = ", "), ".\n", sep = "")
    cat("Estimated parameters for the hit-miss method:\n")
    print(x$hm_params)
  }

  if (!is.null(x$eval_metrics)) {
    cat("========================================================\n")
    cat("Evaluation metrics:\n")
    print(format_linkage_error_rates(x$eval_metrics))
  }

}

#' @import data.table
#' @importFrom utils head
#' @method print mec_blocking
#' @exportS3Method
print.mec_blocking <- function(x, ...) {

  cat("Blocked MEC record linkage based on: ", "\n")
  cat(paste(x$variables, collapse = ", "), ".\n", sep = "")
  cat("========================================================\n")

  if (NROW(x$M_est) == 0) {
    cat("No matches were predicted.\n")
  } else {
    M_est <- data.table::copy(x$M_est)
    data.table::set(M_est, j = "ratio / 1000", value = M_est[["ratio"]] / 1000)
    data.table::set(M_est, j = "ratio", value = NULL)
    M_est_head <- head(M_est, 6)
    cat("The algorithm predicted", NROW(M_est), "matches.\n")
    cat("The first", NROW(M_est_head), "predicted matches are:\n")
    print(M_est_head)
    cat("========================================================\n")
  }

  if (!is.null(x$blocking_eval)) {
    cat("========================================================\n")
    cat("Blocking diagnostics:\n")
    cat("Known matches: ", x$blocking_eval[["true_matches"]], ".\n", sep = "")
    cat("Known matches retained by blocking: ", x$blocking_eval[["preserved_matches"]], ".\n", sep = "")
    cat("Known matches missed by blocking: ", x$blocking_eval[["lost_matches"]], ".\n", sep = "")
    cat("Blocking MMR: ", sprintf("%.4f", x$blocking_eval[["blocking_fnr"]] * 100), " %.\n", sep = "")
    cat("Candidate pairs retained: ", x$blocking_eval[["blocked_pairs"]],
        " of ", x$blocking_eval[["full_pairs"]], ".\n", sep = "")
    pair_reduction <- 1 - x$blocking_eval[["blocked_pairs"]] / x$blocking_eval[["full_pairs"]]
    cat("Candidate pair reduction: ", sprintf("%.4f", pair_reduction * 100), " %.\n", sep = "")
  }

  if (!is.null(x$eval_metrics)) {
    cat("========================================================\n")
    cat("Evaluation metrics:\n")
    print(format_linkage_error_rates(x$eval_metrics))
  }

}

#' @import data.table
#' @importFrom stats runif
#' @importFrom stats predict
#' @importFrom purrr partial
#' @importFrom utils head
#'
#' @title Unsupervised Maximum Entropy Classifier for Record Linkage
#'
#' @author Adam Struzik
#'
#' @description
#' Implements several extensions to the maximum entropy classification (MEC) algorithm for record linkage
#' (see [Lee et al. (2022)](https://www150.statcan.gc.ca/n1/pub/12-001-x/2022001/article/00007-eng.htm)),
#' iteratively estimating probability/density ratios to classify record pairs into matches and non-matches
#' based on comparison vectors.
#'
#' @param A A duplicate-free `data.frame` or `data.table`.
#' @param B A duplicate-free `data.frame` or `data.table`.
#' @param variables A character vector of key variables used to create comparison vectors.
#' @param comparators A named list of functions for comparing pairs of records.
#' @param methods A named list of methods used for estimation (`"binary"`, `"continuous_parametric"`, `"continuous_nonparametric"` or `"hit_miss"`).
#' @param duplicates_in_A Logical indicating whether to allow `A` to have duplicate records.
#' @param start_params Start parameters for the `"binary"`, `"continuous_parametric"` and `"hit_miss"` methods.
#' @param nonpar_hurdle Logical indicating whether to use a hurdle model or not
#' (used only if the `"continuous_nonparametric"` method has been chosen for at least one variable).
#' @param set_construction A method for constructing the predicted set of matches (`"size"`, `"flr"` or `"mmr"`).
#' @param target_rate A target false link rate (FLR) or missing match rate
#' (MMR) (used only if `set_construction == "flr"` or `set_construction == "mmr"`).
#' @param max_iter_bisection A maximum number of iterations for the bisection procedure
#' (used only if `set_construction == "flr"` or `set_construction == "mmr"`).
#' @param tol Error tolerance in the bisection procedure
#' (used only if `set_construction == "flr"` or `set_construction == "mmr"`).
#' @param delta A numeric value specifying the tolerance for the change in the estimated number of matches between iterations.
#' @param eps A numeric value specifying the tolerance for the change in model parameters between iterations.
#' @param max_iter_em A maximum number of iterations for the EM algorithm
#' (used only if the `"hit_miss"` method has been chosen for at least one variable).
#' @param tol_em Error tolerance in the EM algorithm
#' (used only if the `"hit_miss"` method has been chosen for at least one variable).
#' @param controls_nleqslv Controls passed to the \link[nleqslv:nleqslv]{nleqslv()} function
#' (only if the `"continuous_parametric"` method has been chosen for at least one variable).
#' @param controls_kliep Controls passed to the \link[densityratio:kliep]{kliep()} function
#' (only if the `"continuous_nonparametric"` method has been chosen for at least one variable).
#' @param true_matches A `data.frame` or `data.table` indicating known matches.
#' @param verbose Logical indicating whether to print progress messages.
#'
#' @details
#' Consider two datasets without duplicates: \eqn{A} and \eqn{B}.
#' Let the bipartite comparison space \eqn{\Omega = A \times B} consist of
#' matches \eqn{M} and non-matches \eqn{U} between the records in files
#' \eqn{A} and \eqn{B}. For any pair of records \eqn{(a,b) \in \Omega},
#' let \eqn{\pmb{\gamma}_{ab} = (\gamma_{ab}^1,\gamma_{ab}^2,
#' \ldots,\gamma_{ab}^K)'} be the comparison vector between
#' a set of key variables. The original MEC algorithm uses the binary
#' comparison function to evaluate record pairs across two datasets.
#' However, this approach may be insufficient when handling datasets
#' with frequent errors across multiple variables.
#'
#' We propose the use of continuous comparison functions to address
#' the limitations of binary comparison methods. We consider every
#' semi-metric, i.e., a function \eqn{d: A \times B \to \mathbb{R}},
#' satisfying the following conditions:\cr
#' \enumerate{
#' \item{\eqn{d(x,y) \geq 0},}
#' \item{\eqn{d(x,y) = 0} if and only if \eqn{x = y},}
#' \item{\eqn{d(x,y) = d(y,x)}.}
#' }
#' For example, we can use the Jaro-Winkler distance for character variables
#' (which is implemented in the `automatedRecLin` package as [jarowinkler_complement()])
#' or the Euclidean distance for numerical variables. The `automatedRecLin` package allows the use of
#' a different comparison function for each key variable (which should be specified
#' as a list in the `comparators` argument). The default function
#' for each key variable is \link[reclin2:cmp_identical]{cmp_identical()}
#' (the binary comparison function).
#'
#' The [mec()] function offers different approaches to estimate the
#' probability/density ratio between matches and non-matches,
#' which should be specified as a list in the `methods` argument.
#' The available methods suitable for the binary comparison function
#' are `"binary"` and `"hit_miss"`. Both assume that \eqn{\gamma_{ab}^k|M}
#' and \eqn{\gamma_{ab}^k|U} follow Bernoulli distributions.
#' `"binary"` and `"hit_miss"` both estimate the parameters for the matches iteratively,
#' but `"binary"` estimates the parameters for the non-matches
#' only at the start, while `"hit_miss"` does
#' so iteratively using a hit-miss model (for details see
#' [Lee et al. (2022)](https://www150.statcan.gc.ca/n1/pub/12-001-x/2022001/article/00007-eng.htm)).
#' `"binary"` is the default method for each variable.
#'
#' For the continuous semi-metrics we suggest the usage
#' of `"continuous_parametric"` or `"continuous_nonparametric"`
#' method. The `"continuous_parametric"` method assumes that
#' \eqn{\gamma_{ab}^k|M} and \eqn{\gamma_{ab}^k|U} follow
#' hurdle Gamma distributions. The density function of a hurdle
#' Gamma distribution is characterized by three parameters
#' \eqn{p_0 \in (0,1)} and \eqn{\alpha, \beta > 0} as follows:
#' \deqn{
#' f(x;p_0,\alpha,\beta) = p_0^{\mathbb{I}(x = 0)}[(1 - p_0) v(x;\alpha,\beta)]^{\mathbb{I}(x > 0)},
#' }
#' where
#' \deqn{
#' v(x;\alpha,\beta) = \frac{\beta^{\alpha} x^{\alpha - 1} \exp(-\beta x)}
#' {\Gamma(\alpha)}
#' }
#' is the density function of a Gamma distribution
#' (for details see [Vo et al. (2023)](https://ideas.repec.org/a/eee/csdana/v179y2023ics0167947322002365.html)).
#' At the beginning, the algorithm estimates the parameters for the non-matches
#' and then does it iteratively for the matches.
#' The `"continuous_nonparametric"` method does not assume anything about
#' the distributions of the comparison vectors. It iteratively directly
#' estimates the density ratio between the matches and the non-matches, using
#' the Kullback-Leibler Importance Estimation Procedure (KLIEP).
#' For details see [Sugiyama et al. (2008)](https://link.springer.com/article/10.1007/s10463-008-0197-x).
#'
#' The [mec()] function allows the construction of the predicted set
#' of matches using its estimated size or the bisection procedure,
#' described in [Lee et al. (2022)](https://www150.statcan.gc.ca/n1/pub/12-001-x/2022001/article/00007-eng.htm),
#' based on a target false link rate (FLR)
#' or missing match rate (MMR). To use the second option, set `set_construction = "flr"`
#' or `set_construction = "mmr"` and
#' specify a target error rate using the `target_rate` argument.
#'
#' The assumption that \eqn{A} and \eqn{B} contain no duplicate records
#' might be relaxed by allowing \eqn{A} to have duplicates. To do so,
#' set `duplicates_in_A = TRUE`.
#'
#' @return
#' Returns a list containing:\cr
#' \itemize{
#' \item{`M_est` -- a `data.table` with predicted matches,}
#' \item{`n_M_est` -- estimated classification set size,}
#' \item{`flr_est` -- estimated false link rate (FLR),}
#' \item{`mmr_est` -- estimated missing match rate (MMR),}
#' \item{`iter_bisection` -- the number of iterations in the bisection procedure,}
#' \item{`b_vars` -- a character vector of variables used for the `"binary"` method (with the prefix `"gamma_"`),}
#' \item{`cpar_vars` -- a character vector of variables used for the `"continuous_parametric"` method (with the prefix `"gamma_"`),}
#' \item{`cnonpar_vars` -- a character vector of variables used for the `"continuous_nonparametric"` method (with the prefix `"gamma_"`),}
#' \item{`hm_vars` -- a character vector of variables used for the `"hit_miss"` method (with the prefix `"gamma_"`),}
#' \item{`b_params` -- parameters estimated using the `"binary"` method,}
#' \item{`cpar_params` -- parameters estimated using the `"continuous_parametric"` method,}
#' \item{`hm_params` -- parameters estimated using the `"hit_miss"` method,}
#' \item{`ratio_kliep` -- a result of the \link[densityratio:kliep]{kliep()} function,}
#' \item{`variables` -- a character vector of key variables used for comparison,}
#' \item{`set_construction` -- a method for constructing the predicted set of matches,}
#' \item{`eval_metrics` -- standard metrics for quality assessment (if `true_matches` is provided),}
#' \item{`confusion` -- confusion matrix (if `true_matches` is provided).}
#' }
#'
#' @references
#' Lee, D., Zhang, L.-C. and Kim, J. K. (2022). Maximum entropy classification for record linkage.
#' Survey Methodology, Statistics Canada, Catalogue No. 12-001-X, Vol. 48, No. 1.
#'
#' Vo, T. H., Chauvet, G., Happe, A., Oger, E., Paquelet, S., and Garès, V. (2023).
#' Extending the Fellegi-Sunter record linkage model for mixed-type data with application to the French national health data system.
#' Computational Statistics & Data Analysis, 179, 107656.
#'
#' Sugiyama, M., Suzuki, T., Nakajima, S. et al. Direct importance estimation for covariate shift adaptation.
#' Ann Inst Stat Math 60, 699–746 (2008). \doi{10.1007/s10463-008-0197-x}
#'
#' @examples
#' df_1 <- data.frame(
#'   name = c("Emma", "Liam", "Olivia", "Noah", "Ava",
#'            "Ethan", "Sophia", "Mason", "Isabella", "James"),
#'   surname = c("Smith", "Johnson", "Williams", "Brown", "Jones",
#'               "Garcia", "Miller", "Davis", "Rodriguez", "Wilson"),
#'   city = c("New York", "Los Angeles", "Chicago", "Houston", "Phoenix",
#'            "Philadelphia", "San Antonio", "San Diego", "Dallas", "San Jose")
#' )
#'
#' df_2 <- data.frame(
#'   name = c(
#'     "Emma", "Liam", "Olivia", "Noah",
#'     "Ava", "Ehtan", "Sopia", "Mson",
#'     "Charlotte", "Benjamin", "Amelia", "Lucas"
#'   ),
#'   surname = c(
#'      "Smith", "Johnson", "Williams", "Brown",
#'     "Jnes", "Garca", "Miler", "Dvis",
#'     "Martinez", "Lee", "Hernandez", "Clark"
#'   ),
#'   city = c(
#'     "New York", "Los Angeles", "Chicago", "Houston",
#'     "Phonix", "Philadelpia", "San Antnio", "San Dieg",
#'     "Seattle", "Miami", "Boston", "Denver"
#'   )
#' )
#' true_matches <- data.frame(
#'   "a" = 1:8,
#'   "b" = 1:8
#' )
#'
#' variables <- c("name", "surname", "city")
#' comparators <- list(
#'   "name" = jarowinkler_complement(),
#'   "surname" = jarowinkler_complement(),
#'   "city" = jarowinkler_complement()
#' )
#' methods <- list(
#'   "name" = "continuous_parametric",
#'   "surname" = "continuous_parametric",
#'   "city" = "continuous_parametric"
#' )
#'
#' set.seed(1)
#' result <- mec(A = df_1, B = df_2,
#'               variables = variables,
#'               comparators = comparators,
#'               methods = methods,
#'               true_matches = true_matches)
#' result
#' @export
mec <- function(A,
                B,
                variables,
                comparators = NULL,
                methods = NULL,
                duplicates_in_A = FALSE,
                start_params = NULL,
                nonpar_hurdle = TRUE,
                set_construction = NULL,
                target_rate = 0.03,
                max_iter_bisection = 100,
                tol = 0.005,
                delta = 0.5,
                eps = 0.05,
                max_iter_em = 10,
                tol_em = 1,
                controls_nleqslv = list(),
                controls_kliep = control_kliep(),
                true_matches = NULL,
                verbose = FALSE) {

  if (!is.null(start_params)) {
    stopifnot("`start_params` should be a list." =
                is.list(start_params))
  }

  if (is.null(set_construction)) {
    set_construction <- "size"
  }
  validate_choice(set_construction, c("size", "flr", "mmr"), "set_construction")
  methods <- validate_methods(
    methods = methods,
    variables = variables,
    allowed_methods = c("binary", "continuous_parametric", "continuous_nonparametric", "hit_miss")
  )

  data.table::setDT(A)
  data.table::setDT(B)
  true_matches <- sanitize_true_matches(
    true_matches = true_matches,
    n_A = nrow(A),
    n_B = nrow(B),
    arg_name = "true_matches"
  )
  stopifnot("There are no records with perfect agreement on the key variables.
            Please provide relevant datasets." =
              has_perfect_agreement(A, B, variables))
  if (verbose) {
    message(sprintf(
      "Preparing unsupervised MEC for %d records in A, %d records in B, and %d key variable(s).",
      nrow(A),
      nrow(B),
      length(variables)
    ))
  }
  data.table::set(A, j = "a", value = seq_len(nrow(A)))
  data.table::set(B, j = "b", value = seq_len(nrow(B)))
  M <- merge(
    A[, c("a", variables), with = FALSE],
    B[, c("b", variables), with = FALSE],
    by = variables
  )

  preprocessing <- drop_constant_key_variables(
    A = A,
    B = B,
    variables = variables,
    comparators = comparators,
    methods = methods
  )
  A <- preprocessing$A
  B <- preprocessing$B
  variables <- preprocessing$variables
  comparators <- preprocessing$comparators
  methods <- preprocessing$methods
  constant_vars <- preprocessing$constant_vars

  for (var in constant_vars) {
    warning(paste("The variable", var, "has only one unique value and has been removed."))
  }

  vectors <- comparison_vectors(A = A,
                                B = B,
                                variables = variables,
                                comparators = comparators)
  Omega <- vectors$Omega
  comparators <- vectors$comparators
  method_variables <- extract_method_variables(methods, include_hit_miss = TRUE)
  b_vars <- method_variables$b_vars
  cpar_vars <- method_variables$cpar_vars
  cnonpar_vars <- method_variables$cnonpar_vars
  hm_vars <- method_variables$hm_vars

  ratio_kliep <- NULL
  kliep_warning_emitted <- FALSE

  warn_kliep_once <- function(variables, message) {
    if (!kliep_warning_emitted) {
      warn_kliep_issue("mec()", variables, message)
      kliep_warning_emitted <<- TRUE
    }
  }

  exact_match_idx <- data.table(
    omega_idx = seq_len(NROW(Omega)),
    a = Omega[["a"]],
    b = Omega[["b"]]
  )[M[, c("a", "b"), with = FALSE], on = c("a", "b"), nomatch = 0L][["omega_idx"]]
  n <- NROW(Omega)
  if (verbose) {
    message(sprintf(
      "Created %d comparison vector(s) and initialized MEC with %d exact-agreement seed match(es).",
      n,
      length(exact_match_idx)
    ))
  }
  exact_match_mask <- rep(FALSE, n)
  exact_match_mask[exact_match_idx] <- TRUE
  M_idx <- exact_match_idx
  M <- Omega[M_idx]
  n_M <- length(M_idx)
  all_idx <- seq_len(n)
  selection_mask <- rep(FALSE, n)
  data.table::set(Omega, j = "ratio", value = 1)

  if (is.null(start_params)) {

    start_params <- list()

    if (length(b_vars) > 0) {
      start_params[["binary"]] <- data.table(
        variable = b_vars,
        theta = runif(length(b_vars), min = 0.9)
      )

    }

    if (length(cpar_vars) > 0) {
      start_params[["continuous_parametric"]] <- data.table(
        variable = cpar_vars,
        p_0_M = runif(length(cpar_vars), min = 0.8, max = 0.9),
        alpha_M = runif(length(cpar_vars), min = 0.1, max = 1),
        beta_M = runif(length(cpar_vars), min = 10, max = 20)
      )
    }

    if (length(hm_vars) > 0) {
      start_params[["hit_miss"]] <- data.table(
        variable = hm_vars,
        theta = runif(length(hm_vars), min = 0.9)
      )

    }

  }

  if (length(b_vars) > 0) {

    b_params <- align_parameter_table(start_params$binary, b_vars)
    Omega_b <- Omega[, b_vars, with = FALSE]

    eta_b <- binary_formula(Omega_b)
    b_params$eta <- eta_b

    b_denominator <- bernoulli_product(Omega_b, b_params$eta)
    data.table::set(
      Omega,
      j = "ratio",
      value = Omega[["ratio"]] * bernoulli_ratio(Omega_b, b_params$theta, b_params$eta)
    )
    data.table::set(Omega, j = "b_denominator", value = b_denominator)

  }

  if (length(cpar_vars) > 0) {

    cpar_params <- align_parameter_table(start_params$continuous_parametric, cpar_vars)
    Omega_cpar <- Omega[, cpar_vars, with = FALSE]

    p_0_U <- p_0_formula(Omega_cpar)
    gamma_plus_U <- gamma_plus_formula(Omega_cpar)
    modified_nleqslv <- purrr::partial(nleqslv::nleqslv, control = controls_nleqslv)
    alpha_U <- alpha_formula(Omega_cpar, modified_nleqslv)
    beta_U <- alpha_U / gamma_plus_U
    cpar_params$p_0_U <- p_0_U
    cpar_params$alpha_U <- alpha_U
    cpar_params$beta_U <- beta_U

    cpar_denominator <- hurdle_gamma_product(
      Omega_cpar,
      cpar_params$p_0_U,
      cpar_params$alpha_U,
      cpar_params$beta_U
    )
    data.table::set(
      Omega,
      j = "ratio",
      value = Omega[["ratio"]] * hurdle_gamma_ratio(
        Omega_cpar,
        cpar_params$p_0_M,
        cpar_params$alpha_M,
        cpar_params$beta_M,
        cpar_params$p_0_U,
        cpar_params$alpha_U,
        cpar_params$beta_U
      )
    )
    data.table::set(Omega, j = "cpar_denominator", value = cpar_denominator)

  }

  if (length(cnonpar_vars) > 0) {

    Omega_cnonpar <- Omega[, cnonpar_vars, with = FALSE]
    n_exact_matches <- sum(exact_match_mask)

    if (nonpar_hurdle) {

      p_0_M_cnonpar <- runif(length(cnonpar_vars), min = 0.5)
      p_0_U_cnonpar <- p_0_formula(Omega_cnonpar)
      names(p_0_M_cnonpar) <- cnonpar_vars

      ratio_temp <- lapply(cnonpar_vars, function(x) {
        r <- numeric(n)
        r[exact_match_mask] <- stats::runif(n_exact_matches, min = 5, max = 10)
        r[!exact_match_mask] <- stats::runif(n - n_exact_matches, min = 0.1, max = 1)
        r
      })
      names(ratio_temp) <- cnonpar_vars
      cnonpar_ratio_list <- lapply(cnonpar_vars, function(x) {
        gamma_vec <- Omega_cnonpar[[x]]
        ifelse(gamma_vec == 0, p_0_M_cnonpar[x] / p_0_U_cnonpar[x], 1) *
          ifelse(gamma_vec > 0, (1 - p_0_M_cnonpar[x]) / (1 - p_0_U_cnonpar[x]) * ratio_temp[[x]], 1)
      })
      ratio_kliep <- Reduce(`*`, cnonpar_ratio_list)
      data.table::set(Omega, j = "ratio", value = Omega[["ratio"]] * ratio_kliep)

    } else {

      ratio_temp <- as.numeric(Omega$ratio)
      ratio_temp[exact_match_mask] <- ratio_temp[exact_match_mask] * stats::runif(n_exact_matches,
                                                                                  min = 5, max = 10)
      ratio_temp[!exact_match_mask] <- ratio_temp[!exact_match_mask] * stats::runif(n - n_exact_matches,
                                                                                    min = 0.1, max = 5)
      data.table::set(Omega, j = "ratio", value = ratio_temp)

    }

  }

  if (length(hm_vars) > 0) {

    hm_params <- align_parameter_table(start_params$hit_miss, hm_vars)
    Omega_hm <- Omega[, hm_vars, with = FALSE]

    eta_hm <- binary_formula(Omega_hm)
    hm_params$eta <- eta_hm

    hm_denominator <- bernoulli_product(Omega_hm, hm_params$eta)
    data.table::set(
      Omega,
      j = "ratio",
      value = Omega[["ratio"]] * bernoulli_ratio(Omega_hm, hm_params$theta, hm_params$eta)
    )
    data.table::set(Omega, j = "hm_denominator", value = hm_denominator)

    values_list_m <- lapply(hm_vars, function(x) {
      name <- substr(x, 7, nchar(x))
      values <- unique(c(A[[name]], B[[name]]))
      m_est <- sapply(values, function(y) sum(A[[name]] == y) / NROW(A))
      data.table::data.table(
        "value" = values,
        "m_est" = m_est
      )
    })
    names(values_list_m) <- hm_vars
  }

  iter <- 1

  repeat {

    g_est <- pmin(length(M_idx) * Omega$ratio / (length(M_idx) * (Omega$ratio - 1) + n), 1)
    n_M_old <- n_M
    n_M <- sum(g_est)

    if (n_M > min(NROW(A), NROW(B))) {
      n_M <- min(NROW(A), NROW(B))
    }

    data.table::set(Omega, j = "g_est", value = g_est)
    M_idx <- select_mec_indices(
      a = Omega[["a"]],
      b = Omega[["b"]],
      ratio = Omega[["ratio"]],
      n_M = n_M,
      duplicates_in_A = duplicates_in_A
    )
    M <- Omega[M_idx]

    if (length(M_idx) == 0L) {
      break
    }

    selection_mask[M_idx] <- TRUE
    U_idx <- all_idx[!selection_mask]
    selection_mask[M_idx] <- FALSE

    if (iter >= 2) {

      old_params <- c()
      params <- c()
      if (length(b_vars) > 0) {
        old_params <- c(old_params, theta_b_old)
        params <- c(params, theta_b)
      }
      if (length(cpar_vars) > 0) {
        old_params <- c(old_params, p_0_M_old, alpha_M_old, beta_M_old)
        params <- c(params, p_0_M, alpha_M, beta_M)
      }
      if (length(hm_vars) > 0) {
        old_params <- c(old_params, theta_hm_old)
        params <- c(params, theta_hm)
      }

      if (length(cnonpar_vars) == 0) {

        if ((abs(n_M_old - n_M) < delta) || norm(old_params - params, type = "2") < eps) {
          break
        }

      } else {

        if ((abs(n_M_old - n_M) < delta)) {
          break
        }

      }

    }

    data.table::set(Omega, j = "ratio", value = 1)

    if (length(b_vars) > 0) {

      M_b <- Omega_b[M_idx]
      theta_b_old <- b_params$theta
      theta_b <- binary_formula(M_b)
      b_params$theta <- theta_b

      b_numerator <- bernoulli_product(Omega_b, b_params$theta)
      data.table::set(Omega, j = "ratio", value = Omega[["ratio"]] * b_numerator / Omega[["b_denominator"]])
    }

    if (length(cpar_vars) > 0) {
      M_cpar <- Omega_cpar[M_idx]
      p_0_M_old <- cpar_params$p_0_M
      alpha_M_old <- cpar_params$alpha_M
      beta_M_old <- cpar_params$beta_M
      p_0_M <- p_0_formula(M_cpar)
      gamma_plus_M <- gamma_plus_formula(M_cpar)
      alpha_M <- alpha_formula(M_cpar, modified_nleqslv)
      beta_M <- alpha_M / gamma_plus_M
      beta_M[is.nan(beta_M)] <- beta_M_old[is.nan(beta_M)]
      cpar_params$p_0_M <- p_0_M
      cpar_params$alpha_M <- alpha_M
      cpar_params$beta_M <- beta_M
      cpar_numerator <- hurdle_gamma_product(
        Omega_cpar,
        cpar_params$p_0_M,
        cpar_params$alpha_M,
        cpar_params$beta_M
      )
      data.table::set(Omega, j = "ratio", value = Omega[["ratio"]] * cpar_numerator / Omega[["cpar_denominator"]])

    }

    if (length(cnonpar_vars) > 0) {

      M_cnonpar <- Omega_cnonpar[M_idx]
      U_cnonpar <- Omega_cnonpar[U_idx]

      if (nonpar_hurdle) {

        # When the current split is too sparse for KLIEP, keep the last ratio update.
        p_0_M_cnonpar <- p_0_formula(M_cnonpar)
        p_0_U_cnonpar <- p_0_formula(U_cnonpar)
        ratio_kliep_old <- ratio_kliep

        tryCatch({
          ratio_kliep_models <- fit_kliep_hurdle_models(
            df_numerator = M_cnonpar,
            df_denominator = U_cnonpar,
            variables = cnonpar_vars,
            controls_kliep = controls_kliep
          )
          missing_models <- missing_kliep_models(ratio_kliep_models)

          if (length(missing_models) < length(cnonpar_vars)) {
            if (length(missing_models) > 0L) {
              warn_kliep_once(missing_models, "using only the hurdle mass term for those variables in the current iteration.")
            }

            ratio_kliep <- kliep_hurdle_ratio(
              df = Omega_cnonpar,
              variables = cnonpar_vars,
              p_0_numerator = p_0_M_cnonpar,
              p_0_denominator = p_0_U_cnonpar,
              ratio_kliep_list = ratio_kliep_models
            )
          } else {
            ratio_kliep <- ratio_kliep_old
            warn_kliep_once(cnonpar_vars, "could not be fitted in the current iteration; using the previous ratio estimate.")
          }
          data.table::set(Omega, j = "ratio", value = Omega[["ratio"]] * ratio_kliep)
        },
          error = function(e) {
            warn_kliep_once(cnonpar_vars, paste("failed with error:", e$message))
          })

      } else {

        ratio_kliep <- do.call(
          densityratio::kliep,
          c(list(
            df_numerator = M_cnonpar,
            df_denominator = U_cnonpar
          ),
          controls_kliep)
        )

        data.table::set(Omega, j = "ratio", value = Omega[["ratio"]] * stats::predict(ratio_kliep, Omega_cnonpar))

      }

    }

    if (length(hm_vars) > 0) {

      M_hm <- Omega_hm[M_idx]
      theta_hm_old <- hm_params$theta
      theta_hm <- binary_formula(M_hm)
      hm_params$theta <- theta_hm

      p_est <- n_M / max(NROW(A), NROW(B))

      u_est_list <- lapply(hm_vars, function(x) {
        u_est <- runif(NROW(values_list_m[[x]]), 0, 1)
        u_est <- u_est / sum(u_est)
      })
      names(u_est_list) <- hm_vars

      m_bk_prod <- sapply(1:NROW(B), function(x) {
        m_bk_list <- lapply(hm_vars, function(y) {
          var_name <- substr(y, 7, nchar(y))
          d_value <- (B[[var_name]])[x]
          m_kd <- as.vector((values_list_m[[y]])[(values_list_m[[y]])[["value"]] == d_value, ][["m_est"]])
        })
        names(m_bk_list) <- hm_vars
        Reduce(`*`, m_bk_list)
      })

      iter_em <- 1
      while (iter_em <= max_iter_em + 1) {

        if (iter_em >= 2) {
          delta_b_old <- em_params[["delta_b"]]
        }

        delta_m_u <- lapply(1:NROW(B), function(x) {

          u_bk_list <- lapply(hm_vars, function(y) {
            var_name <- substr(y, 7, nchar(y))
            values <- cbind(values_list_m[[y]], u_est_list[[y]])
            colnames(values) <- c("value", "m_est", "u_est")
            d_value <- (B[[var_name]])[x]
            u_kd <- as.vector(values[values[["value"]] == d_value, ][["u_est"]])
          })
          names(u_bk_list) <- hm_vars
          u_bk_prod <- Reduce(`*`, u_bk_list)

          data.table::data.table(delta_b = p_est * m_bk_prod[x] / (p_est * m_bk_prod[x] + (1 - p_est) * u_bk_prod),
                                 m_bk_prod = m_bk_prod[x], u_bk_prod = u_bk_prod)

        })

        em_params <- data.table::rbindlist(delta_m_u)

        if (iter_em >= 3) {
          log_lik_old <- log_lik
        }

        if (iter_em >= 2) {
          log_lik <- sum(
            ifelse(delta_b_old == 0,
                   0,
                   delta_b_old * log(p_est * em_params[["m_bk_prod"]]))
          ) +
            sum(ifelse(em_params[["u_bk_prod"]] == 0,
                       0,
                       (1 - delta_b_old) * log((1 - p_est) * em_params[["u_bk_prod"]]))
                )
        }

        if (iter_em >= 3) {
          if (abs(log_lik - log_lik_old) <= tol_em) break
        }

        u_est_list <- lapply(hm_vars, function(x) {

          var_name <- substr(x, 7, nchar(x))
          z_bk <- B[[var_name]]

          sapply((values_list_m[[x]])[["value"]], function(y) {
            sum((1 - em_params[["delta_b"]]) * (z_bk == y)) / sum(1 - em_params[["delta_b"]])
          })

        })
        names(u_est_list) <- hm_vars

        iter_em <- iter_em + 1

      }

      hm_numerator <- bernoulli_product(Omega_hm, hm_params$theta)
      eta_hm <- sapply(hm_vars, function(col) {
        ((1 - p_est) * sum(u_est_list[[col]] * (values_list_m[[col]])[["m_est"]]) +
           p_est * (1 - 1 / NROW(A)) * sum(((values_list_m[[col]])[["m_est"]]) ^ 2)) /
          (1 - p_est / NROW(A))
      })
      hm_params$eta <- eta_hm

      hm_denominator <- bernoulli_product(Omega_hm, hm_params$eta)
      data.table::set(Omega, j = "hm_denominator", value = hm_denominator)

      data.table::set(Omega, j = "ratio", value = Omega[["ratio"]] * hm_numerator / Omega[["hm_denominator"]])
    }

    iter <- iter + 1

  }

  n_M_est <- n_M
  selection_summary <- summarize_mec_selection(
    a = Omega[["a"]],
    b = Omega[["b"]],
    ratio = Omega[["ratio"]],
    g_est = Omega[["g_est"]],
    n_M_est = n_M_est,
    duplicates_in_A = duplicates_in_A,
    set_construction = set_construction,
    target_rate = target_rate,
    tol = tol,
    max_iter = max_iter_bisection
  )
  M <- Omega[selection_summary$selected_idx]
  flr_est <- selection_summary$flr_est
  mmr_est <- selection_summary$mmr_est
  iter_bisection <- selection_summary$iter
  if (verbose) {
    message(sprintf("Finished MEC with %d predicted match(es).", NROW(M)))
  }

  if (!is.null(true_matches)) {

    data.table::setDT(true_matches)

    eval <- evaluation(M, true_matches, n)
    eval_metrics <- unlist(get_metrics(
      TP = eval$TP,
      FP = eval$FP,
      FN = eval$FN,
      TN = eval$TN
    ))
    confusion <- get_confusion(
      TP = eval$TP,
      FP = eval$FP,
      FN = eval$FN,
      TN = eval$TN
    )

  }

  structure(
    list(
      M_est = M[, c("a", "b", "ratio")],
      n_M_est = n_M_est,
      flr_est = flr_est,
      mmr_est = if (is.null(mmr_est)) NULL else mmr_est,
      iter_bisection = iter_bisection,
      n = n,
      b_vars = if (length(b_vars) == 0) NULL else b_vars,
      cpar_vars = if (length(cpar_vars) == 0) NULL else cpar_vars,
      cnonpar_vars = if (length(cnonpar_vars) == 0) NULL else cnonpar_vars,
      hm_vars = if (length(hm_vars) == 0) NULL else hm_vars,
      b_params = if (length(b_vars) == 0) NULL else b_params,
      cpar_params = if (length(cpar_vars) == 0) NULL else cpar_params,
      hm_params = if (length(hm_vars) == 0) NULL else hm_params,
      ratio_kliep = if (is.null(ratio_kliep)) NULL else ratio_kliep,
      delta = delta,
      eps = eps,
      controls_nleqslv = controls_nleqslv,
      variables = variables,
      set_construction = set_construction,
      eval_metrics = if (is.null(true_matches)) NULL else eval_metrics,
      confusion = if (is.null(true_matches)) NULL else confusion
    ),
    class = "mec_rec_lin"
  )

}


#' @title Blocked Unsupervised Maximum Entropy Classifier for Record Linkage
#'
#' @author Adam Struzik
#'
#' @description
#' Runs graph-based blocking using \link[blocking:blocking]{blocking()},
#' defines a blocking candidate-pair space, and fits an inverted unsupervised
#' maximum entropy classifier (MEC) directly on all candidate pairs.
#'
#' @param A A duplicate-free `data.frame` or `data.table`.
#' @param B A duplicate-free `data.frame` or `data.table`.
#' @param variables A character vector of key variables used to create MEC comparison vectors.
#' @param comparators A named list of functions for comparing pairs of records.
#' @param methods A named list of methods used for estimation (`"binary"` or
#' `"continuous_parametric"`). Other unsupervised MEC methods are not
#' supported by [mec_blocking()] at this stage.
#' @param blocking_x Optional input passed as `x` to \link[blocking:blocking]{blocking()}.
#' @param blocking_y Optional input passed as `y` to \link[blocking:blocking]{blocking()}.
#' @param blocking_variables Variables used to create blocking strings when
#' `blocking_x` and `blocking_y` are not supplied.
#' @param blocking_sep Separator used when concatenating `blocking_variables`.
#' @param controls_blocking A list of additional arguments passed to
#' \link[blocking:blocking]{blocking()}, except `x` and `y`.
#' @param start_params Start parameters for the `"binary"` and
#' `"continuous_parametric"` methods.
#' @param alpha A single numeric value in `[0, 1)` controlling the fraction of
#' the current nonmatch complement dropped from nonmatch-side parameter fitting
#' after the first inverted MEC iteration. The first U-side fit uses the full
#' initial fitting set, and posterior/count formulas continue to use the full
#' current nonmatch count.
#' @param delta A numeric value specifying the tolerance for the change in the
#' estimated number of nonmatches between MEC iterations.
#' @param eps A numeric value specifying the tolerance for the change in model
#' parameters between MEC iterations.
#' @param controls_nleqslv Controls passed to the \link[nleqslv:nleqslv]{nleqslv()} function
#' (only if the `"continuous_parametric"` method has been chosen for at least one
#' variable).
#' @param true_matches A `data.frame` or `data.table` indicating known matches.
#' @param keep_blocking_result Logical indicating whether to store the raw object
#' returned by \link[blocking:blocking]{blocking()}.
#' @param keep_training_data Logical indicating whether to store pooled training
#' comparison vectors.
#' @param verbose Logical indicating whether to print progress messages.
#'
#' @details
#' The function assumes one-to-one linkage. The blocking stage defines disjoint
#' bipartite blocks, and the candidate-pair space \eqn{\Omega_B} is the union of
#' within-block Cartesian products. Duplicate candidate pairs are removed
#' deterministically before MEC fitting.
#'
#' The blocked MEC fit is inverted relative to [mec()]. The initial match set
#' contains at most \eqn{\nu} feasible pairs, where \eqn{\nu} is the structural
#' one-to-one upper bound. Initial feasible matches are selected
#' greedily by an unweighted disagreement norm: binary agreement indicators use
#' `1 - gamma`, while continuous dissimilarities use `gamma` unchanged. At each
#' iteration, match-side parameters are estimated from the current greedy
#' one-to-one match set, and nonmatch-side parameters are estimated from its
#' complement.
#'
#' The `alpha` argument applies only to nonmatch-side distribution estimation.
#' The first U-side fit uses the full initial complement. In later iterations,
#' the least reliable current nonmatches are dropped from the U-side fitting
#' sample, with reliability ranked by the previous nonmatch posterior estimate
#' and then by the inverted density ratio if the posterior is unavailable.
#' The posterior and count updates still use the full current complement size,
#' and the final match set remains one-to-one.
#'
#' The returned `ratio` is \eqn{s = u / m}, where \eqn{u} and \eqn{m} denote
#' the estimated nonmatch and match comparison-vector densities. Smaller values
#' are therefore more match-like. Updated match sets are selected greedily in
#' ascending order of this ratio.
#'
#' If \eqn{N = |\Omega_B|} and \eqn{\nu} is the maximum feasible one-to-one
#' matching size in the candidate graph, the estimated number of nonmatches is
#' bounded below by \eqn{N - \nu}. For the disjoint complete blocks reconstructed
#' by this function, \eqn{\nu = \sum_h \min(n_{Ah}, n_{Bh})}.
#'
#' If the initialized match set exhausts the candidate-pair space, for example
#' when \eqn{N = \nu}, there is no candidate complement from which to estimate
#' nonmatch parameters. In that case the function returns the structurally
#' feasible initialized match set, sets `n_U_est = 0`, and leaves nonmatch-side
#' parameters unavailable.
#'
#' @return
#' Returns a list of class `"mec_blocking"` containing:
#' \itemize{
#' \item{`M_est` -- a `data.table` with predicted matches and columns `a`, `b`, `block`, and `ratio`,}
#' \item{`n_M_est` -- estimated total number of matches across all blocks,}
#' \item{`n_U_est` -- estimated total number of candidate nonmatches,}
#' \item{`alpha` -- fraction of the current nonmatch complement dropped from later U-side fitting,}
#' \item{`candidate_pair_count` -- number of candidate pairs in \eqn{\Omega_B},}
#' \item{`block_estimates` -- a `data.table` with block-level size and match-count diagnostics,}
#' \item{`block_summary` -- a `data.table` describing the final disjoint blocks,}
#' \item{`excluded_records` -- a list with records from `A` and `B` excluded by blocking,}
#' \item{`b_vars` -- variables used for the `"binary"` method, with the prefix `"gamma_"`,}
#' \item{`cpar_vars` -- variables used for the `"continuous_parametric"` method, with the prefix `"gamma_"`,}
#' \item{`b_params` -- parameters estimated using the `"binary"` method,}
#' \item{`cpar_params` -- parameters estimated using the `"continuous_parametric"` method,}
#' \item{`variables` -- key variables used for comparison,}
#' \item{`comparators` -- comparison functions used to create comparison vectors,}
#' \item{`methods` -- MEC estimation methods used for the key variables,}
#' \item{`delta` -- tolerance for changes in the estimated number of nonmatches,}
#' \item{`eps` -- tolerance for changes in nonmatch-side model parameters,}
#' \item{`controls_nleqslv` -- controls passed to \link[nleqslv:nleqslv]{nleqslv()},}
#' \item{`blocking_result` -- raw object returned by \link[blocking:blocking]{blocking()} if `keep_blocking_result = TRUE`; otherwise `NULL`,}
#' \item{`training_Omega` -- candidate-space comparison vectors with inverted scores if `keep_training_data = TRUE`; otherwise `NULL`,}
#' \item{`blocking_eval` -- blocking diagnostics if `true_matches` is provided; otherwise `NULL`,}
#' \item{`mec_eval` -- MEC-selection diagnostics among known matches retained in the candidate-pair space if `true_matches` is provided; otherwise `NULL`,}
#' \item{`eval_metrics` -- empirical linkage quality metrics based on `true_matches`; otherwise `NULL`,}
#' \item{`confusion` -- empirical confusion matrix based on `true_matches`; otherwise `NULL`.}
#' }
#'
#' @examples
#' df_1 <- data.frame(
#'   name = c("Emma", "Liam", "Olivia", "Noah", "Ava"),
#'   surname = c("Smith", "Jones", "Brown", "Davis", "Miller"),
#'   city = c("Boston", "Boston", "Austin", "Austin", "Denver")
#' )
#' df_2 <- data.frame(
#'   name = c("Emma", "Liam", "Olivia", "Noah", "Ava"),
#'   surname = c("Smith", "Jones", "Brown", "Davis", "Miller"),
#'   city = c("Boston", "Boston", "Austin", "Austin", "Denver")
#' )
#'
#' blocking_x <- matrix(
#'   c(1, 0, 0, 1, 1, 1, 2, 0, 0, 2),
#'   ncol = 2,
#'   byrow = TRUE
#' )
#' blocking_y <- blocking_x
#'
#' result <- mec_blocking(
#'   A = df_1,
#'   B = df_2,
#'   variables = c("name", "surname", "city"),
#'   blocking_x = blocking_x,
#'   blocking_y = blocking_y,
#'   controls_blocking = list(
#'     representation = "custom_matrix",
#'     ann = "kd",
#'     distance = "euclidean",
#'     seed = 1
#'   ),
#'   true_matches = data.frame(a = 1:5, b = 1:5)
#' )
#' result
#' @import data.table
#' @importFrom blocking blocking
#' @export
mec_blocking <- function(
    A,
    B,
    variables,
    comparators = NULL,
    methods = NULL,
    blocking_x = NULL,
    blocking_y = NULL,
    blocking_variables = variables,
    blocking_sep = " ",
    controls_blocking = list(),
    start_params = NULL,
    alpha = 0,
    delta = 0.5,
    eps = 0.05,
    controls_nleqslv = list(),
    true_matches = NULL,
    keep_blocking_result = FALSE,
    keep_training_data = FALSE,
    verbose = FALSE) {

  stopifnot("`A` should be a data.frame or a data.table." =
              is.data.frame(A) | is.data.table(A))
  stopifnot("`B` should be a data.frame or a data.table." =
              is.data.frame(B) | is.data.table(B))
  stopifnot("`variables` should be a character vector." =
              is.character(variables))
  stopifnot("`variables` should contain at least one variable." =
              length(variables) > 0L)

  methods <- validate_methods(
    methods = methods,
    variables = variables,
    allowed_methods = c("binary", "continuous_parametric")
  )
  validate_blocking_controls(controls_blocking)
  validate_mec_blocking_alpha(alpha)

  A <- data.table::copy(data.table::as.data.table(A))
  B <- data.table::copy(data.table::as.data.table(B))
  n_A_original <- nrow(A)
  n_B_original <- nrow(B)

  true_matches <- sanitize_true_matches(
    true_matches = true_matches,
    n_A = n_A_original,
    n_B = n_B_original,
    arg_name = "true_matches"
  )

  blocking_inputs <- prepare_blocking_inputs(
    A = A,
    B = B,
    blocking_x = blocking_x,
    blocking_y = blocking_y,
    blocking_variables = blocking_variables,
    blocking_sep = blocking_sep
  )

  if (verbose) {
    message("Running blocking::blocking().")
  }

  blocking_result <- run_blocking(
    blocking_x = blocking_inputs$x,
    blocking_y = blocking_inputs$y,
    controls_blocking = controls_blocking
  )
  blocking_table <- extract_blocking_result_table(blocking_result)
  reconstruction <- reconstruct_block_summary(
    blocking_table = blocking_table,
    n_A = n_A_original,
    n_B = n_B_original
  )
  block_summary <- reconstruction$block_summary
  excluded_records <- reconstruction$excluded_records

  if (verbose) {
    message(sprintf("Reconstructed %d disjoint blocks.", NROW(block_summary)))
  }

  data.table::set(A, j = "a", value = seq_len(n_A_original))
  data.table::set(B, j = "b", value = seq_len(n_B_original))

  preprocessing <- drop_constant_key_variables(
    A = A,
    B = B,
    variables = variables,
    comparators = comparators,
    methods = methods
  )
  A <- preprocessing$A
  B <- preprocessing$B
  variables <- preprocessing$variables
  comparators <- preprocessing$comparators
  methods <- preprocessing$methods
  constant_vars <- preprocessing$constant_vars

  for (var in constant_vars) {
    warning(paste("The variable", var, "has only one unique value and has been removed."))
  }

  candidate_pairs <- make_block_pair_table(block_summary)
  if (NROW(candidate_pairs) == 0L) {
    stop("Blocking produced no candidate pairs for mec_blocking().", call. = FALSE)
  }
  candidate_pairs <- candidate_pairs[, c("a", "b", "block"), with = FALSE]
  data.table::setorder(candidate_pairs, block, a, b)
  candidate_pairs <- unique(candidate_pairs, by = c("a", "b"))
  data.table::setorder(candidate_pairs, block, a, b)

  candidate_pair_count <- NROW(candidate_pairs)
  nu <- sum(pmin(block_summary[["n_A"]], block_summary[["n_B"]]))
  n_U_min <- candidate_pair_count - nu

  if (verbose) {
    message(sprintf(
      "Fitting inverted MEC on %d candidate pair(s) from %d block(s).",
      candidate_pair_count,
      NROW(block_summary)
    ))
  }

  candidate_vectors <- comparison_vectors(
    A = A,
    B = B,
    variables = variables,
    comparators = comparators,
    pairs = candidate_pairs
  )
  candidate_Omega <- candidate_vectors$Omega
  comparators <- candidate_vectors$comparators

  pooled_fit <- fit_mec_blocking_inverted_omega(
    A = A,
    B = B,
    variables = variables,
    comparators = comparators,
    methods = methods,
    Omega = candidate_Omega,
    start_params = start_params,
    delta = delta,
    eps = eps,
    controls_nleqslv = controls_nleqslv,
    n_U_min = n_U_min,
    nu = nu,
    alpha = alpha,
    context = "mec_blocking()"
  )
  pooled_model <- pooled_fit$model
  M_est <- pooled_fit$M_est
  n_M_est <- pooled_fit$n_M_est
  n_U_est <- pooled_fit$n_U_est

  block_estimates <- block_summary[, .(
    block,
    n_A,
    n_B,
    pair_count,
    nonmatches_min
  )]
  data.table::set(block_estimates, j = "n_M_est", value = 0L)
  data.table::set(block_estimates, j = "selected_pairs", value = 0L)
  if (NROW(M_est) > 0L) {
    selected_by_block <- M_est[, .(selected_pairs = .N), by = block]
    selected_counts <- selected_by_block[["selected_pairs"]][
      match(block_estimates[["block"]], selected_by_block[["block"]])
    ]
    selected_counts[is.na(selected_counts)] <- 0L
    data.table::set(block_estimates, j = "n_M_est", value = as.integer(selected_counts))
    data.table::set(block_estimates, j = "selected_pairs", value = as.integer(selected_counts))
  }

  block_pairs <- candidate_pairs
  blocking_eval <- NULL
  mec_eval <- NULL
  eval_metrics <- NULL
  confusion <- NULL

  if (!is.null(true_matches)) {
    blocking_eval <- blocking_diagnostics(
      true_matches = true_matches,
      block_pairs = block_pairs,
      n_full_pairs = n_A_original * n_B_original
    )
    mec_eval <- mec_selection_diagnostics(
      true_matches = true_matches,
      block_pairs = block_pairs,
      M_est = M_est
    )
    eval <- evaluation(M_est, true_matches, n_A_original * n_B_original)
    eval_metrics <- unlist(get_metrics(
      TP = eval$TP,
      FP = eval$FP,
      FN = eval$FN,
      TN = eval$TN
    ))
    confusion <- get_confusion(
      TP = eval$TP,
      FP = eval$FP,
      FN = eval$FN,
      TN = eval$TN
    )
  }

  structure(
    list(
      M_est = M_est,
      n_M_est = n_M_est,
      n_U_est = n_U_est,
      alpha = pooled_fit$alpha,
      candidate_pair_count = pooled_fit$candidate_pair_count,
      block_estimates = block_estimates,
      block_summary = block_summary,
      excluded_records = excluded_records,
      b_vars = pooled_model$b_vars,
      cpar_vars = pooled_model$cpar_vars,
      b_params = pooled_model$b_params,
      cpar_params = pooled_model$cpar_params,
      variables = variables,
      comparators = comparators,
      methods = methods,
      delta = delta,
      eps = eps,
      controls_nleqslv = controls_nleqslv,
      blocking_result = if (keep_blocking_result) blocking_result else NULL,
      training_Omega = if (keep_training_data) pooled_fit$Omega else NULL,
      blocking_eval = blocking_eval,
      mec_eval = mec_eval,
      eval_metrics = eval_metrics,
      confusion = confusion
    ),
    class = "mec_blocking"
  )
}

library(automatedRecLin)
library(data.table)

data("A_example")
data("B_example")

variables <- c("name", "surname")
comparators <- list("name" = jarowinkler_complement(),
                    "surname" = jarowinkler_complement())
methods_cpar <- list("name" = "continuous_parametric",
                     "surname" = "continuous_parametric")
methods_cnonpar <- list("name" = "continuous_nonparametric",
                        "surname" = "continuous_nonparametric")
methods_hm <- list("name" = "hit_miss",
                   "surname" = "hit_miss")
true_matches <- data.table("a" = 1:8, "b" = 1:8)
confusion_b <- matrix(c(4, 0, 4, 112), nrow = 2, ncol = 2)
rownames(confusion_b) <- c("Actual Positive", "Actual Negative")
colnames(confusion_b) <- c("Predicted Positive", "Predicted Negative")

expect_rate_bounds <- function(result) {
  expect_true(is.finite(result$flr_est))
  expect_true(result$flr_est >= 0 && result$flr_est <= 1)
  expect_true(is.finite(result$mmr_est))
  expect_true(result$mmr_est >= 0 && result$mmr_est <= 1)
}

set.seed(1)

expect_silent(
  mec(A = A_example, B = B_example, variables = variables)
)

expect_equal(
  mec(A = A_example, B = B_example, variables = variables)$M_est,
  data.table("a" = 1:4, "b" = 1:4, "ratio" = rep(720, 4))
)

expect_equal(
  mec(A = A_example, B = B_example, variables = variables)$b_params,
  data.table("variable" = c("gamma_name", "gamma_surname"),
             "theta" = c(1, 1), "eta" = c(0.04166666666666667823149, 0.03333333333333331205406))
)

expect_equal(
  mec(A = A_example, B = B_example, variables = variables, true_matches = true_matches)$eval_metrics,
  c("FLR" = 0, "MMR" = 0.5)
)

expect_equal(
  mec(A = A_example, B = B_example, variables = variables, true_matches = true_matches)$confusion,
  confusion_b
)

expect_equal(
  mec(A = A_example, B = B_example, variables = variables, methods = methods_hm)$M_est,
  data.table("a" = 1:4, "b" = 1:4, "ratio" = rep(573.0984617692614619955, 4))
)

expect_equal(
  mec(A = A_example, B = B_example, variables = variables, methods = methods_hm)$hm_params,
  data.table("variable" = c("gamma_name", "gamma_surname"),
             "theta" = c(1, 1), "eta" = c(0.04616298284003408219922, 0.03847198038861038282832))
)

expect_equal(
  mec(A = A_example, B = B_example, variables = variables,
      comparators = comparators, methods = methods_cpar)$M_est[, c("a", "b")],
  data.table("a" = c(7, 6, 8, 5, 1, 2, 3, 4), "b" = c(7, 6, 8, 5, 1, 2, 3, 4))
)

expect_equal(
  mec(A = A_example, B = B_example, variables = variables,
      comparators = comparators, methods = methods_cpar)$cpar_params,
  data.table("variable" = c("gamma_name", "gamma_surname"),
             "p_0_M" = c(0.625, 0.5),
             "alpha_M" = c(138.4622794465536514963, 120.6657058981201657843),
             "beta_M" = c(2199.106791209967013856, 1974.529732878328104562),
             "p_0_U" = c(0.04166666666666666435370, 0.03333333333333333287074),
             "alpha_U" = c(6.516735740295027667912, 4.622775398311523176176),
             "beta_U" = c(11.173089162681424824086, 7.167260899080440061937))
)

cnonpar_fit <- NULL
expect_warning(
  cnonpar_fit <- mec(A = A_example, B = B_example, variables = variables,
                     comparators = comparators, methods = methods_cnonpar)
)

expect_equal(
  cnonpar_fit$M_est,
  data.table("a" = 1:4, "b" = 1:4,
             "ratio" = rep(369.3706815704043719961, 4))
)

expect_equal(
  cnonpar_fit$n_M_est,
  3.835110959839932665005
)

cpar_fit <- mec(A = A_example, B = B_example, variables = variables,
                comparators = comparators, methods = methods_cpar)

expect_equal(
  cpar_fit$mmr_est,
  cpar_fit$flr_est
)

cpar_fit_flr <- mec(A = A_example, B = B_example, variables = variables,
                    comparators = comparators, methods = methods_cpar,
                    set_construction = "flr", target_rate = 0.05)

expect_rate_bounds(cpar_fit_flr)

cpar_fit_mmr <- mec(A = A_example, B = B_example, variables = variables,
                    comparators = comparators, methods = methods_cpar,
                    set_construction = "mmr", target_rate = 0.05)

expect_rate_bounds(cpar_fit_mmr)

library(automatedRecLin)
library(data.table)

options("text2vec.mc.cores" = 1L)

controls_blocking <- list(
  representation = "custom_matrix",
  ann = "kd",
  distance = "euclidean",
  seed = 1
)

expect_blocking_output_contract <- function(result) {
  removed_outputs <- c(
    "n_U_min", "nu", "n_M_init", "n_U_init", "n_U_fit",
    "u_fit_diagnostics", "ratio_orientation", "pooled_model"
  )
  expect_false(any(removed_outputs %in% names(result)))
  expect_false(any(c("flr_est", "mmr_est") %in% names(result)))
  expect_false(any(c("training_rule", "training_blocks", "controls_blocking") %in% names(result)))
  expect_equal(names(result$M_est), c("a", "b", "block", "ratio"))
  expect_equal(result$n_M_est, NROW(result$M_est))
  expect_equal(result$n_U_est, result$candidate_pair_count - result$n_M_est)
  expect_false(any(c("nonmatch_sample_size", "nonmatch_sampling_seed", "prob_ratio") %in% names(result)))
  expect_equal(
    names(result$block_estimates),
    c("block", "n_A", "n_B", "pair_count", "nonmatches_min", "n_M_est", "selected_pairs")
  )
  expect_true(all(result$block_estimates[["n_M_est"]] >= 0))
  expect_true(all(result$block_estimates[["selected_pairs"]] >= 0))
  expect_true(all(result$block_estimates[["n_M_est"]] <=
                    pmin(result$block_estimates[["n_A"]], result$block_estimates[["n_B"]])))
  expect_true(all(result$block_estimates[["selected_pairs"]] <=
                    pmin(result$block_estimates[["n_A"]], result$block_estimates[["n_B"]])))
}

expect_equal(
  automatedRecLin:::select_inverted_mec_indices(
    a = c(1L, 1L, 2L, 2L),
    b = c(1L, 2L, 1L, 2L),
    block = c(1, 1, 1, 1),
    ratio = c(2, 0.1, 0.2, Inf),
    n_M = 2
  ),
  c(2L, 3L)
)
expect_equal(
  automatedRecLin:::select_inverted_mec_indices(
    a = c(1L, 2L),
    b = c(1L, 2L),
    block = c(1, 1),
    ratio = c(NA_real_, 0.5),
    n_M = 1
  ),
  2L
)
expect_equal(
  automatedRecLin:::select_inverted_mec_indices(
    a = c(1L, 2L),
    b = c(1L, 2L),
    block = c(1, 1),
    ratio = c(-1, 0.3),
    n_M = 1
  ),
  2L
)
expect_equal(
  automatedRecLin:::select_inverted_mec_indices(
    a = c(1L, 2L),
    b = c(1L, 2L),
    block = c(1, 1),
    ratio = c(0.1, 0.2),
    n_M = 0
  ),
  integer()
)

unkey <- function(x) {
  data.table::setkey(x, NULL)
  x
}

make_doc_example_data <- function() {
  list(
    A = data.frame(
      name = c("Emma", "Liam", "Olivia", "Noah", "Ava"),
      surname = c("Smith", "Jones", "Brown", "Davis", "Miller"),
      city = c("Boston", "Boston", "Austin", "Austin", "Denver")
    ),
    B = data.frame(
      name = c("Emma", "Liam", "Olivia", "Noah", "Ava"),
      surname = c("Smith", "Jones", "Brown", "Davis", "Miller"),
      city = c("Boston", "Boston", "Austin", "Austin", "Denver")
    ),
    blocking_x = matrix(
      c(1, 0, 0, 1, 1, 1, 2, 0, 0, 2),
      ncol = 2,
      byrow = TRUE
    ),
    true_matches = data.frame(a = 1:5, b = 1:5)
  )
}

confusion_all_links <- matrix(c(5, 0, 0, 20), nrow = 2, ncol = 2)
rownames(confusion_all_links) <- c("Actual Positive", "Actual Negative")
colnames(confusion_all_links) <- c("Predicted Positive", "Predicted Negative")

doc_data <- make_doc_example_data()

set.seed(1)
fit_binary <- mec_blocking(
  A = doc_data$A,
  B = doc_data$B,
  variables = c("name", "surname", "city"),
  blocking_x = doc_data$blocking_x,
  blocking_y = doc_data$blocking_x,
  controls_blocking = controls_blocking,
  true_matches = doc_data$true_matches
)

expect_inherits(fit_binary, "mec_blocking")
expect_equal(fit_binary$n_M_est, 5L)
expect_equal(fit_binary$n_U_est, 0L)
expect_equal(fit_binary$candidate_pair_count, 5L)
expect_equal(
  fit_binary$M_est[, .(a, b, block)],
  data.table(a = 1:5, b = 1:5, block = as.numeric(1:5))
)
expect_true(all(is.finite(fit_binary$M_est[["ratio"]])))
expect_true(all(fit_binary$M_est[["ratio"]] >= 0))
expect_equal(
  unkey(fit_binary$block_summary[, .(block, n_A, n_B, pair_count, nonmatches_min)]),
  data.table(
    block = as.numeric(1:5),
    n_A = rep(1L, 5),
    n_B = rep(1L, 5),
    pair_count = rep(1L, 5),
    nonmatches_min = rep(0L, 5)
  )
)
expect_equal(unlist(fit_binary$block_summary[["A"]]), 1:5)
expect_equal(unlist(fit_binary$block_summary[["B"]]), 1:5)
expect_equal(fit_binary$excluded_records$A, integer())
expect_equal(fit_binary$excluded_records$B, integer())
expect_equal(
  fit_binary$blocking_eval,
  data.table(
    true_matches = 5L,
    preserved_matches = 5L,
    lost_matches = 0L,
    blocking_recall = 1,
    blocking_fnr = 0,
    blocked_pairs = 5L,
    full_pairs = 25L
  )
)
expect_equal(fit_binary$eval_metrics, c(FLR = 0, MMR = 0))
expect_equal(fit_binary$confusion, confusion_all_links)
expect_blocking_output_contract(fit_binary)
expect_equal(fit_binary$block_estimates[["n_M_est"]], rep(1L, 5L))
expect_equal(fit_binary$block_estimates[["selected_pairs"]], rep(1L, 5L))
expect_null(fit_binary$blocking_result)
expect_null(fit_binary$training_Omega)

fit_binary_print <- paste(capture.output(print(fit_binary)), collapse = "\n")
expect_false(grepl("Estimated false link rate", fit_binary_print))
expect_false(grepl("Estimated missing match rate", fit_binary_print))
expect_true(grepl("Evaluation metrics", fit_binary_print))

set.seed(1)
fit_cpar <- mec_blocking(
  A = doc_data$A,
  B = doc_data$B,
  variables = c("name", "surname", "city"),
  comparators = list(
    name = jarowinkler_complement(),
    surname = jarowinkler_complement()
  ),
  methods = list(
    name = "continuous_parametric",
    surname = "continuous_parametric"
  ),
  blocking_x = doc_data$blocking_x,
  blocking_y = doc_data$blocking_x,
  controls_blocking = controls_blocking,
  true_matches = doc_data$true_matches
)

expect_equal(fit_cpar$cpar_vars, c("gamma_name", "gamma_surname"))
expect_equal(fit_cpar$b_vars, "gamma_city")
expect_null(fit_cpar$hm_vars)
expect_null(fit_cpar$cnonpar_vars)
expect_null(fit_cpar$ratio_kliep)
expect_null(fit_cpar$ratio_kliep_list)
expect_equal(
  fit_cpar$M_est[, .(a, b, block)],
  data.table(a = 1:5, b = 1:5, block = as.numeric(1:5))
)
expect_equal(fit_cpar$cpar_params[["variable"]], c("gamma_name", "gamma_surname"))
expect_true(all(is.finite(unlist(fit_cpar$cpar_params[, -"variable"]))))
expect_equal(fit_cpar$b_params[["variable"]], "gamma_city")
expect_true(all(is.finite(fit_cpar$b_params[["theta"]])))
expect_true(all(is.finite(fit_cpar$b_params[["eta"]])))
expect_blocking_output_contract(fit_cpar)

singleton_nonmatch_A <- data.frame(
  name = c("same", "left"),
  surname = c("person", "alpha")
)
singleton_nonmatch_B <- data.frame(
  name = c("same", "right"),
  surname = c("person", "beta")
)
singleton_nonmatch_blocking <- matrix(c(1, 2), ncol = 1)

set.seed(1)
fit_singleton_nonmatch <- mec_blocking(
  A = singleton_nonmatch_A,
  B = singleton_nonmatch_B,
  variables = c("name", "surname"),
  blocking_x = singleton_nonmatch_blocking,
  blocking_y = singleton_nonmatch_blocking,
  controls_blocking = controls_blocking
)

expect_blocking_output_contract(fit_singleton_nonmatch)
expect_equal(fit_singleton_nonmatch$n_M_est, 2L)
expect_equal(fit_singleton_nonmatch$n_U_est, 0L)
expect_equal(fit_singleton_nonmatch$M_est[, .(a, b)], data.table(a = 1:2, b = 1:2))
expect_equal(
  unkey(fit_singleton_nonmatch$block_estimates[, .(block, n_M_est, selected_pairs)]),
  data.table(block = as.numeric(1:2), n_M_est = c(1L, 1L), selected_pairs = c(1L, 1L))
)

threshold_A <- data.frame(
  name = c("A1", "A2", "A3", "B1", "B2", "B3"),
  surname = c("S1", "S2", "S3", "T1", "T2", "T3")
)
threshold_B <- threshold_A
threshold_blocking <- matrix(c(1, 1, 1, 2, 2, 2), ncol = 1)

confusion_threshold <- matrix(c(2, 0, 4, 30), nrow = 2, ncol = 2)
rownames(confusion_threshold) <- c("Actual Positive", "Actual Negative")
colnames(confusion_threshold) <- c("Predicted Positive", "Predicted Negative")

set.seed(1)
fit_threshold <- mec_blocking(
  A = threshold_A,
  B = threshold_B,
  variables = c("name", "surname"),
  blocking_x = threshold_blocking,
  blocking_y = threshold_blocking,
  controls_blocking = controls_blocking,
  true_matches = data.frame(a = 1:6, b = 1:6),
  keep_training_data = TRUE,
  keep_blocking_result = TRUE
)

expect_equal(fit_threshold$candidate_pair_count, 6L)
expect_equal(fit_threshold$n_U_est, 4L)
expect_equal(fit_threshold$n_M_est, 2L)
expect_equal(fit_threshold$block_summary[["block"]], as.numeric(1:2))
expect_equal(NROW(fit_threshold$training_Omega), fit_threshold$candidate_pair_count)
expect_true(all(c("gamma_name", "gamma_surname", "init_disagreement", "ratio", "q_est") %in%
                  names(fit_threshold$training_Omega)))
expect_true(all(fit_threshold$training_Omega[["q_est"]] >= 0 &
                  fit_threshold$training_Omega[["q_est"]] <= 1))
expect_true(!is.null(fit_threshold$blocking_result))
expect_true("result" %in% names(fit_threshold$blocking_result))

included_A <- as.integer(vapply(
  fit_threshold$block_summary[["A"]],
  function(x) x[1L],
  numeric(1)
))

expect_equal(NROW(fit_threshold$block_summary), 2L)
expect_equal(fit_threshold$block_summary[["n_A"]], rep(1L, 2L))
expect_equal(fit_threshold$block_summary[["n_B"]], rep(3L, 2L))
expect_equal(
  fit_threshold$M_est[, .(a, b, block)],
  data.table(
    a = included_A,
    b = included_A,
    block = fit_threshold$block_summary[["block"]]
  )
)
expect_equal(sort(fit_threshold$excluded_records$A), setdiff(seq_len(6L), included_A))
expect_equal(fit_threshold$excluded_records$B, integer())
expect_equal(
  fit_threshold$blocking_eval,
  data.table(
    true_matches = 6L,
    preserved_matches = 2L,
    lost_matches = 4L,
    blocking_recall = 1 / 3,
    blocking_fnr = 2 / 3,
    blocked_pairs = 6L,
    full_pairs = 36L
  )
)
expect_equal(fit_threshold$eval_metrics, c(FLR = 0, MMR = 2 / 3))
expect_equal(fit_threshold$confusion, confusion_threshold)
expect_blocking_output_contract(fit_threshold)

expect_error(
  mec_blocking(
    A = doc_data$A,
    B = doc_data$B,
    variables = c("name", "surname", "city"),
    methods = list(name = "hit_miss"),
    blocking_x = doc_data$blocking_x,
    blocking_y = doc_data$blocking_x,
    controls_blocking = controls_blocking
  )
)

expect_error(
  mec_blocking(
    A = doc_data$A,
    B = doc_data$B,
    variables = c("name", "surname", "city"),
    blocking_x = doc_data$blocking_x,
    controls_blocking = controls_blocking
  )
)

expect_error(
  mec_blocking(
    A = doc_data$A,
    B = doc_data$B,
    variables = c("name", "surname", "city"),
    blocking_x = doc_data$blocking_x,
    blocking_y = doc_data$blocking_x,
    controls_blocking = c(list(x = doc_data$blocking_x), controls_blocking)
  )
)

expect_error(
  mec_blocking(
    A = doc_data$A,
    B = doc_data$B,
    variables = c("name", "surname", "city"),
    blocking_x = doc_data$blocking_x,
    blocking_y = doc_data$blocking_x,
    controls_blocking = controls_blocking,
    nonmatch_sample_size = 26
  )
)

expect_error(
  mec_blocking(
    A = doc_data$A,
    B = doc_data$B,
    variables = c("name", "surname", "city"),
    blocking_x = doc_data$blocking_x,
    blocking_y = doc_data$blocking_x,
    controls_blocking = controls_blocking,
    min_training_pairs = 1
  )
)

#' @title Evaluation for Record Linkage
#'
#' @noRd
evaluation <- function(pred_matches, true_matches, n) {

  pred_matches <- paste0(pred_matches[["a"]], "_", pred_matches[["b"]])
  true_matches <- paste0(true_matches[["a"]], "_", true_matches[["b"]])

  TP <- sum(pred_matches %in% true_matches)
  FP <- length(pred_matches) - TP
  TN <- n - length(true_matches) - FP
  FN <- sum(!(true_matches %in% pred_matches))

  list(
    TP = TP,
    FP = FP,
    FN = FN,
    TN = TN
  )
}

#' @title Evaluation Metrics for Record Linkage
#'
#' @noRd
get_metrics <- function(TP, FP, FN, TN) {

  precision <- if (TP + FP != 0) TP / (TP + FP) else 0
  fnr <- if (FN + TP != 0) FN / (FN + TP) else 0

  flr <- 1 - precision
  mmr <- fnr

  list(FLR = flr,
       MMR = mmr)

}

#' @title Confusion Matrix for Record Linkage
#'
#' @noRd
get_confusion <- function(TP, FP, FN, TN) {

  cm <- matrix(c(TP, FP, FN, TN), nrow = 2)
  colnames(cm) <- c("Predicted Positive",
                    "Predicted Negative")
  rownames(cm) <- c("Actual Positive",
                    "Actual Negative")
  return(cm)
}

# Avoid CRAN notes for data.table column names used through non-standard evaluation.
utils::globalVariables(c(
  ".",
  "a",
  "b",
  "block",
  "cumulative_nonmatches_min",
  "cumulative_pairs",
  "n_A",
  "n_B",
  "nonmatches_min",
  "pair_count",
  "selected_pairs"
))

#' @noRd
binary_formula <- function(df) {
  if (NCOL(df) == 0L) {
    return(numeric())
  }

  colMeans(as.matrix(df))
}

#' @noRd
p_0_formula <- function(df) {
  if (NCOL(df) == 0L) {
    return(numeric())
  }

  df_matrix <- as.matrix(df)
  colSums(df_matrix == 0) / NROW(df_matrix)
}

#' @noRd
f_alpha <- function(alpha, gamma) {
  gamma <- gamma[gamma > 0]
  gamma_mean <- mean(gamma)
  sum(log(gamma) - log(gamma_mean) - digamma(alpha) + log(alpha))
}

#' @noRd
f_alpha_iterative <- function(alpha, beta, gamma) {
  gamma <- gamma[gamma > 0]
  sum(log(gamma) + log(beta) - digamma(alpha))
}

#' @noRd
gamma_plus_formula <- function(df) {
  if (NCOL(df) == 0L) {
    return(numeric())
  }

  vapply(df, function(x) {
    x <- x[x > 0]
    mean(x)
  }, numeric(1))
}

#' @noRd
alpha_formula <- function(df, fun) {
  if (NCOL(df) == 0L) {
    return(numeric())
  }

  vapply(df, function(y) {
    fun(x = 1, fn = f_alpha, gamma = y, method = "Newton")$x
  }, numeric(1))
}

#' @noRd
alpha_formula_summary <- function(n_positive, positive_sum, log_positive_sum, fun) {
  vapply(seq_along(n_positive), function(i) {
    if (n_positive[i] < 2L) {
      stop("The nonmatch sample should contain at least two positive continuous comparisons for each continuous parametric variable.")
    }

    positive_mean <- positive_sum[i] / n_positive[i]
    f_alpha_summary <- function(alpha) {
      log_positive_sum[i] - n_positive[i] * log(positive_mean) -
        n_positive[i] * digamma(alpha) + n_positive[i] * log(alpha)
    }

    fun(x = 1, fn = f_alpha_summary, method = "Newton")$x
  }, numeric(1))
}

#' @noRd
alpha_formula_iterative <- function(df, fun, beta) {
  K <- length(beta)
  par <- sapply(1:K, function(x) {
    fun(x = 1, fn = f_alpha_iterative, gamma = df[, as.numeric(x)], beta = beta[x], method = "Newton")$x
  })
  par
}

#' @importFrom stats dgamma
#'
#' @noRd
hurdle_gamma_density <- function(x, p_0, alpha, beta) {
  ifelse(x == 0, p_0, 1) *
    ifelse(x > 0, (1 - p_0) * stats::dgamma(x = x, shape = alpha, scale = 1 / beta), 1)
}

#' @noRd
# Check for any exact agreement using only key variables to reduce copying.
has_perfect_agreement <- function(A, B, variables) {
  A_keys <- unique(A[, variables, with = FALSE])
  B_keys <- unique(B[, variables, with = FALSE])

  NROW(A_keys[B_keys, on = variables, nomatch = 0L, mult = "first"]) > 0
}

#' @noRd
positive_gamma_table <- function(gamma, variable) {
  gamma_dt <- data.table::data.table(value = gamma[gamma > 0])
  data.table::setnames(gamma_dt, "value", variable)
  gamma_dt
}

#' @noRd
validate_match_pairs <- function(matches, n_A, n_B, arg_name = "matches") {
  if (!(is.data.frame(matches) || is.data.table(matches))) {
    stop(sprintf("`%s` should be a data.frame or a data.table.", arg_name))
  }

  data.table::setDT(matches)

  if (!(length(colnames(matches)) == 2L && all(colnames(matches) == c("a", "b")))) {
    stop(sprintf("`%s` should consist of two columns: a, b.", arg_name))
  }

  if (anyNA(matches[["a"]]) || anyNA(matches[["b"]])) {
    stop(sprintf("`%s` cannot contain missing values.", arg_name))
  }

  if (!is.numeric(matches[["a"]]) || !is.numeric(matches[["b"]])) {
    stop(sprintf("`%s` should contain numeric row indices.", arg_name))
  }

  if (any(matches[["a"]] < 1) || any(matches[["b"]] < 1)) {
    stop(sprintf("`%s` should contain positive row indices.", arg_name))
  }

  if (any(matches[["a"]] != as.integer(matches[["a"]])) ||
      any(matches[["b"]] != as.integer(matches[["b"]]))) {
    stop(sprintf("`%s` should contain integer row indices.", arg_name))
  }

  if (any(matches[["a"]] > n_A) || any(matches[["b"]] > n_B)) {
    stop(sprintf("`%s` contains row indices outside the input datasets.", arg_name))
  }

  if (anyDuplicated(matches[, c("a", "b"), with = FALSE]) > 0L) {
    stop(sprintf("`%s` should not contain duplicate record pairs.", arg_name))
  }

  invisible(matches)
}

#' @noRd
validate_choice <- function(value, choices, arg_name) {
  if (length(value) != 1L || is.na(value) || !(value %in% choices)) {
    stop(sprintf("`%s` should be one of: %s.", arg_name, paste(choices, collapse = ", ")))
  }

  value
}

#' @noRd
validate_methods <- function(methods, variables, allowed_methods, default_method = "binary") {
  if (is.null(methods)) {
    methods <- list()
  } else if (!is.list(methods)) {
    stop("`methods` should be a list.")
  }

  methods <- unlist(methods, use.names = TRUE)
  missing_variables <- variables[!(variables %in% names(methods))]
  methods[missing_variables] <- default_method
  methods <- methods[variables]

  invalid_methods <- unique(methods[!(methods %in% allowed_methods)])
  if (length(invalid_methods) > 0L) {
    stop(
      sprintf(
        "`methods` contains unsupported values: %s. Allowed methods are: %s.",
        paste(invalid_methods, collapse = ", "),
        paste(allowed_methods, collapse = ", ")
      )
    )
  }

  methods
}

#' @noRd
extract_method_variables <- function(methods, include_hit_miss = FALSE) {
  gamma_names_for <- function(method_name) {
    variable_names <- names(methods[methods == method_name])

    if (length(variable_names) == 0L) {
      character()
    } else {
      paste0("gamma_", variable_names)
    }
  }

  method_variables <- list(
    b_vars = gamma_names_for("binary"),
    cpar_vars = gamma_names_for("continuous_parametric"),
    cnonpar_vars = gamma_names_for("continuous_nonparametric")
  )

  if (include_hit_miss) {
    method_variables[["hm_vars"]] <- gamma_names_for("hit_miss")
  }

  method_variables
}

#' @noRd
drop_constant_key_variables <- function(A, B, variables, comparators = NULL, methods = NULL) {
  unique_values <- vapply(variables, function(col) {
    data.table::uniqueN(c(A[[col]], B[[col]]))
  }, integer(1))
  constant_vars <- names(unique_values)[unique_values == 1L]

  if (length(constant_vars) > 0L) {
    A[, (constant_vars) := NULL]
    B[, (constant_vars) := NULL]
    variables <- variables[!(variables %in% constant_vars)]

    if (!is.null(comparators)) {
      comparators[constant_vars] <- NULL
    }

    if (!is.null(methods)) {
      methods <- methods[!(names(methods) %in% constant_vars)]
    }
  }

  if (length(variables) == 0L) {
    stop("All key variables have only one unique value and cannot be used for record linkage.", call. = FALSE)
  }

  list(
    A = A,
    B = B,
    variables = variables,
    comparators = comparators,
    methods = methods,
    constant_vars = constant_vars
  )
}

#' @noRd
sanitize_true_matches <- function(true_matches, n_A, n_B, arg_name = "true_matches") {
  if (is.null(true_matches)) {
    return(NULL)
  }

  if (!(is.data.frame(true_matches) || is.data.table(true_matches))) {
    warning(sprintf("`%s` should be a data.frame or a data.table. Setting `%s` to `NULL`.", arg_name, arg_name))
    return(NULL)
  }

  true_matches <- data.table::copy(data.table::as.data.table(true_matches))

  tryCatch(
    {
      validate_match_pairs(true_matches, n_A, n_B, arg_name = arg_name)
      true_matches
    },
    error = function(e) {
      warning(sprintf("%s Setting `%s` to `NULL`.", conditionMessage(e), arg_name))
      NULL
    }
  )
}

#' @noRd
align_parameter_table <- function(params, variables) {
  params[match(variables, params[["variable"]])]
}

#' @noRd
fit_kliep_hurdle_model <- function(gamma_numerator, gamma_denominator, variable, controls_kliep) {
  gamma_num_dt <- positive_gamma_table(gamma_numerator, variable)
  gamma_den_dt <- positive_gamma_table(gamma_denominator, variable)

  if (NROW(gamma_num_dt) < 2L || NROW(gamma_den_dt) < 2L) {
    return(NULL)
  }

  do.call(
    densityratio::kliep,
    c(
      list(
        df_numerator = gamma_num_dt,
        df_denominator = gamma_den_dt
      ),
      controls_kliep
    )
  )
}

#' @noRd
fit_kliep_hurdle_models <- function(df_numerator, df_denominator, variables, controls_kliep) {
  ratio_kliep_list <- lapply(variables, function(variable) {
    fit_kliep_hurdle_model(
      gamma_numerator = df_numerator[[variable]],
      gamma_denominator = df_denominator[[variable]],
      variable = variable,
      controls_kliep = controls_kliep
    )
  })
  names(ratio_kliep_list) <- variables
  ratio_kliep_list
}

#' @noRd
missing_kliep_models <- function(ratio_kliep_list) {
  names(which(vapply(ratio_kliep_list, is.null, logical(1))))
}

#' @noRd
predict_kliep_positive <- function(model, gamma, variable) {
  kliep_pred <- rep(1, length(gamma))
  positive_idx <- which(gamma > 0)

  if (length(positive_idx) == 0L) {
    return(kliep_pred)
  }

  gamma_dt <- data.table::data.table(value = gamma[positive_idx])
  data.table::setnames(gamma_dt, "value", variable)
  kliep_pred[positive_idx] <- as.vector(stats::predict(model, gamma_dt))
  kliep_pred
}

#' @noRd
warn_kliep_issue <- function(context, variables, message) {
  warning(
    sprintf(
      "KLIEP issue in %s for variable%s %s: %s",
      context,
      if (length(variables) == 1L) "" else "s",
      paste(variables, collapse = ", "),
      message
    ),
    call. = FALSE
  )
}

#' @noRd
# Keep aligned parameter vectors to avoid repeated lookups inside the MEC loop.
bernoulli_product <- function(df, probs) {
  density_product <- rep(1, NROW(df))

  for (i in seq_along(probs)) {
    density_product <- density_product * stats::dbinom(
      x = df[[i]],
      size = 1,
      prob = probs[i]
    )
  }

  density_product
}

#' @noRd
bernoulli_ratio <- function(df, numerator_probs, denominator_probs) {
  bernoulli_product(df, numerator_probs) / bernoulli_product(df, denominator_probs)
}

#' @noRd
hurdle_gamma_product <- function(df, p_0, alpha, beta) {
  density_product <- rep(1, NROW(df))

  for (i in seq_along(p_0)) {
    density_product <- density_product * hurdle_gamma_density(
      x = df[[i]],
      p_0 = p_0[i],
      alpha = alpha[i],
      beta = beta[i]
    )
  }

  density_product
}

#' @noRd
hurdle_gamma_ratio <- function(df,
                               p_0_numerator,
                               alpha_numerator,
                               beta_numerator,
                               p_0_denominator,
                               alpha_denominator,
                               beta_denominator) {
  hurdle_gamma_product(df, p_0_numerator, alpha_numerator, beta_numerator) /
    hurdle_gamma_product(df, p_0_denominator, alpha_denominator, beta_denominator)
}

#' @noRd
kliep_hurdle_ratio <- function(df, variables, p_0_numerator, p_0_denominator, ratio_kliep_list) {
  ratio_list <- lapply(seq_along(variables), function(i) {
    variable <- variables[i]
    gamma_vec <- df[[variable]]

    if (!is.null(ratio_kliep_list[[variable]])) {
      kliep_pred <- predict_kliep_positive(ratio_kliep_list[[variable]], gamma_vec, variable)
      ifelse(gamma_vec == 0, p_0_numerator[i] / p_0_denominator[i], 1) *
        ifelse(gamma_vec > 0, (1 - p_0_numerator[i]) / (1 - p_0_denominator[i]) * kliep_pred, 1)
    } else {
      ifelse(gamma_vec == 0, p_0_numerator[i] / p_0_denominator[i], 1)
    }
  })

  Reduce(`*`, ratio_list)
}

#' @noRd
# Summarize the selected pairs and error rates without rebuilding large subsets in each step.
summarize_mec_selection <- function(a,
                                    b,
                                    ratio,
                                    g_est,
                                    n_M_est,
                                    duplicates_in_A = FALSE,
                                    set_construction = "size",
                                    target_rate = 0.03,
                                    tol = 0.005,
                                    max_iter = 50) {
  validate_choice(set_construction, c("size", "flr", "mmr"), "set_construction")

  order_idx <- order(ratio, decreasing = TRUE)
  ratio_sorted <- ratio[order_idx]
  g_est_sorted <- g_est[order_idx]
  g_est_cumsum <- cumsum(g_est_sorted)

  if (set_construction == "size") {
    selected_idx <- select_mec_indices(
      a = a,
      b = b,
      ratio = ratio,
      n_M = n_M_est,
      duplicates_in_A = duplicates_in_A
    )
    n_selected <- length(selected_idx)
    selected_g_est <- g_est[selected_idx]
    # Report the size-based rates for the returned classification set.
    flr_est <- if (n_selected == 0L) Inf else mean(1 - selected_g_est)
    mmr_est <- if (n_selected == 0L) 1 else max(0, min(1, 1 - sum(selected_g_est) / n_selected))

    return(list(
      selected_idx = selected_idx,
      flr_est = flr_est,
      mmr_est = mmr_est,
      iter = NULL
    ))
  }

  min_treshold <- min(ratio)
  max_treshold <- max(ratio)
  treshold <- (min_treshold + max_treshold) / 2
  iter <- 0

  while (iter < max_iter) {
    n_selected <- findInterval(-treshold, -ratio_sorted)

    if (set_construction == "flr") {
      current_rate <- if (n_selected == 0L) Inf else (n_selected - g_est_cumsum[n_selected]) / n_selected
    } else {
      current_rate <- if (n_selected == 0L || n_M_est <= 0) 1 else 1 - g_est_cumsum[n_selected] / n_M_est
    }

    if (abs(current_rate - target_rate) <= tol) {
      break
    } else if (current_rate < target_rate) {
      if (set_construction == "flr") {
        max_treshold <- treshold
      } else {
        min_treshold <- treshold
      }
    } else {
      if (set_construction == "flr") {
        min_treshold <- treshold
      } else {
        max_treshold <- treshold
      }
    }

    treshold <- (min_treshold + max_treshold) / 2
    iter <- iter + 1
  }

  n_selected <- findInterval(-treshold, -ratio_sorted)
  selected_idx <- if (n_selected == 0L) integer() else order_idx[seq_len(n_selected)]
  flr_est <- if (n_selected == 0L) Inf else (n_selected - g_est_cumsum[n_selected]) / n_selected
  mmr_est <- if (n_selected == 0L || n_M_est <= 0) 1 else 1 - g_est_cumsum[n_selected] / n_M_est

  list(
    selected_idx = selected_idx,
    flr_est = flr_est,
    mmr_est = mmr_est,
    iter = iter
  )
}

#' @noRd
# Rebuild the greedy MEC set from row indices instead of repeated table rbinding.
select_mec_indices <- function(a, b, ratio, n_M, duplicates_in_A = FALSE) {
  order_idx <- order(ratio, decreasing = TRUE)
  n_target <- round(n_M)

  if (n_target <= 0L || length(order_idx) == 0L) {
    return(integer())
  }

  selected_idx <- integer(length(order_idx))
  n_selected <- 0L
  used_a <- rep(FALSE, max(a))

  if (!duplicates_in_A) {
    used_b <- rep(FALSE, max(b))
  }

  for (idx in order_idx) {
    current_a <- a[idx]

    if (duplicates_in_A) {
      if (!used_a[current_a]) {
        n_selected <- n_selected + 1L
        selected_idx[n_selected] <- idx
        used_a[current_a] <- TRUE
      }
    } else {
      current_b <- b[idx]

      if (!used_a[current_a] && !used_b[current_b]) {
        n_selected <- n_selected + 1L
        selected_idx[n_selected] <- idx
        used_a[current_a] <- TRUE
        used_b[current_b] <- TRUE
      }
    }

    if (n_selected >= n_M) {
      break
    }
  }

  utils::head(selected_idx[seq_len(n_selected)], n_target)
}

#' @noRd
fixed_n_M <- function(n, ratio_gamma) {
  function(n_M) {
    sum(pmin(n_M * ratio_gamma / (n_M * (ratio_gamma - 1) + n), 1))
  }
}

#' @noRd
validate_blocking_controls <- function(controls_blocking) {
  if (!is.list(controls_blocking)) {
    stop("`controls_blocking` should be a list.")
  }

  protected_args <- intersect(names(controls_blocking), c("x", "y"))
  if (length(protected_args) > 0L) {
    stop("`controls_blocking` should not contain `x` or `y`.")
  }

  controls_blocking
}

#' @noRd
blocking_input_length <- function(x) {
  if (is.null(dim(x))) {
    length(x)
  } else {
    nrow(x)
  }
}

#' @noRd
build_blocking_input <- function(data, blocking_variables, blocking_sep) {
  if (!is.character(blocking_variables) || length(blocking_variables) == 0L) {
    stop("`blocking_variables` should be a non-empty character vector.")
  }

  if (!all(blocking_variables %in% names(data))) {
    stop("Not all `blocking_variables` are present in the input datasets.")
  }

  if (length(blocking_sep) != 1L || is.na(blocking_sep)) {
    stop("`blocking_sep` should be a single non-missing character value.")
  }

  if (length(blocking_variables) == 1L) {
    return(as.character(data[[blocking_variables]]))
  }

  blocking_data <- data[, blocking_variables, with = FALSE]
  do.call(paste, c(blocking_data, sep = blocking_sep))
}

#' @noRd
prepare_blocking_inputs <- function(A,
                                    B,
                                    blocking_x,
                                    blocking_y,
                                    blocking_variables,
                                    blocking_sep) {
  if (is.null(blocking_x) && is.null(blocking_y)) {
    blocking_x <- build_blocking_input(A, blocking_variables, blocking_sep)
    blocking_y <- build_blocking_input(B, blocking_variables, blocking_sep)
  } else if (is.null(blocking_x) || is.null(blocking_y)) {
    stop("`blocking_x` and `blocking_y` should both be supplied or both be `NULL`.")
  }

  if (blocking_input_length(blocking_x) != nrow(A)) {
    stop("`blocking_x` should have one row or element for each record in `A`.")
  }

  if (blocking_input_length(blocking_y) != nrow(B)) {
    stop("`blocking_y` should have one row or element for each record in `B`.")
  }

  list(x = blocking_x, y = blocking_y)
}

#' @noRd
run_blocking <- function(blocking_x, blocking_y, controls_blocking) {
  controls_blocking <- validate_blocking_controls(controls_blocking)

  do.call(
    blocking::blocking,
    c(
      list(x = blocking_x, y = blocking_y),
      controls_blocking
    )
  )
}

#' @noRd
extract_blocking_result_table <- function(blocking_result) {
  if (is.null(blocking_result[["result"]])) {
    stop("The object returned by `blocking::blocking()` should contain a `result` element.")
  }

  result <- data.table::as.data.table(blocking_result[["result"]])
  required_cols <- c("x", "y", "block")

  if (!all(required_cols %in% names(result))) {
    stop("`blocking::blocking()` result should contain columns x, y, and block.")
  }

  if (NROW(result) == 0L) {
    stop("`blocking::blocking()` did not return any candidate pairs.")
  }

  result[, required_cols, with = FALSE]
}

#' @noRd
reconstruct_block_summary <- function(blocking_table, n_A, n_B) {
  blocking_table <- data.table::copy(blocking_table)
  data.table::setnames(blocking_table, c("x", "y"), c("a", "b"))

  if (anyNA(blocking_table[["a"]]) || anyNA(blocking_table[["b"]]) ||
      anyNA(blocking_table[["block"]])) {
    stop("Blocking result cannot contain missing values in x, y, or block.")
  }

  if (!is.numeric(blocking_table[["a"]]) || !is.numeric(blocking_table[["b"]])) {
    stop("Blocking result columns x and y should contain numeric row indices.")
  }

  if (any(blocking_table[["a"]] != as.integer(blocking_table[["a"]])) ||
      any(blocking_table[["b"]] != as.integer(blocking_table[["b"]]))) {
    stop("Blocking result columns x and y should contain integer row indices.")
  }

  if (any(blocking_table[["a"]] < 1) || any(blocking_table[["a"]] > n_A) ||
      any(blocking_table[["b"]] < 1) || any(blocking_table[["b"]] > n_B)) {
    stop("Blocking result contains row indices outside the input datasets.")
  }

  x_blocks <- unique(blocking_table[, .(a, block)])
  y_blocks <- unique(blocking_table[, .(b, block)])

  if (anyDuplicated(x_blocks[["a"]]) > 0L ||
      anyDuplicated(y_blocks[["b"]]) > 0L) {
    stop("Final blocks should be disjoint on both input files.")
  }

  A_by_block <- x_blocks[order(block, a), .(A = list(a), n_A = .N), by = block]
  B_by_block <- y_blocks[order(block, b), .(B = list(b), n_B = .N), by = block]
  block_summary <- merge(A_by_block, B_by_block, by = "block", all = FALSE)
  data.table::setorder(block_summary, block)
  block_summary[, pair_count := n_A * n_B]
  block_summary[, nonmatches_min := pair_count - pmin(n_A, n_B)]

  excluded_records <- list(
    A = setdiff(seq_len(n_A), x_blocks[["a"]]),
    B = setdiff(seq_len(n_B), y_blocks[["b"]])
  )

  list(
    block_summary = block_summary,
    excluded_records = excluded_records
  )
}

#' @noRd
make_block_pair_table <- function(block_summary, blocks = NULL) {
  if (is.null(blocks)) {
    selected_blocks <- block_summary[["block"]]
  } else {
    selected_blocks <- blocks
  }

  block_summary <- block_summary[block %in% selected_blocks]
  pair_list <- lapply(seq_len(NROW(block_summary)), function(i) {
    pairs <- data.table::CJ(
      a = block_summary[["A"]][[i]],
      b = block_summary[["B"]][[i]]
    )
    data.table::setkey(pairs, NULL)
    pairs[, block := block_summary[["block"]][i]]
    pairs[]
  })

  data.table::rbindlist(pair_list, use.names = TRUE)
}

#' @noRd
exact_match_pairs <- function(A, B, variables) {
  merge(
    A[, c("a", variables), with = FALSE],
    B[, c("b", variables), with = FALSE],
    by = variables
  )[, c("a", "b"), with = FALSE]
}

#' @noRd
score_mec_ratio <- function(Omega, model) {
  data.table::set(Omega, j = "ratio", value = 1)

  if (!is.null(model$b_vars)) {
    b_vars <- model$b_vars
    b_params <- align_parameter_table(model$b_params, b_vars)
    Omega_b <- Omega[, b_vars, with = FALSE]
    data.table::set(
      Omega,
      j = "ratio",
      value = Omega[["ratio"]] * bernoulli_ratio(Omega_b, b_params$theta, b_params$eta)
    )
  }

  if (!is.null(model$cpar_vars)) {
    cpar_vars <- model$cpar_vars
    cpar_params <- data.table::copy(align_parameter_table(model$cpar_params, cpar_vars))
    if ("p_0_U" %in% names(cpar_params)) {
      p_0_denominator <- cpar_params$p_0_U
      alpha_denominator <- cpar_params$alpha_U
      beta_denominator <- cpar_params$beta_U
    } else {
      p_0_denominator <- cpar_params$p_0_Omega
      alpha_denominator <- cpar_params$alpha_Omega
      beta_denominator <- cpar_params$beta_Omega
    }
    Omega_cpar <- Omega[, cpar_vars, with = FALSE]
    data.table::set(
      Omega,
      j = "ratio",
      value = Omega[["ratio"]] * hurdle_gamma_ratio(
        Omega_cpar,
        cpar_params$p_0_M,
        cpar_params$alpha_M,
        cpar_params$beta_M,
        p_0_denominator,
        alpha_denominator,
        beta_denominator
      )
    )
  }

  if (!is.null(model$cnonpar_vars)) {
    cnonpar_vars <- model$cnonpar_vars
    Omega_cnonpar <- Omega[, cnonpar_vars, with = FALSE]

    if (!is.null(model$ratio_kliep_list)) {
      ratio_kliep_list <- model$ratio_kliep_list
      p_0_M_cnonpar <- model$cnonpar_params[["p_0_M_cnonpar"]]
      p_0_U_cnonpar <- model$cnonpar_params[["p_0_U_cnonpar"]]
      names(p_0_M_cnonpar) <- cnonpar_vars
      names(p_0_U_cnonpar) <- cnonpar_vars

      ratio_kliep <- kliep_hurdle_ratio(
        Omega_cnonpar,
        cnonpar_vars,
        p_0_M_cnonpar,
        p_0_U_cnonpar,
        ratio_kliep_list
      )
      data.table::set(Omega, j = "ratio", value = Omega[["ratio"]] * ratio_kliep)
    } else if (!is.null(model$ratio_kliep)) {
      data.table::set(
        Omega,
        j = "ratio",
        value = Omega[["ratio"]] * stats::predict(model$ratio_kliep, Omega_cnonpar)
      )
    }
  }

  if (!is.null(model$hm_vars)) {
    hm_vars <- model$hm_vars
    hm_params <- align_parameter_table(model$hm_params, hm_vars)
    Omega_hm <- Omega[, hm_vars, with = FALSE]
    data.table::set(
      Omega,
      j = "ratio",
      value = Omega[["ratio"]] * bernoulli_ratio(Omega_hm, hm_params$theta, hm_params$eta)
    )
  }

  Omega
}

#' @noRd
blocking_diagnostics <- function(true_matches, block_pairs, n_full_pairs) {
  true_matches <- data.table::copy(true_matches)
  data.table::setkey(true_matches, a, b)
  block_pairs_key <- data.table::copy(block_pairs[, c("a", "b"), with = FALSE])
  data.table::setkey(block_pairs_key, a, b)

  preserved <- true_matches[block_pairs_key, nomatch = 0L]
  n_true <- NROW(true_matches)
  n_preserved <- NROW(preserved)

  data.table::data.table(
    true_matches = n_true,
    preserved_matches = n_preserved,
    lost_matches = n_true - n_preserved,
    blocking_recall = if (n_true == 0L) NA_real_ else n_preserved / n_true,
    blocking_fnr = if (n_true == 0L) NA_real_ else 1 - n_preserved / n_true,
    blocked_pairs = NROW(block_pairs),
    full_pairs = n_full_pairs
  )
}

#' @noRd
mec_selection_diagnostics <- function(true_matches, block_pairs, M_est) {
  true_matches <- data.table::copy(data.table::as.data.table(true_matches))
  block_pairs <- data.table::copy(data.table::as.data.table(block_pairs))[, .(a, b)]
  M_pairs <- data.table::copy(data.table::as.data.table(M_est))[, .(a, b)]

  candidate_matches <- NROW(block_pairs[true_matches, on = .(a, b), nomatch = 0L])
  selected_candidate_matches <- NROW(M_pairs[true_matches, on = .(a, b), nomatch = 0L])
  selected_pairs <- NROW(M_pairs)
  true_match_count <- NROW(true_matches)
  blocking_lost_matches <- true_match_count - candidate_matches
  mec_missed_candidate_matches <- candidate_matches - selected_candidate_matches
  false_links <- selected_pairs - selected_candidate_matches

  data.table::data.table(
    true_matches = true_match_count,
    candidate_matches = candidate_matches,
    blocking_lost_matches = blocking_lost_matches,
    selected_candidate_matches = selected_candidate_matches,
    mec_missed_candidate_matches = mec_missed_candidate_matches,
    mec_candidate_recall = if (candidate_matches == 0L) NA_real_ else selected_candidate_matches / candidate_matches,
    mec_candidate_fnr = if (candidate_matches == 0L) NA_real_ else mec_missed_candidate_matches / candidate_matches,
    selected_pairs = selected_pairs,
    false_links = false_links,
    empirical_flr = if (selected_pairs == 0L) NA_real_ else false_links / selected_pairs
  )
}

#' @noRd
fit_mec_unsupervised_omega <- function(A,
                                       B,
                                       variables,
                                       comparators,
                                       methods,
                                       Omega,
                                       exact_matches,
                                       start_params = NULL,
                                       nonmatch_params = NULL,
                                       nonpar_hurdle = TRUE,
                                       delta = 0.5,
                                       eps = 0.05,
                                       max_iter_em = 10,
                                       tol_em = 1,
                                       controls_nleqslv = list(),
                                       controls_kliep = control_kliep(),
                                       context = "mec_blocking()") {
  if (!is.null(start_params)) {
    stopifnot("`start_params` should be a list." =
                is.list(start_params))
  }
  if (!is.null(nonmatch_params)) {
    stopifnot("`nonmatch_params` should be a list." =
                is.list(nonmatch_params))
  }

  method_variables <- extract_method_variables(methods, include_hit_miss = TRUE)
  b_vars <- method_variables$b_vars
  cpar_vars <- method_variables$cpar_vars
  cnonpar_vars <- method_variables$cnonpar_vars
  hm_vars <- method_variables$hm_vars

  exact_match_idx <- data.table(
    omega_idx = seq_len(NROW(Omega)),
    a = Omega[["a"]],
    b = Omega[["b"]]
  )[exact_matches[, c("a", "b"), with = FALSE], on = c("a", "b"), nomatch = 0L][["omega_idx"]]

  if (length(exact_match_idx) == 0L) {
    stop("The MEC training space contains no exact agreements on the key variables.")
  }

  n <- NROW(Omega)
  exact_match_mask <- rep(FALSE, n)
  exact_match_mask[exact_match_idx] <- TRUE
  M_idx <- exact_match_idx
  n_M <- length(M_idx)
  all_idx <- seq_len(n)
  selection_mask <- rep(FALSE, n)

  b_params <- NULL
  cpar_params <- NULL
  cnonpar_params <- NULL
  hm_params <- NULL
  ratio_kliep <- NULL
  ratio_kliep_list <- NULL
  kliep_warning_emitted <- FALSE

  warn_kliep_once <- function(variables, message) {
    if (!kliep_warning_emitted) {
      warn_kliep_issue(context, variables, message)
      kliep_warning_emitted <<- TRUE
    }
  }

  data.table::set(Omega, j = "ratio", value = 1)

  if (is.null(start_params)) {
    start_params <- list()

    if (length(b_vars) > 0L) {
      start_params[["binary"]] <- data.table(
        variable = b_vars,
        theta = runif(length(b_vars), min = 0.9)
      )
    }

    if (length(cpar_vars) > 0L) {
      start_params[["continuous_parametric"]] <- data.table(
        variable = cpar_vars,
        p_0_M = runif(length(cpar_vars), min = 0.8, max = 0.9),
        alpha_M = runif(length(cpar_vars), min = 0.1, max = 1),
        beta_M = runif(length(cpar_vars), min = 10, max = 20)
      )
    }

    if (length(hm_vars) > 0L) {
      start_params[["hit_miss"]] <- data.table(
        variable = hm_vars,
        theta = runif(length(hm_vars), min = 0.9)
      )
    }
  }

  if (length(b_vars) > 0L) {
    b_params <- align_parameter_table(start_params$binary, b_vars)
    Omega_b <- Omega[, b_vars, with = FALSE]
    if (!is.null(nonmatch_params) && !is.null(nonmatch_params$binary)) {
      eta_b <- align_parameter_table(nonmatch_params$binary, b_vars)$eta
    } else {
      eta_b <- binary_formula(Omega_b)
    }
    b_params$eta <- eta_b
    b_denominator <- bernoulli_product(Omega_b, b_params$eta)
    data.table::set(
      Omega,
      j = "ratio",
      value = Omega[["ratio"]] * bernoulli_ratio(Omega_b, b_params$theta, b_params$eta)
    )
    data.table::set(Omega, j = "b_denominator", value = b_denominator)
  }

  if (length(cpar_vars) > 0L) {
    cpar_params <- align_parameter_table(start_params$continuous_parametric, cpar_vars)
    Omega_cpar <- Omega[, cpar_vars, with = FALSE]
    modified_nleqslv <- purrr::partial(nleqslv::nleqslv, control = controls_nleqslv)
    if (!is.null(nonmatch_params) && !is.null(nonmatch_params$continuous_parametric)) {
      cpar_nonmatch_params <- align_parameter_table(nonmatch_params$continuous_parametric, cpar_vars)
      p_0_U <- cpar_nonmatch_params$p_0_U
      alpha_U <- cpar_nonmatch_params$alpha_U
      beta_U <- cpar_nonmatch_params$beta_U
    } else {
      p_0_U <- p_0_formula(Omega_cpar)
      gamma_plus_U <- gamma_plus_formula(Omega_cpar)
      alpha_U <- alpha_formula(Omega_cpar, modified_nleqslv)
      beta_U <- alpha_U / gamma_plus_U
    }
    cpar_params$p_0_U <- p_0_U
    cpar_params$alpha_U <- alpha_U
    cpar_params$beta_U <- beta_U
    cpar_denominator <- hurdle_gamma_product(
      Omega_cpar,
      cpar_params$p_0_U,
      cpar_params$alpha_U,
      cpar_params$beta_U
    )
    data.table::set(
      Omega,
      j = "ratio",
      value = Omega[["ratio"]] * hurdle_gamma_ratio(
        Omega_cpar,
        cpar_params$p_0_M,
        cpar_params$alpha_M,
        cpar_params$beta_M,
        cpar_params$p_0_U,
        cpar_params$alpha_U,
        cpar_params$beta_U
      )
    )
    data.table::set(Omega, j = "cpar_denominator", value = cpar_denominator)
  }

  if (length(cnonpar_vars) > 0L) {
    Omega_cnonpar <- Omega[, cnonpar_vars, with = FALSE]
    n_exact_matches <- sum(exact_match_mask)

    if (nonpar_hurdle) {
      p_0_M_cnonpar <- runif(length(cnonpar_vars), min = 0.5)
      p_0_U_cnonpar <- p_0_formula(Omega_cnonpar)
      names(p_0_M_cnonpar) <- cnonpar_vars
      cnonpar_params <- data.table(
        variable = cnonpar_vars,
        p_0_M_cnonpar = p_0_M_cnonpar,
        p_0_U_cnonpar = p_0_U_cnonpar
      )

      ratio_temp <- lapply(cnonpar_vars, function(x) {
        r <- numeric(n)
        r[exact_match_mask] <- stats::runif(n_exact_matches, min = 5, max = 10)
        r[!exact_match_mask] <- stats::runif(n - n_exact_matches, min = 0.1, max = 1)
        r
      })
      names(ratio_temp) <- cnonpar_vars
      cnonpar_ratio_list <- lapply(cnonpar_vars, function(x) {
        gamma_vec <- Omega_cnonpar[[x]]
        ifelse(gamma_vec == 0, p_0_M_cnonpar[x] / p_0_U_cnonpar[x], 1) *
          ifelse(gamma_vec > 0, (1 - p_0_M_cnonpar[x]) / (1 - p_0_U_cnonpar[x]) * ratio_temp[[x]], 1)
      })
      ratio_kliep <- Reduce(`*`, cnonpar_ratio_list)
      data.table::set(Omega, j = "ratio", value = Omega[["ratio"]] * ratio_kliep)
    } else {
      ratio_temp <- as.numeric(Omega$ratio)
      ratio_temp[exact_match_mask] <- ratio_temp[exact_match_mask] * stats::runif(
        n_exact_matches,
        min = 5,
        max = 10
      )
      ratio_temp[!exact_match_mask] <- ratio_temp[!exact_match_mask] * stats::runif(
        n - n_exact_matches,
        min = 0.1,
        max = 5
      )
      data.table::set(Omega, j = "ratio", value = ratio_temp)
    }
  }

  if (length(hm_vars) > 0L) {
    hm_params <- align_parameter_table(start_params$hit_miss, hm_vars)
    Omega_hm <- Omega[, hm_vars, with = FALSE]
    eta_hm <- binary_formula(Omega_hm)
    hm_params$eta <- eta_hm
    hm_denominator <- bernoulli_product(Omega_hm, hm_params$eta)
    data.table::set(
      Omega,
      j = "ratio",
      value = Omega[["ratio"]] * bernoulli_ratio(Omega_hm, hm_params$theta, hm_params$eta)
    )
    data.table::set(Omega, j = "hm_denominator", value = hm_denominator)

    values_list_m <- lapply(hm_vars, function(x) {
      name <- substr(x, 7, nchar(x))
      values <- unique(c(A[[name]], B[[name]]))
      m_est <- sapply(values, function(y) sum(A[[name]] == y) / NROW(A))
      data.table::data.table(
        "value" = values,
        "m_est" = m_est
      )
    })
    names(values_list_m) <- hm_vars
  }

  iter <- 1

  repeat {
    g_est <- pmin(length(M_idx) * Omega$ratio / (length(M_idx) * (Omega$ratio - 1) + n), 1)
    n_M_old <- n_M
    n_M <- sum(g_est)

    if (n_M > min(NROW(A), NROW(B))) {
      n_M <- min(NROW(A), NROW(B))
    }

    data.table::set(Omega, j = "g_est", value = g_est)
    M_idx <- select_mec_indices(
      a = Omega[["a"]],
      b = Omega[["b"]],
      ratio = Omega[["ratio"]],
      n_M = n_M,
      duplicates_in_A = FALSE
    )

    if (length(M_idx) == 0L) {
      break
    }

    selection_mask[M_idx] <- TRUE
    U_idx <- all_idx[!selection_mask]
    selection_mask[M_idx] <- FALSE

    if (iter >= 2) {
      old_params <- c()
      params <- c()
      if (length(b_vars) > 0L) {
        old_params <- c(old_params, theta_b_old)
        params <- c(params, theta_b)
      }
      if (length(cpar_vars) > 0L) {
        old_params <- c(old_params, p_0_M_old, alpha_M_old, beta_M_old)
        params <- c(params, p_0_M, alpha_M, beta_M)
      }
      if (length(hm_vars) > 0L) {
        old_params <- c(old_params, theta_hm_old)
        params <- c(params, theta_hm)
      }

      if (length(cnonpar_vars) == 0L) {
        if ((abs(n_M_old - n_M) < delta) || norm(old_params - params, type = "2") < eps) {
          break
        }
      } else if ((abs(n_M_old - n_M) < delta)) {
        break
      }
    }

    data.table::set(Omega, j = "ratio", value = 1)

    if (length(b_vars) > 0L) {
      M_b <- Omega_b[M_idx]
      theta_b_old <- b_params$theta
      theta_b <- binary_formula(M_b)
      b_params$theta <- theta_b
      b_numerator <- bernoulli_product(Omega_b, b_params$theta)
      data.table::set(Omega, j = "ratio", value = Omega[["ratio"]] * b_numerator / Omega[["b_denominator"]])
    }

    if (length(cpar_vars) > 0L) {
      M_cpar <- Omega_cpar[M_idx]
      p_0_M_old <- cpar_params$p_0_M
      alpha_M_old <- cpar_params$alpha_M
      beta_M_old <- cpar_params$beta_M
      p_0_M <- p_0_formula(M_cpar)
      gamma_plus_M <- gamma_plus_formula(M_cpar)
      alpha_M <- alpha_formula(M_cpar, modified_nleqslv)
      beta_M <- alpha_M / gamma_plus_M
      beta_M[is.nan(beta_M)] <- beta_M_old[is.nan(beta_M)]
      cpar_params$p_0_M <- p_0_M
      cpar_params$alpha_M <- alpha_M
      cpar_params$beta_M <- beta_M
      cpar_numerator <- hurdle_gamma_product(
        Omega_cpar,
        cpar_params$p_0_M,
        cpar_params$alpha_M,
        cpar_params$beta_M
      )
      data.table::set(Omega, j = "ratio", value = Omega[["ratio"]] * cpar_numerator / Omega[["cpar_denominator"]])
    }

    if (length(cnonpar_vars) > 0L) {
      M_cnonpar <- Omega_cnonpar[M_idx]
      U_cnonpar <- Omega_cnonpar[U_idx]

      if (nonpar_hurdle) {
        p_0_M_cnonpar <- p_0_formula(M_cnonpar)
        p_0_U_cnonpar <- p_0_formula(U_cnonpar)
        cnonpar_params <- data.table(
          variable = cnonpar_vars,
          p_0_M_cnonpar = p_0_M_cnonpar,
          p_0_U_cnonpar = p_0_U_cnonpar
        )
        ratio_kliep_old <- ratio_kliep

        tryCatch({
          ratio_kliep_models <- fit_kliep_hurdle_models(
            df_numerator = M_cnonpar,
            df_denominator = U_cnonpar,
            variables = cnonpar_vars,
            controls_kliep = controls_kliep
          )
          missing_models <- missing_kliep_models(ratio_kliep_models)

          if (length(missing_models) < length(cnonpar_vars)) {
            if (length(missing_models) > 0L) {
              warn_kliep_once(missing_models, "using only the hurdle mass term for those variables in the current iteration.")
            }

            ratio_kliep_list <- ratio_kliep_models
            ratio_kliep <- kliep_hurdle_ratio(
              df = Omega_cnonpar,
              variables = cnonpar_vars,
              p_0_numerator = p_0_M_cnonpar,
              p_0_denominator = p_0_U_cnonpar,
              ratio_kliep_list = ratio_kliep_models
            )
          } else {
            ratio_kliep <- ratio_kliep_old
            warn_kliep_once(cnonpar_vars, "could not be fitted in the current iteration; using the previous ratio estimate.")
          }
          data.table::set(Omega, j = "ratio", value = Omega[["ratio"]] * ratio_kliep)
        },
          error = function(e) {
            warn_kliep_once(cnonpar_vars, paste("failed with error:", e$message))
          })
      } else {
        ratio_kliep <- do.call(
          densityratio::kliep,
          c(list(
            df_numerator = M_cnonpar,
            df_denominator = U_cnonpar
          ),
          controls_kliep)
        )
        data.table::set(Omega, j = "ratio", value = Omega[["ratio"]] * stats::predict(ratio_kliep, Omega_cnonpar))
      }
    }

    if (length(hm_vars) > 0L) {
      M_hm <- Omega_hm[M_idx]
      theta_hm_old <- hm_params$theta
      theta_hm <- binary_formula(M_hm)
      hm_params$theta <- theta_hm

      p_est <- n_M / max(NROW(A), NROW(B))
      u_est_list <- lapply(hm_vars, function(x) {
        u_est <- runif(NROW(values_list_m[[x]]), 0, 1)
        u_est <- u_est / sum(u_est)
      })
      names(u_est_list) <- hm_vars

      m_bk_prod <- sapply(seq_len(NROW(B)), function(x) {
        m_bk_list <- lapply(hm_vars, function(y) {
          var_name <- substr(y, 7, nchar(y))
          d_value <- (B[[var_name]])[x]
          as.vector((values_list_m[[y]])[(values_list_m[[y]])[["value"]] == d_value, ][["m_est"]])
        })
        names(m_bk_list) <- hm_vars
        Reduce(`*`, m_bk_list)
      })

      iter_em <- 1
      while (iter_em <= max_iter_em + 1) {
        if (iter_em >= 2) {
          delta_b_old <- em_params[["delta_b"]]
        }

        delta_m_u <- lapply(seq_len(NROW(B)), function(x) {
          u_bk_list <- lapply(hm_vars, function(y) {
            var_name <- substr(y, 7, nchar(y))
            values <- cbind(values_list_m[[y]], u_est_list[[y]])
            colnames(values) <- c("value", "m_est", "u_est")
            d_value <- (B[[var_name]])[x]
            as.vector(values[values[["value"]] == d_value, ][["u_est"]])
          })
          names(u_bk_list) <- hm_vars
          u_bk_prod <- Reduce(`*`, u_bk_list)
          data.table::data.table(
            delta_b = p_est * m_bk_prod[x] / (p_est * m_bk_prod[x] + (1 - p_est) * u_bk_prod),
            m_bk_prod = m_bk_prod[x],
            u_bk_prod = u_bk_prod
          )
        })

        em_params <- data.table::rbindlist(delta_m_u)

        if (iter_em >= 3) {
          log_lik_old <- log_lik
        }

        if (iter_em >= 2) {
          log_lik <- sum(
            ifelse(delta_b_old == 0,
                   0,
                   delta_b_old * log(p_est * em_params[["m_bk_prod"]]))
          ) +
            sum(ifelse(em_params[["u_bk_prod"]] == 0,
                       0,
                       (1 - delta_b_old) * log((1 - p_est) * em_params[["u_bk_prod"]]))
                )
        }

        if (iter_em >= 3) {
          if (abs(log_lik - log_lik_old) <= tol_em) break
        }

        u_est_list <- lapply(hm_vars, function(x) {
          var_name <- substr(x, 7, nchar(x))
          z_bk <- B[[var_name]]

          sapply((values_list_m[[x]])[["value"]], function(y) {
            sum((1 - em_params[["delta_b"]]) * (z_bk == y)) / sum(1 - em_params[["delta_b"]])
          })
        })
        names(u_est_list) <- hm_vars

        iter_em <- iter_em + 1
      }

      hm_numerator <- bernoulli_product(Omega_hm, hm_params$theta)
      eta_hm <- sapply(hm_vars, function(col) {
        ((1 - p_est) * sum(u_est_list[[col]] * (values_list_m[[col]])[["m_est"]]) +
           p_est * (1 - 1 / NROW(A)) * sum(((values_list_m[[col]])[["m_est"]]) ^ 2)) /
          (1 - p_est / NROW(A))
      })
      hm_params$eta <- eta_hm
      hm_denominator <- bernoulli_product(Omega_hm, hm_params$eta)
      data.table::set(Omega, j = "hm_denominator", value = hm_denominator)
      data.table::set(Omega, j = "ratio", value = Omega[["ratio"]] * hm_numerator / Omega[["hm_denominator"]])
    }

    iter <- iter + 1
  }

  n_M_est <- n_M
  selection_summary <- summarize_mec_selection(
    a = Omega[["a"]],
    b = Omega[["b"]],
    ratio = Omega[["ratio"]],
    g_est = Omega[["g_est"]],
    n_M_est = n_M_est,
    duplicates_in_A = FALSE,
    set_construction = "size"
  )
  M <- Omega[selection_summary$selected_idx]

  model <- list(
    b_vars = if (length(b_vars) == 0L) NULL else b_vars,
    cpar_vars = if (length(cpar_vars) == 0L) NULL else cpar_vars,
    cnonpar_vars = if (length(cnonpar_vars) == 0L) NULL else cnonpar_vars,
    hm_vars = if (length(hm_vars) == 0L) NULL else hm_vars,
    b_params = if (is.null(b_params)) NULL else b_params,
    cpar_params = if (is.null(cpar_params)) NULL else cpar_params,
    cnonpar_params = if (is.null(cnonpar_params)) NULL else cnonpar_params,
    hm_params = if (is.null(hm_params)) NULL else hm_params,
    ratio_kliep = if (is.null(ratio_kliep) || is.numeric(ratio_kliep)) NULL else ratio_kliep,
    ratio_kliep_list = if (is.null(ratio_kliep_list)) NULL else ratio_kliep_list,
    variables = variables,
    comparators = comparators,
    methods = methods,
    n_M_est = n_M_est,
    prob_est = n_M_est / n
  )

  list(
    model = model,
    Omega = Omega,
    M_est = M[, c("a", "b", "ratio"), with = FALSE],
    n_M_est = n_M_est,
    flr_est = selection_summary$flr_est,
    mmr_est = selection_summary$mmr_est
  )
}

#' @noRd
validate_mec_blocking_alpha <- function(alpha) {
  if (!is.numeric(alpha) || length(alpha) != 1L || is.na(alpha) ||
      !is.finite(alpha) || alpha < 0 || alpha >= 1) {
    stop("`alpha` should be a single numeric value in [0, 1).", call. = FALSE)
  }

  invisible(alpha)
}

#' @noRd
initial_inverted_match_count <- function(nu) {
  as.integer(max(0L, nu))
}

#' @noRd
prepare_inverted_start_params <- function(start_params, cpar_vars) {
  if (!is.null(start_params)) {
    stopifnot("`start_params` should be a list." =
                is.list(start_params))
    return(start_params)
  }

  start_params <- list()
  if (length(cpar_vars) > 0L) {
    start_params[["continuous_parametric"]] <- data.table(
      variable = cpar_vars,
      p_0_M = stats::runif(length(cpar_vars), min = 0.8, max = 0.9),
      alpha_M = stats::runif(length(cpar_vars), min = 0.1, max = 1),
      beta_M = stats::runif(length(cpar_vars), min = 10, max = 20)
    )
  }

  start_params
}

#' @noRd
get_hurdle_gamma_fallback <- function(fallback_params, variables, side) {
  if (is.null(fallback_params)) {
    return(NULL)
  }

  alpha_col <- paste0("alpha_", side)
  beta_col <- paste0("beta_", side)
  if (!all(c("variable", alpha_col, beta_col) %in% names(fallback_params))) {
    return(NULL)
  }

  fallback_params <- align_parameter_table(fallback_params, variables)
  list(
    alpha = fallback_params[[alpha_col]],
    beta = fallback_params[[beta_col]]
  )
}

#' @noRd
estimate_hurdle_gamma_params <- function(df,
                                         variables,
                                         side,
                                         fallback_params = NULL,
                                         controls_nleqslv = list(),
                                         context = "mec_blocking()") {
  if (length(variables) == 0L) {
    return(NULL)
  }

  p_0 <- p_0_formula(df)
  alpha <- stats::setNames(rep(NA_real_, length(variables)), variables)
  beta <- stats::setNames(rep(NA_real_, length(variables)), variables)
  fallback <- get_hurdle_gamma_fallback(fallback_params, variables, side)
  modified_nleqslv <- purrr::partial(nleqslv::nleqslv, control = controls_nleqslv)

  for (i in seq_along(variables)) {
    variable <- variables[i]
    gamma <- df[[variable]]
    positive_gamma <- gamma[gamma > 0]

    if (length(positive_gamma) >= 2L) {
      estimated_alpha <- tryCatch(
        alpha_formula(df[, variable, with = FALSE], modified_nleqslv)[1L],
        error = function(e) NA_real_
      )
      estimated_beta <- estimated_alpha / mean(positive_gamma)

      if (is.finite(estimated_alpha) && is.finite(estimated_beta) &&
          estimated_alpha > 0 && estimated_beta > 0) {
        alpha[i] <- estimated_alpha
        beta[i] <- estimated_beta
        next
      }
    }

    has_fallback <- !is.null(fallback) &&
      is.finite(fallback$alpha[i]) && is.finite(fallback$beta[i]) &&
      fallback$alpha[i] > 0 && fallback$beta[i] > 0

    if (!has_fallback) {
      stop(
        sprintf(
          "%s cannot estimate %s-side Gamma parameters for `%s`; at least two positive continuous comparisons or finite fallback parameters are required.",
          context,
          if (side == "M") "match" else "nonmatch",
          variable
        ),
        call. = FALSE
      )
    }

    alpha[i] <- fallback$alpha[i]
    beta[i] <- fallback$beta[i]
  }

  params <- data.table(variable = variables)
  data.table::set(params, j = paste0("p_0_", side), value = as.numeric(p_0))
  data.table::set(params, j = paste0("alpha_", side), value = as.numeric(alpha))
  data.table::set(params, j = paste0("beta_", side), value = as.numeric(beta))
  params
}

#' @noRd
estimate_inverted_match_parameters <- function(Omega,
                                               M_idx,
                                               b_vars,
                                               cpar_vars,
                                               previous_params,
                                               start_params,
                                               controls_nleqslv,
                                               context = "mec_blocking()") {
  if (length(M_idx) == 0L) {
    M_idx <- seq_len(NROW(Omega))
  }

  match_Omega <- Omega[M_idx]
  cpar_fallback <- start_params$continuous_parametric
  if (!is.null(previous_params$continuous_parametric)) {
    cpar_fallback <- previous_params$continuous_parametric
  }

  b_params <- NULL
  if (length(b_vars) > 0L) {
    b_params <- data.table(
      variable = b_vars,
      theta = binary_formula(match_Omega[, b_vars, with = FALSE])
    )
  }

  cpar_params <- NULL
  if (length(cpar_vars) > 0L) {
    cpar_params <- estimate_hurdle_gamma_params(
      df = match_Omega[, cpar_vars, with = FALSE],
      variables = cpar_vars,
      side = "M",
      fallback_params = cpar_fallback,
      controls_nleqslv = controls_nleqslv,
      context = context
    )
  }

  list(binary = b_params, continuous_parametric = cpar_params)
}

#' @noRd
estimate_inverted_nonmatch_parameters <- function(Omega,
                                                  U_idx,
                                                  b_vars,
                                                  cpar_vars,
                                                  previous_params,
                                                  controls_nleqslv,
                                                  context = "mec_blocking()") {
  if (length(U_idx) == 0L) {
    stop(sprintf("%s cannot estimate nonmatch parameters from an empty candidate-pair complement.", context),
         call. = FALSE)
  }

  b_params <- NULL
  if (length(b_vars) > 0L) {
    b_params <- data.table(
      variable = b_vars,
      eta = binary_formula(Omega[U_idx, b_vars, with = FALSE])
    )
  }

  cpar_params <- NULL
  if (length(cpar_vars) > 0L) {
    cpar_params <- estimate_hurdle_gamma_params(
      df = Omega[U_idx, cpar_vars, with = FALSE],
      variables = cpar_vars,
      side = "U",
      fallback_params = previous_params$continuous_parametric,
      controls_nleqslv = controls_nleqslv,
      context = context
    )
  }

  list(binary = b_params, continuous_parametric = cpar_params)
}

#' @noRd
combine_inverted_parameters <- function(match_params, nonmatch_params, b_vars, cpar_vars) {
  b_params <- NULL
  if (length(b_vars) > 0L) {
    b_params <- data.table(
      variable = b_vars,
      theta = match_params$binary$theta,
      eta = nonmatch_params$binary$eta
    )
  }

  cpar_params <- NULL
  if (length(cpar_vars) > 0L) {
    cpar_params <- data.table::copy(match_params$continuous_parametric)
    data.table::set(cpar_params, j = "p_0_U", value = nonmatch_params$continuous_parametric$p_0_U)
    data.table::set(cpar_params, j = "alpha_U", value = nonmatch_params$continuous_parametric$alpha_U)
    data.table::set(cpar_params, j = "beta_U", value = nonmatch_params$continuous_parametric$beta_U)
  }

  list(binary = b_params, continuous_parametric = cpar_params)
}

#' @noRd
inverted_nonmatch_param_vector <- function(nonmatch_params) {
  params <- numeric()

  if (!is.null(nonmatch_params$binary)) {
    params <- c(params, nonmatch_params$binary$eta)
  }

  if (!is.null(nonmatch_params$continuous_parametric)) {
    params <- c(
      params,
      nonmatch_params$continuous_parametric$p_0_U,
      nonmatch_params$continuous_parametric$alpha_U,
      nonmatch_params$continuous_parametric$beta_U
    )
  }

  params
}

#' @noRd
inverted_match_param_vector <- function(match_params) {
  params <- numeric()

  if (!is.null(match_params$binary)) {
    params <- c(params, match_params$binary$theta)
  }

  if (!is.null(match_params$continuous_parametric)) {
    params <- c(
      params,
      match_params$continuous_parametric$p_0_M,
      match_params$continuous_parametric$alpha_M,
      match_params$continuous_parametric$beta_M
    )
  }

  params
}

#' @noRd
inverted_parameter_vector <- function(match_params, nonmatch_params) {
  c(
    inverted_match_param_vector(match_params),
    inverted_nonmatch_param_vector(nonmatch_params)
  )
}

#' @noRd
score_inverted_mec_ratio <- function(Omega, b_vars, cpar_vars, b_params, cpar_params) {
  data.table::set(Omega, j = "ratio", value = 1)

  if (length(b_vars) > 0L) {
    Omega_b <- Omega[, b_vars, with = FALSE]
    data.table::set(
      Omega,
      j = "ratio",
      value = Omega[["ratio"]] * bernoulli_ratio(Omega_b, b_params$eta, b_params$theta)
    )
  }

  if (length(cpar_vars) > 0L) {
    Omega_cpar <- Omega[, cpar_vars, with = FALSE]
    data.table::set(
      Omega,
      j = "ratio",
      value = Omega[["ratio"]] * hurdle_gamma_ratio(
        Omega_cpar,
        cpar_params$p_0_U,
        cpar_params$alpha_U,
        cpar_params$beta_U,
        cpar_params$p_0_M,
        cpar_params$alpha_M,
        cpar_params$beta_M
      )
    )
  }

  ratio <- Omega[["ratio"]]
  ratio[is.na(ratio) | ratio < 0] <- Inf
  data.table::set(Omega, j = "ratio", value = ratio)
  Omega
}

#' @noRd
blocking_disagreement_norm <- function(Omega, b_vars, cpar_vars) {
  components <- list()

  if (length(b_vars) > 0L) {
    components[["binary"]] <- 1 - as.matrix(Omega[, b_vars, with = FALSE])
  }

  if (length(cpar_vars) > 0L) {
    components[["continuous_parametric"]] <- as.matrix(Omega[, cpar_vars, with = FALSE])
  }

  disagreement <- do.call(cbind, components)
  sqrt(rowSums(disagreement ^ 2))
}

#' @noRd
select_inverted_mec_indices <- function(a, b, block, ratio, n_M) {
  n_target <- round(n_M)

  if (n_target <= 0L || length(ratio) == 0L) {
    return(integer())
  }

  ratio_order <- ratio
  ratio_order[is.na(ratio_order) | ratio_order < 0] <- Inf
  order_idx <- order(ratio_order, a, b, block)
  selected_idx <- integer(length(order_idx))
  n_selected <- 0L
  used_a <- rep(FALSE, max(a))
  used_b <- rep(FALSE, max(b))

  for (idx in order_idx) {
    current_a <- a[idx]
    current_b <- b[idx]

    if (!used_a[current_a] && !used_b[current_b]) {
      n_selected <- n_selected + 1L
      selected_idx[n_selected] <- idx
      used_a[current_a] <- TRUE
      used_b[current_b] <- TRUE
    }

    if (n_selected >= n_target) {
      break
    }
  }

  if (n_selected == 0L) {
    return(integer())
  }

  selected_idx[seq_len(n_selected)]
}

#' @noRd
estimate_inverted_q <- function(ratio, n_U_current, N) {
  denominator <- n_U_current * (ratio - 1) + N
  q_est <- n_U_current * ratio / denominator
  q_est[is.infinite(ratio) & ratio > 0] <- 1
  q_est[ratio == 0 & denominator > 0] <- 0
  q_est <- pmin(q_est, 1)
  q_est <- pmax(q_est, 0)
  q_est[!is.finite(q_est)] <- 1
  q_est
}

#' @noRd
has_valid_u_cpar_fallback <- function(previous_params, cpar_vars) {
  if (length(cpar_vars) == 0L) {
    return(logical())
  }

  if (is.null(previous_params)) {
    return(stats::setNames(rep(FALSE, length(cpar_vars)), cpar_vars))
  }

  cpar_params <- previous_params$continuous_parametric
  if (is.null(cpar_params) ||
      !all(c("variable", "alpha_U", "beta_U") %in% names(cpar_params))) {
    return(stats::setNames(rep(FALSE, length(cpar_vars)), cpar_vars))
  }

  cpar_params <- align_parameter_table(cpar_params, cpar_vars)
  valid <- is.finite(cpar_params[["alpha_U"]]) &
    is.finite(cpar_params[["beta_U"]]) &
    cpar_params[["alpha_U"]] > 0 &
    cpar_params[["beta_U"]] > 0
  stats::setNames(valid, cpar_vars)
}

#' @noRd
has_sufficient_u_fit_sample <- function(Omega,
                                        U_fit_idx,
                                        cpar_vars,
                                        previous_params) {
  if (length(U_fit_idx) == 0L) {
    return(FALSE)
  }

  if (length(cpar_vars) == 0L) {
    return(TRUE)
  }

  valid_fallback <- has_valid_u_cpar_fallback(previous_params, cpar_vars)
  for (variable in cpar_vars) {
    gamma <- Omega[[variable]][U_fit_idx]
    n_positive <- sum(is.finite(gamma) & gamma > 0)
    if (n_positive < 2L && !isTRUE(valid_fallback[[variable]])) {
      return(FALSE)
    }
  }

  TRUE
}

#' @noRd
make_u_fit_diagnostic <- function(iter,
                                  n_U_current,
                                  n_U_base,
                                  alpha,
                                  requested_n_drop,
                                  requested_n_keep,
                                  actual_n_drop,
                                  n_U_fit,
                                  alpha_applied,
                                  reason) {
  data.table(
    iter = as.integer(iter),
    n_U_current = as.integer(n_U_current),
    n_U_base = as.integer(n_U_base),
    alpha = as.numeric(alpha),
    requested_n_drop = as.integer(requested_n_drop),
    requested_n_keep = as.integer(requested_n_keep),
    actual_n_drop = as.integer(actual_n_drop),
    actual_drop_fraction = if (n_U_base == 0L) NA_real_ else actual_n_drop / n_U_base,
    n_U_fit = as.integer(n_U_fit),
    alpha_applied = as.logical(alpha_applied),
    reason = reason
  )
}

#' @noRd
rank_u_fit_candidates <- function(Omega, U_fit_idx) {
  score_name <- if ("q_est" %in% names(Omega)) "q_est" else "ratio"
  score <- Omega[[score_name]][U_fit_idx]
  if (all(is.na(score)) && "ratio" %in% names(Omega)) {
    score <- Omega[["ratio"]][U_fit_idx]
  }
  score[is.na(score)] <- -Inf
  U_fit_idx[order(-score, Omega[["a"]][U_fit_idx], Omega[["b"]][U_fit_idx], Omega[["block"]][U_fit_idx])]
}

#' @noRd
select_inverted_u_fit_indices <- function(Omega,
                                          U_idx,
                                          U_base_idx,
                                          iter,
                                          alpha,
                                          cpar_vars,
                                          previous_params) {
  n_U_current <- length(U_idx)
  n_U_base <- length(U_base_idx)
  requested_n_drop <- floor(alpha * n_U_current)
  requested_n_keep <- n_U_current - requested_n_drop

  if (iter == 1L) {
    return(list(
      U_fit_idx = U_base_idx,
      diagnostic = make_u_fit_diagnostic(
        iter = iter,
        n_U_current = n_U_current,
        n_U_base = n_U_base,
        alpha = alpha,
        requested_n_drop = 0L,
        requested_n_keep = n_U_current,
        actual_n_drop = 0L,
        n_U_fit = n_U_base,
        alpha_applied = FALSE,
        reason = "first_u_fit_full"
      )
    ))
  }

  if (alpha == 0 || requested_n_drop == 0L) {
    reason <- if (alpha == 0) "alpha_zero" else "requested_drop_zero"
    return(list(
      U_fit_idx = U_base_idx,
      diagnostic = make_u_fit_diagnostic(
        iter = iter,
        n_U_current = n_U_current,
        n_U_base = n_U_base,
        alpha = alpha,
        requested_n_drop = requested_n_drop,
        requested_n_keep = requested_n_keep,
        actual_n_drop = 0L,
        n_U_fit = n_U_base,
        alpha_applied = FALSE,
        reason = reason
      )
    ))
  }

  if (n_U_base <= requested_n_keep) {
    return(list(
      U_fit_idx = U_base_idx,
      diagnostic = make_u_fit_diagnostic(
        iter = iter,
        n_U_current = n_U_current,
        n_U_base = n_U_base,
        alpha = alpha,
        requested_n_drop = requested_n_drop,
        requested_n_keep = requested_n_keep,
        actual_n_drop = 0L,
        n_U_fit = n_U_base,
        alpha_applied = FALSE,
        reason = "base_smaller_than_requested_keep"
      )
    ))
  }

  ranked_idx <- rank_u_fit_candidates(Omega, U_base_idx)
  retained_idx <- ranked_idx[seq_len(requested_n_keep)]
  if (!has_sufficient_u_fit_sample(
    Omega = Omega,
    U_fit_idx = retained_idx,
    cpar_vars = cpar_vars,
    previous_params = previous_params
  )) {
    return(list(
      U_fit_idx = U_base_idx,
      diagnostic = make_u_fit_diagnostic(
        iter = iter,
        n_U_current = n_U_current,
        n_U_base = n_U_base,
        alpha = alpha,
        requested_n_drop = requested_n_drop,
        requested_n_keep = requested_n_keep,
        actual_n_drop = 0L,
        n_U_fit = n_U_base,
        alpha_applied = FALSE,
        reason = "minimum_sample_full_base"
      )
    ))
  }

  actual_n_drop <- n_U_base - length(retained_idx)
  list(
    U_fit_idx = retained_idx,
    diagnostic = make_u_fit_diagnostic(
      iter = iter,
      n_U_current = n_U_current,
      n_U_base = n_U_base,
      alpha = alpha,
      requested_n_drop = requested_n_drop,
      requested_n_keep = requested_n_keep,
      actual_n_drop = actual_n_drop,
      n_U_fit = length(retained_idx),
      alpha_applied = actual_n_drop > 0L,
      reason = "alpha_reliability_drop"
    )
  )
}

#' @noRd
fit_mec_blocking_inverted_omega <- function(A,
                                            B,
                                            variables,
                                            comparators,
                                            methods,
                                            Omega,
                                            start_params = NULL,
                                            delta = 0.5,
                                            eps = 0.05,
                                            controls_nleqslv = list(),
                                            n_U_min,
                                            nu,
                                            alpha = 0,
                                            context = "mec_blocking()") {
  method_variables <- extract_method_variables(methods, include_hit_miss = FALSE)
  b_vars <- method_variables$b_vars
  cpar_vars <- method_variables$cpar_vars
  N <- NROW(Omega)
  all_idx <- seq_len(N)
  start_params <- prepare_inverted_start_params(start_params, cpar_vars)
  max_iter_inverted <- 1000L

  init_disagreement <- blocking_disagreement_norm(Omega, b_vars, cpar_vars)
  data.table::set(Omega, j = "init_disagreement", value = init_disagreement)
  n_M_init_target <- initial_inverted_match_count(nu)
  M_idx <- select_inverted_mec_indices(
    a = Omega[["a"]],
    b = Omega[["b"]],
    block = Omega[["block"]],
    ratio = Omega[["init_disagreement"]],
    n_M = n_M_init_target
  )
  U_idx <- setdiff(all_idx, M_idx)
  n_M_init <- length(M_idx)
  n_U_init <- length(U_idx)

  if (length(U_idx) == 0L) {
    if (N != nu) {
      stop(sprintf("%s initialized an empty nonmatch complement before reaching the structural one-to-one bound.",
                   context),
           call. = FALSE)
    }

    match_params <- estimate_inverted_match_parameters(
      Omega = Omega,
      M_idx = M_idx,
      b_vars = b_vars,
      cpar_vars = cpar_vars,
      previous_params = list(),
      start_params = start_params,
      controls_nleqslv = controls_nleqslv,
      context = context
    )
    data.table::set(Omega, j = "ratio", value = 0)
    data.table::set(Omega, j = "q_est", value = 0)
    M_est <- Omega[M_idx, c("a", "b", "block", "ratio"), with = FALSE]
    u_fit_diagnostics <- make_u_fit_diagnostic(
      iter = 0L,
      n_U_current = 0L,
      n_U_base = 0L,
      alpha = alpha,
      requested_n_drop = 0L,
      requested_n_keep = 0L,
      actual_n_drop = 0L,
      n_U_fit = 0L,
      alpha_applied = FALSE,
      reason = "structural_no_nonmatch_complement"
    )
    model <- list(
      b_vars = if (length(b_vars) == 0L) NULL else b_vars,
      cpar_vars = if (length(cpar_vars) == 0L) NULL else cpar_vars,
      b_params = match_params$binary,
      cpar_params = match_params$continuous_parametric,
      nonmatch_params = NULL,
      variables = variables,
      comparators = comparators,
      methods = methods,
      n_M_est = NROW(M_est),
      n_U_est = 0L,
      n_U_min = n_U_min,
      nu = nu,
      n_M_init = n_M_init,
      n_U_init = n_U_init,
      candidate_pair_count = N,
      prob_est = if (N == 0L) NA_real_ else NROW(M_est) / N,
      ratio_orientation = "u_over_m",
      alpha = alpha,
      n_U_fit = 0L,
      u_fit_diagnostics = u_fit_diagnostics,
      iter = 0L,
      convergence_reason = "structural_no_nonmatch_complement"
    )

    return(list(
      model = model,
      Omega = Omega,
      M_est = M_est,
      n_M_est = NROW(M_est),
      n_U_est = 0L,
      n_U_min = n_U_min,
      nu = nu,
      n_M_init = n_M_init,
      n_U_init = n_U_init,
      candidate_pair_count = N,
      alpha = alpha,
      n_U_fit = 0L,
      u_fit_diagnostics = u_fit_diagnostics,
      iter = 0L,
      convergence_reason = "structural_no_nonmatch_complement"
    ))
  }

  iter <- 1L
  n_U_old <- length(U_idx)
  previous_match_params <- NULL
  previous_nonmatch_params <- NULL
  previous_param_vector <- NULL
  convergence_reason <- "max_iter"
  n_U_fit <- length(U_idx)
  u_fit_diagnostics <- data.table()

  repeat {
    match_params <- estimate_inverted_match_parameters(
      Omega = Omega,
      M_idx = M_idx,
      b_vars = b_vars,
      cpar_vars = cpar_vars,
      previous_params = if (is.null(previous_match_params)) list() else previous_match_params,
      start_params = start_params,
      controls_nleqslv = controls_nleqslv,
      context = context
    )
    u_fit_selection <- select_inverted_u_fit_indices(
      Omega = Omega,
      U_idx = U_idx,
      U_base_idx = U_idx,
      iter = iter,
      alpha = alpha,
      cpar_vars = cpar_vars,
      previous_params = previous_nonmatch_params
    )
    U_fit_idx <- u_fit_selection$U_fit_idx
    u_fit_diagnostics <- rbindlist(
      list(u_fit_diagnostics, u_fit_selection$diagnostic),
      use.names = TRUE
    )
    n_U_fit <- length(U_fit_idx)
    nonmatch_params <- estimate_inverted_nonmatch_parameters(
      Omega = Omega,
      U_idx = U_fit_idx,
      b_vars = b_vars,
      cpar_vars = cpar_vars,
      previous_params = previous_nonmatch_params,
      controls_nleqslv = controls_nleqslv,
      context = context
    )
    combined_params <- combine_inverted_parameters(
      match_params = match_params,
      nonmatch_params = nonmatch_params,
      b_vars = b_vars,
      cpar_vars = cpar_vars
    )
    Omega <- score_inverted_mec_ratio(
      Omega = Omega,
      b_vars = b_vars,
      cpar_vars = cpar_vars,
      b_params = combined_params$binary,
      cpar_params = combined_params$continuous_parametric
    )

    q_est <- estimate_inverted_q(
      ratio = Omega[["ratio"]],
      n_U_current = length(U_idx),
      N = N
    )
    data.table::set(Omega, j = "q_est", value = q_est)
    n_U_raw <- sum(q_est)
    n_U_est <- max(n_U_min, min(N, round(n_U_raw)))
    n_M_est <- N - n_U_est
    M_idx_new <- select_inverted_mec_indices(
      a = Omega[["a"]],
      b = Omega[["b"]],
      block = Omega[["block"]],
      ratio = Omega[["ratio"]],
      n_M = n_M_est
    )
    U_idx_new <- setdiff(all_idx, M_idx_new)
    param_vector <- inverted_parameter_vector(match_params, nonmatch_params)
    can_check_convergence <- iter >= 2L

    if (can_check_convergence && abs(n_U_est - n_U_old) < delta) {
      convergence_reason <- "n_U_delta"
    } else if (can_check_convergence && identical(sort(M_idx_new), sort(M_idx))) {
      convergence_reason <- "match_set_unchanged"
    } else if (can_check_convergence &&
               !is.null(previous_param_vector) &&
               norm(previous_param_vector - param_vector, type = "2") < eps) {
      convergence_reason <- "nonmatch_parameter_eps"
    } else if (iter >= max_iter_inverted) {
      warning("mec_blocking() reached the maximum number of inverted MEC iterations.")
      convergence_reason <- "max_iter"
    } else if (length(U_idx_new) == 0L) {
      convergence_reason <- "structural_no_nonmatch_complement"
    } else if (length(M_idx_new) == 0L) {
      convergence_reason <- "empty_match_set"
    } else {
      previous_match_params <- match_params
      previous_nonmatch_params <- nonmatch_params
      previous_param_vector <- param_vector
      n_U_old <- n_U_est
      M_idx <- M_idx_new
      U_idx <- U_idx_new
      iter <- iter + 1L
      next
    }

    M_idx <- M_idx_new
    U_idx <- U_idx_new
    previous_match_params <- match_params
    previous_nonmatch_params <- nonmatch_params
    previous_param_vector <- param_vector
    break
  }

  M_est <- Omega[M_idx, c("a", "b", "block", "ratio"), with = FALSE]
  n_M_selected <- NROW(M_est)
  n_U_selected <- N - n_M_selected
  model <- list(
    b_vars = if (length(b_vars) == 0L) NULL else b_vars,
    cpar_vars = if (length(cpar_vars) == 0L) NULL else cpar_vars,
    b_params = combined_params$binary,
    cpar_params = combined_params$continuous_parametric,
    nonmatch_params = previous_nonmatch_params,
    variables = variables,
    comparators = comparators,
    methods = methods,
    n_M_est = n_M_selected,
    n_U_est = n_U_selected,
    n_U_min = n_U_min,
    nu = nu,
    n_M_init = n_M_init,
    n_U_init = n_U_init,
    candidate_pair_count = N,
    prob_est = n_M_selected / N,
    ratio_orientation = "u_over_m",
    alpha = alpha,
    n_U_fit = n_U_fit,
    u_fit_diagnostics = u_fit_diagnostics,
    iter = iter,
    convergence_reason = convergence_reason
  )

  list(
    model = model,
    Omega = Omega,
    M_est = M_est,
    n_M_est = n_M_selected,
    n_U_est = n_U_selected,
    n_U_min = n_U_min,
    nu = nu,
    n_M_init = n_M_init,
    n_U_init = n_U_init,
    candidate_pair_count = N,
    alpha = alpha,
    n_U_fit = n_U_fit,
    u_fit_diagnostics = u_fit_diagnostics,
    iter = iter,
    convergence_reason = convergence_reason
  )
}

#' @import data.table
#' @importFrom nleqslv nleqslv
#' @importFrom purrr partial
#' @importFrom densityratio kliep
#'
#' @title Train a Record Linkage Model
#'
#' @author Adam Struzik
#'
#' @description Trains a supervised record linkage model using probability or density ratio estimation,
#' based on [Lee et al. (2022)](https://www150.statcan.gc.ca/n1/pub/12-001-x/2022001/article/00007-eng.htm),
#' with several extensions.
#'
#' @param A A duplicate-free `data.frame` or `data.table`.
#' @param B A duplicate-free `data.frame` or `data.table`.
#' @param matches A `data.frame` or `data.table` indicating known matches.
#' @param variables A character vector of key variables used to create comparison vectors.
#' @param comparators A named list of functions for comparing pairs of records.
#' @param methods A named list of methods used for estimation (`"binary"`, `"continuous_parametric"` or `"continuous_nonparametric"`).
#' @param prob_ratio Probability/density ratio type (`"1"` or `"2"`).
#' @param nonpar_hurdle Logical indicating whether to use a hurdle model or not
#' (used only if the `"continuous_nonparametric"` method has been chosen for at least one variable).
#' @param controls_nleqslv Controls passed to the \link[nleqslv:nleqslv]{nleqslv()} function (only if the `"continuous_parametric"` method has been chosen for at least one variable).
#' @param controls_kliep Controls passed to the \link[densityratio:kliep]{kliep()} function (only if the `"continuous_nonparametric"` method has been chosen for at least one variable).
#' @param verbose Logical indicating whether to print progress messages.
#'
#' @details
#' Consider two datasets: \eqn{A} and \eqn{B}.
#' Let the bipartite comparison space \eqn{\Omega = A \times B} consist of
#' matches \eqn{M} and non-matches \eqn{U} between the records in files
#' \eqn{A} and \eqn{B}. For any pair of records \eqn{(a,b) \in \Omega},
#' let \eqn{\pmb{\gamma}_{ab} = (\gamma_{ab}^1,\gamma_{ab}^2,
#' \ldots,\gamma_{ab}^K)'} be the comparison vector between
#' a set of key variables. The original MEC algorithm uses the binary
#' comparison function to evaluate record pairs across two datasets.
#' However, this approach may be insufficient when handling datasets
#' with frequent errors across multiple variables.
#'
#' We propose the use of continuous comparison functions to address
#' the limitations of binary comparison methods. We consider every
#' semi-metric, i.e., a function \eqn{d: A \times B \to \mathbb{R}},
#' satisfying the following conditions:\cr
#' \enumerate{
#' \item{\eqn{d(x,y) \geq 0},}
#' \item{\eqn{d(x,y) = 0} if and only if \eqn{x = y},}
#' \item{\eqn{d(x,y) = d(y,x)}.}
#' }
#' For example, we can use the Jaro-Winkler distance for character variables
#' (which is implemented in the `automatedRecLin` package as [jarowinkler_complement()])
#' or the Euclidean distance for numerical variables. The `automatedRecLin` package allows the use of
#' a different comparison function for each key variable (which should be specified
#' as a list in the `comparators` argument). The default function
#' for each key variable is \link[reclin2:cmp_identical]{cmp_identical()}
#' (the binary comparison function).
#'
#' The [train_rec_lin()] function is used to train a record linkage model,
#' when \eqn{M} and \eqn{U} are known (which might later serve as a classifier
#' for pairs outside \eqn{\Omega}). It offers different approaches to estimate the
#' probability/density ratio between matches and non-matches, which should be
#' specified as a list in the `methods` argument. The method suitable for the binary
#' comparison function is `"binary"`, which is also the default method for each
#' variable.
#'
#' For the continuous semi-metrics we suggest the usage
#' of `"continuous_parametric"` or `"continuous_nonparametric"`
#' method. The `"continuous_parametric"` method assumes that
#' \eqn{\gamma_{ab}^k|M} and \eqn{\gamma_{ab}^k|U} follow
#' hurdle Gamma distributions. The density function of a hurdle
#' Gamma distribution is characterized by three parameters
#' \eqn{p_0 \in (0,1)} and \eqn{\alpha, \beta > 0} as follows:
#' \deqn{
#' f(x;p_0,\alpha,\beta) = p_0^{\mathbb{I}(x = 0)}[(1 - p_0) v(x;\alpha,\beta)]^{\mathbb{I}(x > 0)},
#' }
#' where
#' \deqn{
#' v(x;\alpha,\beta) = \frac{\beta^{\alpha} x^{\alpha - 1} \exp(-\beta x)}
#' {\Gamma(\alpha)}
#' }
#' is the density function of a Gamma distribution
#' (for details see [Vo et al. (2023)](https://ideas.repec.org/a/eee/csdana/v179y2023ics0167947322002365.html)).
#' The `"continuous_nonparametric"` method does not assume anything about
#' the distributions of the comparison vectors. It directly
#' estimates the density ratio between the matches and the non-matches, using
#' the Kullback-Leibler Importance Estimation Procedure (KLIEP).
#' For details see [Sugiyama et al. (2008)](https://link.springer.com/article/10.1007/s10463-008-0197-x).
#'
#' @return
#' Returns a list containing:\cr
#' \itemize{
#' \item{`b_vars` -- a character vector of variables used for the `"binary"` method (with the prefix `"gamma_"`),}
#' \item{`cpar_vars` -- a character vector of variables used for the `"continuous_parametric"` method (with the prefix `"gamma_"`),}
#' \item{`cnonpar_vars` -- a character vector of variables used for the `"continuous_nonparametric"` method (with the prefix `"gamma_"`),}
#' \item{`b_params` -- parameters estimated using the `"binary"` method,}
#' \item{`cpar_params` -- parameters estimated using the `"continuous_parametric"` method,}
#' \item{`cnonpar_params` -- probability of exact matching estimated using the `"continuous_nonparametric"` method,}
#' \item{`ratio_kliep` -- a result of the \link[densityratio:kliep]{kliep()} function,}
#' \item{`ratio_kliep_list` -- an object containing the results of the \link[densityratio:kliep]{kliep()} function,}
#' \item{`ml_model` -- here `NULL`,}
#' \item{`pi_est` -- a prior probability of matching,}
#' \item{`match_prop` -- proportion of matches in the smaller dataset,}
#' \item{`variables` -- a character vector of key variables used for comparison,}
#' \item{`comparators` -- a list of functions used to compare pairs of records,}
#' \item{`methods` -- a list of methods used for estimation,}
#' \item{`"prob_ratio"` -- probability/density ratio type.}
#' }
#'
#' @references
#' Lee, D., Zhang, L.-C. and Kim, J. K. (2022). Maximum entropy classification for record linkage.
#' Survey Methodology, Statistics Canada, Catalogue No. 12-001-X, Vol. 48, No. 1.
#'
#' Vo, T. H., Chauvet, G., Happe, A., Oger, E., Paquelet, S., and Garès, V. (2023).
#' Extending the Fellegi-Sunter record linkage model for mixed-type data with application to the French national health data system.
#' Computational Statistics & Data Analysis, 179, 107656.
#'
#' Sugiyama, M., Suzuki, T., Nakajima, S. et al. Direct importance estimation for covariate shift adaptation.
#' Ann Inst Stat Math 60, 699–746 (2008). \doi{10.1007/s10463-008-0197-x}
#'
#' @examples
#' df_1 <- data.frame(
#'   "name" = c("James", "Emma", "William", "Olivia", "Thomas",
#'   "Sophie", "Harry", "Amelia", "George", "Isabella"),
#'   "surname" = c("Smith", "Johnson", "Brown", "Taylor", "Wilson",
#'   "Davis", "Clark", "Harris", "Lewis", "Walker")
#' )
#'  df_2 <- data.frame(
#'   "name" = c("James", "Ema", "Wimliam", "Olivia", "Charlotte",
#'   "Henry", "Lucy", "Edward", "Alice", "Jack"),
#'   "surname" = c("Smith", "Johnson", "Bron", "Tailor", "Moore",
#'   "Evans", "Hall", "Wright", "Green", "King")
#' )
#' comparators <- list("name" = jarowinkler_complement(),
#'                     "surname" = jarowinkler_complement())
#' matches <- data.frame("a" = 1:4, "b" = 1:4)
#' methods <- list("name" = "continuous_nonparametric",
#'                 "surname" = "continuous_nonparametric")
#' model <- train_rec_lin(A = df_1, B = df_2, matches = matches,
#'                        variables = c("name", "surname"),
#'                        comparators = comparators,
#'                        methods = methods)
#' model
#' @export
train_rec_lin <- function(
    A,
    B,
    matches,
    variables,
    comparators = NULL,
    methods = NULL,
    prob_ratio = NULL,
    nonpar_hurdle = TRUE,
    controls_nleqslv = list(),
    controls_kliep = control_kliep(),
    verbose = FALSE) {

  data.table::setDT(A)
  data.table::setDT(B)
  validate_match_pairs(matches, nrow(A), nrow(B))

  stopifnot("There are no records with perfect agreement on the key variables. Please provide relevant datasets." =
              has_perfect_agreement(A, B, variables))

  if (is.null(prob_ratio)) {
    prob_ratio <- "1"
  }
  validate_choice(prob_ratio, c("1", "2"), "prob_ratio")

  methods <- validate_methods(
    methods = methods,
    variables = variables,
    allowed_methods = c("binary", "continuous_parametric", "continuous_nonparametric")
  )
  if (verbose) {
    message(sprintf(
      "Preparing supervised record linkage model for %d records in A, %d records in B, and %d key variable(s).",
      nrow(A),
      nrow(B),
      length(variables)
    ))
  }
  preprocessing <- drop_constant_key_variables(
    A = A,
    B = B,
    variables = variables,
    comparators = comparators,
    methods = methods
  )
  A <- preprocessing$A
  B <- preprocessing$B
  variables <- preprocessing$variables
  comparators <- preprocessing$comparators
  methods <- preprocessing$methods
  constant_vars <- preprocessing$constant_vars

  for (var in constant_vars) {
    warning(paste("The variable", var, "has only one unique value and has been removed."))
  }

  vectors <- comparison_vectors(A = A,
                                B = B,
                                variables = variables,
                                comparators = comparators,
                                matches = matches)
  Omega <- vectors$Omega
  comparators <- vectors$comparators

  n <- NROW(Omega)
  M_idx <- which(Omega[["match"]] == 1)
  U_idx <- which(Omega[["match"]] == 0)
  n_M <- length(M_idx)
  pi_est <- n_M / n
  if (verbose) {
    message(sprintf(
      "Created %d comparison vector(s), including %d known match(es) and %d known nonmatch(es).",
      n,
      n_M,
      n - n_M
    ))
  }

  if (prob_ratio == "2") {
    # Reuse the computed comparison vectors instead of building another full Cartesian product.
    common_values <- vapply(variables, function(var) {
      gamma_name <- paste0("gamma_", var)

      if (methods[[var]] == "binary") {
        any(Omega[[gamma_name]][U_idx] == 1)
      } else {
        any(Omega[[gamma_name]][U_idx] == 0)
      }
    }, logical(1))

    if (any(!common_values)) {
      prob_ratio <- "1"
      warning("Some variables lack common values between the unmatches. Switching the probability ratio to \"1\".")
    }
  }

  method_variables <- extract_method_variables(methods)
  b_vars <- method_variables$b_vars
  cpar_vars <- method_variables$cpar_vars
  cnonpar_vars <- method_variables$cnonpar_vars

  b_params <- NULL
  cpar_params <- NULL
  cnonpar_params <- NULL
  ratio_kliep <- NULL
  ratio_kliep_list <- NULL
  kliep_warning_emitted <- FALSE

  warn_kliep_once <- function(variables, message) {
    if (!kliep_warning_emitted) {
      warn_kliep_issue("train_rec_lin()", variables, message)
      kliep_warning_emitted <<- TRUE
    }
  }

  if (any(methods == "binary")) {

    b_vars <- paste0("gamma_", names(which(methods == "binary")))
    Omega_b <- Omega[, b_vars, with = FALSE]
    M_b <- Omega_b[M_idx]
    theta_b <- binary_formula(M_b)
    if (prob_ratio == "1") {
      eta_b <- binary_formula(Omega_b)
    } else if (prob_ratio == "2") {
      U_b <- Omega_b[U_idx]
      eta_b <- binary_formula(U_b)
    }

    b_params <- data.table(
      variable = b_vars,
      theta = theta_b,
      eta = eta_b
    )

  }

  if (any(methods == "continuous_parametric")) {

    cpar_vars <- paste0("gamma_", names(which(methods == "continuous_parametric")))
    modified_nleqslv <- purrr::partial(nleqslv::nleqslv, control = controls_nleqslv)
    Omega_cpar <- Omega[, cpar_vars, with = FALSE]
    M_cpar <- Omega_cpar[M_idx]
    p_0_M <- p_0_formula(M_cpar)
    gamma_plus_M <- gamma_plus_formula(M_cpar)
    alpha_M <- alpha_formula(M_cpar, modified_nleqslv)
    beta_M <- alpha_M / gamma_plus_M
    if (prob_ratio == "1") {
      p_0_Omega <- p_0_formula(Omega_cpar)
      gamma_plus_Omega <- gamma_plus_formula(Omega_cpar)
      alpha_Omega <- alpha_formula(Omega_cpar, modified_nleqslv)
      beta_Omega <- alpha_Omega / gamma_plus_Omega

      cpar_params <- data.table(
        variable = cpar_vars,
        p_0_M = p_0_M,
        p_0_Omega = p_0_Omega,
        alpha_M = alpha_M,
        alpha_Omega = alpha_Omega,
        beta_M = beta_M,
        beta_Omega = beta_Omega
      )
    } else if (prob_ratio == "2") {
      U_cpar <- Omega_cpar[U_idx]
      p_0_U <- p_0_formula(U_cpar)
      gamma_plus_U <- gamma_plus_formula(U_cpar)
      alpha_U <- alpha_formula(U_cpar, modified_nleqslv)
      beta_U <- alpha_U / gamma_plus_U

      cpar_params <- data.table(
        variable = cpar_vars,
        p_0_M = p_0_M,
        p_0_U = p_0_U,
        alpha_M = alpha_M,
        alpha_U = alpha_U,
        beta_M = beta_M,
        beta_U = beta_U
      )
    }

  }

  if (any(methods == "continuous_nonparametric")) {

    cnonpar_vars <- paste0("gamma_", names(which(methods == "continuous_nonparametric")))
    Omega_cnonpar <- Omega[, cnonpar_vars, with = FALSE]
    M_cnonpar <- Omega_cnonpar[M_idx]
    if (prob_ratio == "1") {

      if (nonpar_hurdle) {

        p_0_M_cnonpar <- p_0_formula(M_cnonpar)
        p_0_U_cnonpar <- p_0_formula(Omega_cnonpar)

        ratio_kliep_list <- fit_kliep_hurdle_models(
          df_numerator = M_cnonpar,
          df_denominator = Omega_cnonpar,
          variables = cnonpar_vars,
          controls_kliep = controls_kliep
        )
        missing_models <- missing_kliep_models(ratio_kliep_list)

        if (length(missing_models) > 0L) {
          warn_kliep_once(missing_models, "insufficient positive comparisons; using only the hurdle mass term for those variables.")
        }

        cnonpar_params <- data.table(
          variable = cnonpar_vars,
          p_0_M_cnonpar = p_0_M_cnonpar,
          p_0_U_cnonpar = p_0_U_cnonpar
        )

      } else {

        ratio_kliep <- do.call(
          densityratio::kliep,
          c(list(
            df_numerator = M_cnonpar,
            df_denominator = Omega_cnonpar
          ),
          controls_kliep)
        )

      }

    } else if (prob_ratio == "2") {
      U_cnonpar <- Omega_cnonpar[U_idx]

      if (nonpar_hurdle) {

        p_0_M_cnonpar <- p_0_formula(M_cnonpar)
        p_0_U_cnonpar <- p_0_formula(U_cnonpar)

        ratio_kliep_list <- fit_kliep_hurdle_models(
          df_numerator = M_cnonpar,
          df_denominator = U_cnonpar,
          variables = cnonpar_vars,
          controls_kliep = controls_kliep
        )
        missing_models <- missing_kliep_models(ratio_kliep_list)

        if (length(missing_models) > 0L) {
          warn_kliep_once(missing_models, "insufficient positive comparisons; using only the hurdle mass term for those variables.")
        }

        cnonpar_params <- data.table(
          variable = cnonpar_vars,
          p_0_M_cnonpar = p_0_M_cnonpar,
          p_0_U_cnonpar = p_0_U_cnonpar
        )

      } else {

        ratio_kliep <- do.call(
          densityratio::kliep,
          c(list(
            df_numerator = M_cnonpar,
            df_denominator = U_cnonpar
          ),
          controls_kliep)
        )

      }

    }

  }

  if (verbose) {
    method_names <- paste(unique(unlist(methods, use.names = FALSE)), collapse = ", ")
    message(sprintf("Finished estimating model components for methods: %s.", method_names))
  }

  structure(
    list(
      b_vars = if (length(b_vars) == 0) NULL else b_vars,
      cpar_vars = if (length(cpar_vars) == 0) NULL else cpar_vars,
      cnonpar_vars = if (length(cnonpar_vars) == 0) NULL else cnonpar_vars,
      b_params = if (is.null(b_params)) NULL else b_params,
      cpar_params = if (is.null(cpar_params)) NULL else cpar_params,
      cnonpar_params = if (is.null(cnonpar_params)) NULL else cnonpar_params,
      ratio_kliep = if (is.null(ratio_kliep)) NULL else ratio_kliep,
      ratio_kliep_list = if (is.null(ratio_kliep_list)) NULL else ratio_kliep_list,
      ml_model = NULL,
      pi_est = pi_est,
      match_prop = vectors$match_prop,
      variables = variables,
      comparators = comparators,
      methods = methods,
      prob_ratio = prob_ratio
    ),
    class = "rec_lin_model"
  )

}

#' @importFrom methods is
#'
#' @title Create a Custom Record Linkage Model
#'
#' @author Adam Struzik
#'
#' @description
#' Creates a supervised record linkage model using a custom machine learning (ML) classifier.
#'
#' @param ml_model A trained ML model that predicts the probability of a match based on comparison vectors.
#' @param vectors An object of class `comparison_vectors` (a result of [comparison_vectors()]), used for training the `ml_model`.
#'
#' @details
#' The [custom_rec_lin_model()] function creates a custom record linkage model,
#' based on known matches and non-matches (which might later serve as a classifier
#' for pairs outside training data). The procedure of creating a custom model
#' based on training data is as follows.
#' \enumerate{
#' \item{Use [comparison_vectors()] to compare pairs of records.}
#' \item{Train a machine learning classifier using the `Omega` element
#' of the output of [comparison_vectors()]. The classifier should
#' predict the probability of matching based on a given vector.}
#' \item{Use [custom_rec_lin_model()] with
#' appropriate arguments.}
#' }
#'
#' @return
#' Returns a list containing:\cr
#' \itemize{
#' \item{`b_vars` -- here `NULL`,}
#' \item{`cpar_vars` -- here `NULL`,}
#' \item{`cnonpar_vars` -- here `NULL`,}
#' \item{`b_params` -- here `NULL`,}
#' \item{`cpar_params` -- here `NULL`,}
#' \item{`cnonpar_params` -- here `NULL`,}
#' \item{`ratio_kliep` -- here `NULL`,}
#' \item{`ratio_kliep_list` -- here `NULL`,}
#' \item{`ml_model` -- ML model used for creating the record linkage model,}
#' \item{`pi_est` -- a prior probability of matching,}
#' \item{`match_prop` -- proportion of matches in the smaller dataset,}
#' \item{`variables` -- a character vector of key variables used for comparison,}
#' \item{`comparators` -- a list of functions used to compare pairs of records,}
#' \item{`methods` -- here `NULL`,}
#' \item{`prob_ratio` -- here `"2"`.}
#' }
#'
#' @examples
#' if (requireNamespace("xgboost", quietly = TRUE)) {
#'   df_1 <- data.frame(
#'     "name" = c("James", "Emma", "William", "Olivia", "Thomas",
#'     "Sophie", "Harry", "Amelia", "George", "Isabella"),
#'     "surname" = c("Smith", "Johnson", "Brown", "Taylor", "Wilson",
#'     "Davis", "Clark", "Harris", "Lewis", "Walker")
#'   )
#'   df_2 <- data.frame(
#'     "name" = c("James", "Ema", "Wimliam", "Olivia", "Charlotte",
#'     "Henry", "Lucy", "Edward", "Alice", "Jack"),
#'     "surname" = c("Smith", "Johnson", "Bron", "Tailor", "Moore",
#'     "Evans", "Hall", "Wright", "Green", "King")
#'   )
#'   comparators <- list("name" = jarowinkler_complement(),
#'                       "surname" = jarowinkler_complement())
#'   matches <- data.frame("a" = 1:4, "b" = 1:4)
#'   vectors <- comparison_vectors(A = df_1, B = df_2, variables = c("name", "surname"),
#'                                comparators = comparators, matches = matches)
#'   model_xgb <- xgboost::xgboost(x = as.matrix(vectors$Omega[, c("gamma_name", "gamma_surname")]),
#'                        y = factor(vectors$Omega$match),
#'                        objective = "binary:logistic", eval_metric = "logloss",
#'                        nrounds = 100, verbosity = 0, nthread = 1)
#'   custom_xgb_model <- custom_rec_lin_model(model_xgb, vectors)
#'   custom_xgb_model
#' }
#' @export
custom_rec_lin_model <- function(ml_model, vectors) {

  stopifnot("`vectors` must be of class `comparison_vectors`" =
            is(vectors, "comparison_vectors"))

  Omega <- vectors$Omega
  n <- NROW(Omega)
  n_M <- sum(Omega[["match"]] == 1)
  pi_est <- n_M / n

  variables <- vectors$variables

  structure(
    list(
      b_vars = NULL,
      cpar_vars = NULL,
      cnonpar_vars = NULL,
      b_params = NULL,
      cpar_params = NULL,
      cnonpar_params = NULL,
      ratio_kliep = NULL,
      ratio_kliep_list = NULL,
      ml_model = ml_model,
      pi_est = pi_est,
      match_prop = vectors$match_prop,
      variables = variables,
      comparators = vectors$comparators,
      methods = NULL,
      prob_ratio = "2"
    ),
    class = "rec_lin_model"
  )

}

library(automatedRecLin)
library(data.table)
library(xgboost)

df_1 <- data.frame(
  "name" = c("James", "Emma", "William", "Olivia", "Thomas",
             "Sophie", "Harry", "Amelia", "George", "Isabella"),
  "surname" = c("Smith", "Johnson", "Brown", "Taylor", "Wilson",
                "Davis", "Clark", "Harris", "Lewis", "Walker")
)
df_2 <- data.frame(
  "name" = c("James", "Ema", "Wimliam", "Olivia", "Charlotte",
             "Henry", "Lucy", "Edward", "Alice", "Jack"),
  "surname" = c("Smith", "Johnson", "Bron", "Tailor", "Moore",
                "Evans", "Hall", "Wright", "Green", "King")
)
matches <- data.frame("a" = 1:4, "b" = 1:4)
df_new_1 <- data.frame(
  "name" = c("John", "Emily", "Mark", "Anna", "David"),
  "surname" = c("Smith", "Johnson", "Taylor", "Williams", "Brown")
)
df_new_2 <- data.frame(
  "name" = c("John", "Emely", "Mark", "Michael"),
  "surname" = c("Smitth", "Johnson", "Tailor", "Henders")
)
true_matches <- data.frame("a" = 1:3, "b" = 1:3)

comparators = list("name" = jarowinkler_complement(),
                   "surname" = jarowinkler_complement())
methods_cpar <- list("name" = "continuous_parametric",
                     "surname" = "continuous_parametric")
methods_cnonpar <- list("name" = "continuous_nonparametric",
                        "surname" = "continuous_nonparametric")

confusion_cpar <- matrix(c(2, 0, 1, 17), nrow = 2, ncol = 2)
rownames(confusion_cpar) <- c("Actual Positive", "Actual Negative")
colnames(confusion_cpar) <- c("Predicted Positive", "Predicted Negative")

expect_rate_bounds <- function(result) {
  expect_true(is.finite(result$flr_est))
  expect_true(result$flr_est >= 0 && result$flr_est <= 1)
  expect_true(is.finite(result$mmr_est))
  expect_true(result$mmr_est >= 0 && result$mmr_est <= 1)
}

expect_silent(
  train_rec_lin(A = df_1, B = df_2,
                variables = c("name", "surname"),
                matches = matches)
)

expect_warning(
  train_rec_lin(A = df_1, B = df_2,
                variables = c("name", "surname"),
                matches = matches,
                prob_ratio = "2")
)

expect_equal(
  train_rec_lin(A = df_1, B = df_2,
                variables = c("name", "surname"),
                matches = matches)$b_params,
  data.table("variable" = c("gamma_name", "gamma_surname"),
             "theta" = c(0.5, 0.5),
             "eta" = c(0.02, 0.02))
)

expect_equal(
  train_rec_lin(A = df_1, B = df_2,
                variables = c("name", "surname"),
                matches = matches)$pi_est,
  0.04
)

expect_equal(
  train_rec_lin(A = df_1, B = df_2,
                variables = c("name", "surname"),
                matches = matches)$match_prop,
  0.4
)

expect_silent(
  train_rec_lin(A = df_1, B = df_2,
                variables = c("name", "surname"),
                matches = matches,
                comparators = comparators,
                methods = methods_cpar)
)

expect_warning(
  train_rec_lin(A = df_1, B = df_2,
                variables = c("name", "surname"),
                matches = matches,
                comparators = comparators,
                methods = methods_cpar,
                prob_ratio = "2")
)

expect_equal(
  train_rec_lin(A = df_1, B = df_2,
                variables = c("name", "surname"),
                matches = matches,
                comparators = comparators,
                methods = methods_cpar)$cpar_params,
  data.table("variable" = c("gamma_name", "gamma_surname"),
             "p_0_M" = c(0.5, 0.5),
             "p_0_Omega" = c(0.02000000000000000041633, 0.02000000000000000041633),
             "alpha_M" = c(224.66614858286149569722, 15.65946723172959487158),
             "alpha_Omega" = c(5.407152306652986517577, 6.105019794944631428280),
             "beta_M" = c(2516.2608641280485244351, 176.1690063569579081104),
             "beta_Omega" = c(7.975415092480718115553, 9.363290038376906210260))
)

expect_warning(
  train_rec_lin(A = df_1, B = df_2,
                variables = c("name", "surname"),
                matches = matches,
                comparators = comparators,
                methods = methods_cnonpar)
)

expect_equal(
  train_rec_lin(A = df_1, B = df_2,
                variables = c("name", "surname"),
                matches = matches,
                comparators = comparators,
                methods = methods_cnonpar)$cnonpar_params,
  data.table("variable" = c("gamma_name", "gamma_surname"),
             "p_0_M_cnonpar" = c(0.5, 0.5),
             "p_0_U_cnonpar" = c(0.02, 0.02))
)

kliep_model <- densityratio::kliep(
  df_numerator = data.table("gamma_name" = c(0.1, 0.2, 0.3)),
  df_denominator = data.table("gamma_name" = c(0.2, 0.4, 0.6)),
  progressbar = FALSE,
  nfold = 2
)
kliep_pred <- as.vector(stats::predict(
  kliep_model,
  data.table("gamma_name" = 0.4)
))

expect_equal(
  automatedRecLin:::kliep_hurdle_ratio(
    df = data.table("gamma_name" = c(0, 0.4)),
    variables = "gamma_name",
    p_0_numerator = 0.4,
    p_0_denominator = 0.2,
    ratio_kliep_list = list(
      "gamma_name" = kliep_model
    )
  ),
  c(2, (1 - 0.4) / (1 - 0.2) * kliep_pred)
)

model_b <- train_rec_lin(A = df_1, B = df_2,
                         variables = c("name", "surname"),
                         matches = matches)

expect_equal(
  predict(model_b, df_new_1, df_new_2)$M_est,
  data.table("a" = 1:3, "b" = 1:3, "ratio" = rep(12.75510204081633602868, 3))
)

expect_equal(
  predict(model_b, df_new_1, df_new_2)$n_M_est,
  3.35401915868387945352
)

model_cpar <- train_rec_lin(A = df_1, B = df_2,
                            variables = c("name", "surname"),
                            matches = matches,
                            comparators = comparators,
                            methods = methods_cpar)

expect_equal(
  predict(model_cpar, df_new_1, df_new_2)$M_est,
  data.table("a" = c(1, 3), "b" = c(1, 3),
             "ratio" = c(59126.609831455622042995, 4201.052699951693284675))
)

expect_equal(
  predict(model_cpar, df_new_1, df_new_2)$n_M_est,
  2.000000561079755989624
)

expect_equal(
  predict(model_cpar, df_new_1, df_new_2)$mmr_est,
  0
)

expect_equal(
  predict(model_cpar, df_new_1, df_new_2, true_matches = true_matches)$eval_metrics,
  c("FLR" = 0, "MMR" = 1/3)
)

expect_equal(
  predict(model_cpar, df_new_1, df_new_2, true_matches = true_matches)$confusion,
  confusion_cpar
)

model_cnonpar <- train_rec_lin(A = df_1, B = df_2,
                               variables = c("name", "surname"),
                               matches = matches,
                               comparators = comparators,
                               methods = methods_cnonpar)

expect_equal(
  predict(model_cnonpar, df_new_1, df_new_2)$M_est,
  data.table("a" = c(2, 3, 1), "b" = c(2, 3, 1),
             "ratio" = c(47.61403316943127350669, 29.29835603699235235808, 26.65717991654741325647))
)

expect_equal(
  predict(model_cnonpar, df_new_1, df_new_2)$mmr_est,
  0
)

pred_cpar_flr <- predict(model_cpar, df_new_1, df_new_2,
                         set_construction = "flr", target_rate = 0.1)
expect_rate_bounds(pred_cpar_flr)

pred_cpar_mmr <- predict(model_cpar, df_new_1, df_new_2,
                         set_construction = "mmr", target_rate = 0.1)
expect_rate_bounds(pred_cpar_mmr)

# custom model

vectors <- comparison_vectors(A = df_1, B = df_2, variables = c("name", "surname"),
                              comparators = comparators, matches = matches)
model_xgb <- xgboost::xgboost(x = as.matrix(vectors$Omega[, c("gamma_name", "gamma_surname")]),
                              y = factor(vectors$Omega$match),
                              objective = "binary:logistic", eval_metric = "logloss",
                              nrounds = 100, verbosity = 0, nthread = 1)

expect_silent(
  custom_rec_lin_model(model_xgb, vectors)
)

expect_equal(
  custom_rec_lin_model(model_xgb, vectors)$pi_est,
  0.04
)

expect_equal(
  custom_rec_lin_model(model_xgb, vectors)$match_prop,
  0.4
)

custom_xgb_model <- custom_rec_lin_model(model_xgb, vectors)

expect_equal(
  predict(custom_xgb_model, df_new_1, df_new_2,
          data_type = "matrix", type = "response")$M_est,
  data.table("a" = 1:3, "b" = 1:3,
             "ratio" = rep(29.94770389694980394779, 3))
)

expect_equal(
  predict(custom_xgb_model, df_new_1, df_new_2,
          data_type = "matrix", type = "response")$flr_est,
  0.1591118016965802350882
)

expect_equal(
  predict(custom_xgb_model, df_new_1, df_new_2,
          data_type = "matrix", type = "response")$mmr_est,
  0.1591118016965802350882
)

#' @import data.table
#' @import stats
#' @importFrom utils head
#' @noRd

# Parametric bootstrap prototype for MEC standard error estimation.
# Work in progress; kept internal until the interface and limitations are final.
est_se_bootstrap <- function(mec_result,
                             B = 100,
                             alpha = 0.05) {

  # TODO add `mec_rec_lin` class validation

  # TODO add information about no support for cnonpar and hm

  n <- mec_result$n
  n_M_original <- mec_result$n_M_est
  n_M_est <- round(n_M_original)

  pi_hat <- n_M_original / n

  b_vars <- mec_result$b_vars
  cpar_vars <- mec_result$cpar_vars

  b_params <- mec_result$b_params
  cpar_params <- mec_result$cpar_params

  delta <- mec_result$delta
  eps <- mec_result$eps
  controls_nleqslv <- mec_result$controls_nleqslv

  n_M_boot <- numeric(B)

  for (b in seq_len(B)) {

    g <- sample(c(rep(1, n_M_est), rep(0, n - n_M_est)))
    # g <- rbinom(n, 1, pi_hat)
    Omega <- data.table(g = g)

    if (length(b_vars) > 0) {
      for (var in b_vars) {

        theta <- as.numeric(b_params[b_params[["variable"]] == var, "theta"])
        eta <- as.numeric(b_params[b_params[["variable"]] == var, "eta"])

        prob_vec <- fifelse(g == 1, theta, eta)
        set(Omega, j = var, value = rbinom(n, 1, prob_vec))

      }
    }

    if (length(cpar_vars) > 0) {
      for(var in cpar_vars) {

        p_0_M <- as.numeric(cpar_params[cpar_params[["variable"]] == var, "p_0_M"])
        alpha_M <- as.numeric(cpar_params[cpar_params[["variable"]] == var, "alpha_M"])
        beta_M <- as.numeric(cpar_params[cpar_params[["variable"]] == var, "beta_M"])
        p_0_U <- as.numeric(cpar_params[cpar_params[["variable"]] == var, "p_0_U"])
        alpha_U <- as.numeric(cpar_params[cpar_params[["variable"]] == var, "alpha_U"])
        beta_U <- as.numeric(cpar_params[cpar_params[["variable"]] == var, "beta_U"])

        p_0_vec <- fifelse(g == 1, p_0_M, p_0_U)
        alpha_vec <- fifelse(g == 1, alpha_M, alpha_U)
        beta_vec <- fifelse(g == 1, beta_M, beta_U)

        is_positive <- rbinom(n, 1, 1 - p_0_vec)
        gamma_values <- rgamma(n, shape = alpha_vec, rate = beta_vec)
        final_values <- fifelse(is_positive == 1, gamma_values, 0)
        set(Omega, j = var, value = final_values)

      }
    }

    n_M_boot[b] <- mec_relaxed(
      Omega = Omega,
      n = n,
      n_M_est = n_M_est,
      b_vars = b_vars,
      cpar_vars = cpar_vars,
      b_params_init = b_params,
      cpar_params_init = cpar_params,
      controls_nleqslv = controls_nleqslv,
      delta = delta,
      eps = eps
    )

  }

  se_n_M <- sqrt(1 / (B - 1) * sum((n_M_boot - n_M_original)^2))

  z_val <- qnorm(1 - alpha / 2)
  ci_normal <- c(
    lower = n_M_original - z_val * se_n_M,
    upper = n_M_original + z_val * se_n_M
  )

  ci_percentile <- quantile(n_M_boot, probs = c(alpha / 2, 1 - alpha / 2))
  names(ci_percentile) <- c("lower", "upper")

  structure(
    list(
      se = se_n_M,
      ci_normal = ci_normal,
      ci_percentile = ci_percentile,
      boot_dist = n_M_boot,
      n_M_original = n_M_original,
      B = B,
      alpha = alpha
    ),
    class = "mec_bootstrap"
  )

}

#' @noRd
mec_relaxed <- function(Omega,
                        n,
                        n_M_est,
                        b_vars,
                        cpar_vars,
                        b_params_init,
                        cpar_params_init,
                        controls_nleqslv,
                        delta,
                        eps) {

  b_params <- copy(b_params_init)
  cpar_params <- copy(cpar_params_init)

  set(Omega, j = "ratio", value = 1)

  if (length(b_vars) > 0) {

    Omega_b <- Omega[, b_vars, with = FALSE]

    b_params$eta <- binary_formula(Omega_b)

    b_numerator_list <- lapply(b_vars, function(col) {
      dbinom(x = Omega_b[[col]],
             size = 1,
             prob = as.numeric(b_params[b_params[["variable"]] == col, "theta"]))
    })
    b_numerator <- Reduce(`*`, b_numerator_list)
    b_denominator_list <- lapply(b_vars, function(col) {
      dbinom(x = Omega_b[[col]],
             size = 1,
             prob = as.numeric(b_params[b_params[["variable"]] == col, "eta"]))
    })
    b_denominator <- Reduce(`*`, b_denominator_list)

    set(Omega, j = "ratio", value = Omega[["ratio"]] * b_numerator / b_denominator)
    set(Omega, j = "b_denominator", value = b_denominator)

  }

  if (length(cpar_vars) > 0) {

    Omega_cpar <- Omega[, cpar_vars, with = FALSE]

    cpar_params$p_0_U <- p_0_formula(Omega_cpar)
    gamma_plus_U <- gamma_plus_formula(Omega_cpar)
    modified_nleqslv <- purrr::partial(nleqslv::nleqslv, control = controls_nleqslv)
    alpha_U <- alpha_formula(Omega_cpar, modified_nleqslv)
    cpar_params$alpha_U <- alpha_U
    cpar_params$beta_U <- alpha_U / gamma_plus_U

    cpar_numerator_list <- lapply(cpar_vars, function(col) {
      hurdle_gamma_density(x = Omega_cpar[[col]],
                           p_0 = as.numeric(cpar_params[cpar_params[["variable"]] == col, "p_0_M"]),
                           alpha = as.numeric(cpar_params[cpar_params[["variable"]] == col, "alpha_M"]),
                           beta = as.numeric(cpar_params[cpar_params[["variable"]] == col, "beta_M"]))
    })
    cpar_numerator <- Reduce(`*`, cpar_numerator_list)
    cpar_denominator_list <- lapply(cpar_vars, function(col) {
      hurdle_gamma_density(x = Omega_cpar[[col]],
                           p_0 = as.numeric(cpar_params[cpar_params[["variable"]] == col, "p_0_U"]),
                           alpha = as.numeric(cpar_params[cpar_params[["variable"]] == col, "alpha_U"]),
                           beta = as.numeric(cpar_params[cpar_params[["variable"]] == col, "beta_U"]))
    })
    cpar_denominator <- Reduce(`*`, cpar_denominator_list)

    set(Omega, j = "ratio", value = Omega[["ratio"]] * cpar_numerator / cpar_denominator)
    set(Omega, j = "cpar_denominator", value = cpar_denominator)

  }

  iter <- 1

  repeat {

    if (iter == 1) {
      g_est <- pmin(n_M_est * Omega[["ratio"]] / (n_M_est * (Omega[["ratio"]] - 1) + n), 1)
      n_M_old <- n_M_est
    } else {
      g_est <- pmin(NROW(M) * Omega[["ratio"]] / (NROW(M) * (Omega[["ratio"]] - 1) + n), 1)
      n_M_old <- n_M
    }

    n_M <- sum(g_est)
    set(Omega, j = "g_est", value = g_est)

    Omega <- Omega[order(-get("ratio")), ]
    M <- head(Omega, round(n_M))

    if (NROW(M) == 0) {
      break
    }

    if (iter >= 2) {

      old_params <- c()
      params <- c()
      if (length(b_vars) > 0) {
        old_params <- c(old_params, theta_b_old)
        params <- c(params, theta_b)
      }
      if (length(cpar_vars) > 0) {
        old_params <- c(old_params, p_0_M_old, alpha_M_old, beta_M_old)
        params <- c(params, p_0_M, alpha_M, beta_M)
      }
      if ((abs(n_M_old - n_M) < delta) || norm(old_params - params, type = "2") < eps) {
        break
      }

    }

    set(Omega, j = "ratio", value = 1)

    if (length(b_vars) > 0) {

      Omega_b <- Omega[, b_vars, with = FALSE]
      M_b <- M[, b_vars, with = FALSE]

      theta_b_old <- b_params$theta
      theta_b <- binary_formula(M_b)

      b_params$theta <- theta_b

      b_numerator_list <- lapply(b_vars, function(col) {
        stats::dbinom(x = Omega[[col]],
                      size = 1,
                      prob = as.numeric(b_params[b_params[["variable"]] == col, "theta"]))
      })
      b_numerator <- Reduce(`*`, b_numerator_list)
      set(Omega, j = "ratio", value = Omega[["ratio"]] * b_numerator / Omega[["b_denominator"]])

    }

    if (length(cpar_vars) > 0) {

      Omega_cpar <- Omega[, cpar_vars, with = FALSE]
      M_cpar <- M[, cpar_vars, with = FALSE]

      p_0_M_old <- cpar_params$p_0_M
      alpha_M_old <- cpar_params$alpha_M
      beta_M_old <- cpar_params$beta_M

      p_0_M <- p_0_formula(M_cpar)
      gamma_plus_M <- gamma_plus_formula(M_cpar)
      alpha_M <- alpha_formula(M_cpar, modified_nleqslv)
      beta_M <- alpha_M / gamma_plus_M
      beta_M[is.nan(beta_M)] <- beta_M_old[is.nan(beta_M)]

      cpar_params$p_0_M <- p_0_M
      cpar_params$alpha_M <- alpha_M
      cpar_params$beta_M <- beta_M

      cpar_numerator_list <- lapply(cpar_vars, function(col) {
        hurdle_gamma_density(x = Omega[[col]],
                             p_0 = as.numeric(cpar_params[cpar_params[["variable"]] == col, "p_0_M"]),
                             alpha = as.numeric(cpar_params[cpar_params[["variable"]] == col, "alpha_M"]),
                             beta = as.numeric(cpar_params[cpar_params[["variable"]] == col, "beta_M"]))
      })
      cpar_numerator <- Reduce(`*`, cpar_numerator_list)

      set(Omega, j = "ratio", value = Omega[["ratio"]] * cpar_numerator / Omega[["cpar_denominator"]])

    }

    iter <- iter + 1

  }

  return(n_M)

}

#' @title A_example dataset
#'
#' @description
#' An example dataset containing artificial personal data.
#'
#' @format
#' A `data.frame` with 10 records. Each row represents one record, with the following columns:
#' `name`, `surname`, and `city`. Some records can be matched with records in the
#' `B_example` dataset.
#'
#' @docType data
#' @keywords datasets
#' @name A_example
#' @rdname A_example
#' @examples
#' data("A_example")
#' A_example
"A_example"

#' @title B_example dataset
#'
#' @description
#' An example dataset containing artificial personal data.
#'
#' @format
#' A `data.frame` with 12 records. Each row represents one record, with the following columns:
#' `name`, `surname`, and `city`. Some records can be matched with records in the
#' `A_example` dataset.
#'
#' @docType data
#' @keywords datasets
#' @name B_example
#' @rdname B_example
#' @examples
#' data("B_example")
#' B_example
"B_example"

#' @title Fictional census data
#'
#' @description
#' This dataset was created by Paula McLeod, Dick Heasman and Ian Forbes, ONS,
#' for the ESSnet DI on-the-job training course, Southampton, 25-28 January 2011.
#' It contains fictional data representing some observations from a decennial Census.
#'
#' @format A `data.table` with 25343 records. Each row represents one record, with the following columns:
#' \itemize{
#' \item{`person_id` -- a unique number for each person, consisting of postcode, house number and person number,}
#' \item{`pername1` -- forename,}
#' \item{`pername2` -- surname,}
#' \item{`sex` -- gender (M/F),}
#' \item{`dob_day` -- day of birth,}
#' \item{`dob_mon` -- month of birth,}
#' \item{`dob_year` -- year of birth,}
#' \item{`hse_num` -- house number, a numeric label for each house within a street,}
#' \item{`enumcap` -- an address consisting of house number and street name,}
#' \item{`enumpc` -- postcode,}
#' \item{`str_nam` -- street name of person's household's street,}
#' \item{`cap_add` -- full address, consisting of house number, street name and postcode,}
#' \item{`census_id` -- person ID with "CENS" added in front.}
#' }
#'
#' @references
#' McLeod, P., Heasman, D., Forbes, I. (2011). Simulated data for the ESSnet DI on-the-job training course,
#' Southampton, 25-28 January 2011.
#' \url{https://wayback.archive-it.org/12090/20231221144450/https://cros-legacy.ec.europa.eu/content/job-training_en}
#'
#' @docType data
#' @keywords datasets
#' @name census
#' @rdname census
#' @examples
#'
#' data("census")
#' head(census)
#'
"census"

#' @title Fictional customer data
#'
#' @description
#' This dataset was created by Paula McLeod, Dick Heasman and Ian Forbes, ONS,
#' for the ESSnet DI on-the-job training course, Southampton, 25-28 January 2011.
#' It contains fictional observations from Customer Information System,
#' which is combined administrative data from the tax and benefit systems.
#'
#' @format A `data.table` with 24613 records. Each row represents one record, with the following columns:
#' \itemize{
#' \item{`person_id` -- a unique number for each person, consisting of postcode, house number and person number,}
#' \item{`pername1` -- forename,}
#' \item{`pername2` -- surname,}
#' \item{`sex` -- gender (M/F),}
#' \item{`dob_day` -- day of birth,}
#' \item{`dob_mon` -- month of birth,}
#' \item{`dob_year` -- year of birth,}
#' \item{`enumcap` -- an address consisting of house number and street name,}
#' \item{`enumpc` -- postcode,}
#' \item{`cis_id` -- person ID with "CIS" added in front.}
#' }
#'
#' @references
#' McLeod, P., Heasman, D., Forbes, I. (2011). Simulated data for the ESSnet DI on-the-job training course,
#' Southampton, 25-28 January 2011.
#' \url{https://wayback.archive-it.org/12090/20231221144450/https://cros-legacy.ec.europa.eu/content/job-training_en}
#'
#' @docType data
#' @keywords datasets
#' @name cis
#' @rdname cis
#' @examples
#'
#' data("cis")
#' head(cis)
#'
"cis"

#' @title Controls for the \link[densityratio:kliep]{kliep()} Function
#'
#' @author Adam Struzik
#'
#' @description
#' Controls for the \link[densityratio:kliep]{kliep()} function used in the package.
#'
#' @param scale `"numerator"`, `"denominator"` or `NULL`, indicating whether to standardize each numeric variable according
#' to the numerator means and standard deviations, the denominator means and standard deviations,
#' or apply no standardization at all.
#' @param progressbar Logical indicating whether or not to display a progress bar.
#' @param nfold Number of cross-validation folds used in order to calculate the optimal sigma value (default is 2-fold cross-validation).
#' @param ... Additional arguments.
#'
#' @return
#' Returns a list with parameters.
#'
#' @export
control_kliep <- function(scale = NULL,
                          progressbar = FALSE,
                          nfold = 2,
                          ...) {
  append(list(scale = scale, progressbar = progressbar, nfold = nfold),
         list(...))
}

#' @title Absolute Distance Comparison Function
#'
#' @author Adam Struzik
#'
#' @description
#' Creates a function that calculates the absolute distance between two values.
#'
#' @return
#' Returns a function taking two arguments, `x` and `y`, and returning their absolute difference.
#'
#' @examples
#' cmp <- abs_distance()
#' cmp(1, 5) # returns 4
#' @export
abs_distance <- function() {
  function(x, y) {
    abs(x - y)
  }
}

#' @importFrom reclin2 cmp_jarowinkler
#'
#' @title Jaro-Winkler Distance
#'
#' @author Adam Struzik
#'
#' @description
#' Creates a function that calculates the Jaro-Winkler distance between
#' two strings, defined as
#' \eqn{1 - \text{Jaro-Winkler similarity}}.
#'
#' @return
#' Returns a function taking two string arguments, `x` and `y`,
#' and returning the Jaro-Winkler distance.
#'
#' @export
jarowinkler_complement <- function() {
  temp <- reclin2::cmp_jarowinkler()
  function(x, y) {
    1 - temp(x, y)
  }
}

if ( requireNamespace("tinytest", quietly=TRUE) ){
  old_omp_thread_limit <- Sys.getenv("OMP_THREAD_LIMIT", unset = NA_character_)
  Sys.setenv("OMP_THREAD_LIMIT" = "1")
  on.exit({
    if (is.na(old_omp_thread_limit)) {
      Sys.unsetenv("OMP_THREAD_LIMIT")
    } else {
      Sys.setenv("OMP_THREAD_LIMIT" = old_omp_thread_limit)
    }
  }, add = TRUE)

  if ( requireNamespace("data.table", quietly=TRUE) ){
    old_dt_threads <- data.table::getDTthreads()
    data.table::setDTthreads(1L)
    on.exit(data.table::setDTthreads(old_dt_threads), add = TRUE)
  }

  tinytest::test_package("automatedRecLin", ncpu = NULL)
}

---
title: "MEC with blocking"
author: "Adam Struzik"
output: 
    html_vignette:
        df_print: kable
        toc: true
        number_sections: true
        fig_width: 6
        fig_height: 4
vignette: >
  %\VignetteIndexEntry{MEC with blocking}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Setup

Load required packages.

```{r setup, message = FALSE}
library(automatedRecLin)
library(data.table)

options("text2vec.mc.cores" = 1L)
```

## Data

We use the full example Census and Customer Information System (CIS) datasets
from [McLeod et al. (2011)](https://wayback.archive-it.org/12090/20231221144450/https://cros-legacy.ec.europa.eu/content/job-training_en).
The goal is to link records from CIS to records from Census.

```{r data}
data("census", package = "automatedRecLin")
data("cis", package = "automatedRecLin")
setDT(census)
setDT(cis)

NROW(cis)
NROW(census)
```

The `person_id` variable identifies the correct linkage.
We use this information only to evaluate the result.

```{r true-matches}
cis[is.na(cis)] <- ""
census[is.na(census)] <- ""

cis[, pername1 := gsub("-", "", pername1)]
census[, pername1 := gsub("-", "", pername1)]

true_matches <- merge(
  x = cis[, .(a = .I, person_id)],
  y = census[, .(b = .I, person_id)],
  by = "person_id"
)[, .(a, b)]

NROW(true_matches)
```

## MEC with blocking

We compare forename and surname using the Jaro-Winkler distance. These two
comparison variables are modeled with the continuous parametric MEC method.
Sex and date-of-birth variables use the default binary method. Address fields
are used only to construct blocks.

```{r model-specification}
variables <- c(
  "pername1", "pername2", "sex",
  "dob_day", "dob_mon", "dob_year"
)

comparators <- list(
  "pername1" = jarowinkler_complement(),
  "pername2" = jarowinkler_complement()
)

methods <- list(
  "pername1" = "continuous_parametric",
  "pername2" = "continuous_parametric"
)

blocking_variables <- c(variables, "enumcap", "enumpc")
```

Run blocked MEC. The model is trained on all candidate pairs retained by
blocking.

```{r mec-blocking}
set.seed(1)

result <- mec_blocking(
  A = cis,
  B = census,
  variables = variables,
  comparators = comparators,
  methods = methods,
  blocking_variables = blocking_variables,
  blocking_sep = " ",
  controls_blocking = list(seed = 1, n_threads = 1),
  alpha = 0.5,
  true_matches = true_matches
)

result
```

## Blocking efficiency and linkage results

The full Cartesian product contains `r format(NROW(cis) * NROW(census), big.mark = ",")`
record pairs. Blocking reduces this to `r format(result$blocking_eval[["blocked_pairs"]], big.mark = ",")`
candidate pairs, while retaining `r sprintf("%.2f%%", 100 * result$blocking_eval[["blocking_recall"]])`
of known links. The final linkage set contains `r format(NROW(result$M_est), big.mark = ",")`
predicted matches.

```{r results, echo = FALSE}
data.table(
  step = c("Training", "Blocking", "Linkage"),
  result = c(
    paste0(
      "all_candidate_pairs on ",
      format(NROW(result$block_summary), big.mark = ","),
      " blocks"
    ),
    paste0(
      format(result$blocking_eval[["preserved_matches"]], big.mark = ","),
      " of ",
      format(result$blocking_eval[["true_matches"]], big.mark = ","),
      " known links retained"
    ),
    paste0(
      "FLR = ", sprintf("%.2f%%", 100 * result$eval_metrics[["FLR"]]),
      "; MMR = ", sprintf("%.2f%%", 100 * result$eval_metrics[["MMR"]])
    )
  )
)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/comparators.R
\name{abs_distance}
\alias{abs_distance}
\title{Absolute Distance Comparison Function}
\usage{
abs_distance()
}
\value{
Returns a function taking two arguments, \code{x} and \code{y}, and returning their absolute difference.
}
\description{
Creates a function that calculates the absolute distance between two values.
}
\examples{
cmp <- abs_distance()
cmp(1, 5) # returns 4
}
\author{
Adam Struzik
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised_learning.R
\name{mec_blocking}
\alias{mec_blocking}
\title{Blocked Unsupervised Maximum Entropy Classifier for Record Linkage}
\usage{
mec_blocking(
  A,
  B,
  variables,
  comparators = NULL,
  methods = NULL,
  blocking_x = NULL,
  blocking_y = NULL,
  blocking_variables = variables,
  blocking_sep = " ",
  controls_blocking = list(),
  start_params = NULL,
  alpha = 0,
  delta = 0.5,
  eps = 0.05,
  controls_nleqslv = list(),
  true_matches = NULL,
  keep_blocking_result = FALSE,
  keep_training_data = FALSE,
  verbose = FALSE
)
}
\arguments{
\item{A}{A duplicate-free \code{data.frame} or \code{data.table}.}

\item{B}{A duplicate-free \code{data.frame} or \code{data.table}.}

\item{variables}{A character vector of key variables used to create MEC comparison vectors.}

\item{comparators}{A named list of functions for comparing pairs of records.}

\item{methods}{A named list of methods used for estimation (\code{"binary"} or
\code{"continuous_parametric"}). Other unsupervised MEC methods are not
supported by \code{\link[=mec_blocking]{mec_blocking()}} at this stage.}

\item{blocking_x}{Optional input passed as \code{x} to \link[blocking:blocking]{blocking()}.}

\item{blocking_y}{Optional input passed as \code{y} to \link[blocking:blocking]{blocking()}.}

\item{blocking_variables}{Variables used to create blocking strings when
\code{blocking_x} and \code{blocking_y} are not supplied.}

\item{blocking_sep}{Separator used when concatenating \code{blocking_variables}.}

\item{controls_blocking}{A list of additional arguments passed to
\link[blocking:blocking]{blocking()}, except \code{x} and \code{y}.}

\item{start_params}{Start parameters for the \code{"binary"} and
\code{"continuous_parametric"} methods.}

\item{alpha}{A single numeric value in \verb{[0, 1)} controlling the fraction of
the current nonmatch complement dropped from nonmatch-side parameter fitting
after the first inverted MEC iteration. The first U-side fit uses the full
initial fitting set, and posterior/count formulas continue to use the full
current nonmatch count.}

\item{delta}{A numeric value specifying the tolerance for the change in the
estimated number of nonmatches between MEC iterations.}

\item{eps}{A numeric value specifying the tolerance for the change in model
parameters between MEC iterations.}

\item{controls_nleqslv}{Controls passed to the \link[nleqslv:nleqslv]{nleqslv()} function
(only if the \code{"continuous_parametric"} method has been chosen for at least one
variable).}

\item{true_matches}{A \code{data.frame} or \code{data.table} indicating known matches.}

\item{keep_blocking_result}{Logical indicating whether to store the raw object
returned by \link[blocking:blocking]{blocking()}.}

\item{keep_training_data}{Logical indicating whether to store pooled training
comparison vectors.}

\item{verbose}{Logical indicating whether to print progress messages.}
}
\value{
Returns a list of class \code{"mec_blocking"} containing:
\itemize{
\item{\code{M_est} -- a \code{data.table} with predicted matches and columns \code{a}, \code{b}, \code{block}, and \code{ratio},}
\item{\code{n_M_est} -- estimated total number of matches across all blocks,}
\item{\code{n_U_est} -- estimated total number of candidate nonmatches,}
\item{\code{alpha} -- fraction of the current nonmatch complement dropped from later U-side fitting,}
\item{\code{candidate_pair_count} -- number of candidate pairs in \eqn{\Omega_B},}
\item{\code{block_estimates} -- a \code{data.table} with block-level size and match-count diagnostics,}
\item{\code{block_summary} -- a \code{data.table} describing the final disjoint blocks,}
\item{\code{excluded_records} -- a list with records from \code{A} and \code{B} excluded by blocking,}
\item{\code{b_vars} -- variables used for the \code{"binary"} method, with the prefix \code{"gamma_"},}
\item{\code{cpar_vars} -- variables used for the \code{"continuous_parametric"} method, with the prefix \code{"gamma_"},}
\item{\code{b_params} -- parameters estimated using the \code{"binary"} method,}
\item{\code{cpar_params} -- parameters estimated using the \code{"continuous_parametric"} method,}
\item{\code{variables} -- key variables used for comparison,}
\item{\code{comparators} -- comparison functions used to create comparison vectors,}
\item{\code{methods} -- MEC estimation methods used for the key variables,}
\item{\code{delta} -- tolerance for changes in the estimated number of nonmatches,}
\item{\code{eps} -- tolerance for changes in nonmatch-side model parameters,}
\item{\code{controls_nleqslv} -- controls passed to \link[nleqslv:nleqslv]{nleqslv()},}
\item{\code{blocking_result} -- raw object returned by \link[blocking:blocking]{blocking()} if \code{keep_blocking_result = TRUE}; otherwise \code{NULL},}
\item{\code{training_Omega} -- candidate-space comparison vectors with inverted scores if \code{keep_training_data = TRUE}; otherwise \code{NULL},}
\item{\code{blocking_eval} -- blocking diagnostics if \code{true_matches} is provided; otherwise \code{NULL},}
\item{\code{mec_eval} -- MEC-selection diagnostics among known matches retained in the candidate-pair space if \code{true_matches} is provided; otherwise \code{NULL},}
\item{\code{eval_metrics} -- empirical linkage quality metrics based on \code{true_matches}; otherwise \code{NULL},}
\item{\code{confusion} -- empirical confusion matrix based on \code{true_matches}; otherwise \code{NULL}.}
}
}
\description{
Runs graph-based blocking using \link[blocking:blocking]{blocking()},
defines a blocking candidate-pair space, and fits an inverted unsupervised
maximum entropy classifier (MEC) directly on all candidate pairs.
}
\details{
The function assumes one-to-one linkage. The blocking stage defines disjoint
bipartite blocks, and the candidate-pair space \eqn{\Omega_B} is the union of
within-block Cartesian products. Duplicate candidate pairs are removed
deterministically before MEC fitting.

The blocked MEC fit is inverted relative to \code{\link[=mec]{mec()}}. The initial match set
contains at most \eqn{\nu} feasible pairs, where \eqn{\nu} is the structural
one-to-one upper bound. Initial feasible matches are selected
greedily by an unweighted disagreement norm: binary agreement indicators use
\code{1 - gamma}, while continuous dissimilarities use \code{gamma} unchanged. At each
iteration, match-side parameters are estimated from the current greedy
one-to-one match set, and nonmatch-side parameters are estimated from its
complement.

The \code{alpha} argument applies only to nonmatch-side distribution estimation.
The first U-side fit uses the full initial complement. In later iterations,
the least reliable current nonmatches are dropped from the U-side fitting
sample, with reliability ranked by the previous nonmatch posterior estimate
and then by the inverted density ratio if the posterior is unavailable.
The posterior and count updates still use the full current complement size,
and the final match set remains one-to-one.

The returned \code{ratio} is \eqn{s = u / m}, where \eqn{u} and \eqn{m} denote
the estimated nonmatch and match comparison-vector densities. Smaller values
are therefore more match-like. Updated match sets are selected greedily in
ascending order of this ratio.

If \eqn{N = |\Omega_B|} and \eqn{\nu} is the maximum feasible one-to-one
matching size in the candidate graph, the estimated number of nonmatches is
bounded below by \eqn{N - \nu}. For the disjoint complete blocks reconstructed
by this function, \eqn{\nu = \sum_h \min(n_{Ah}, n_{Bh})}.

If the initialized match set exhausts the candidate-pair space, for example
when \eqn{N = \nu}, there is no candidate complement from which to estimate
nonmatch parameters. In that case the function returns the structurally
feasible initialized match set, sets \code{n_U_est = 0}, and leaves nonmatch-side
parameters unavailable.
}
\examples{
df_1 <- data.frame(
  name = c("Emma", "Liam", "Olivia", "Noah", "Ava"),
  surname = c("Smith", "Jones", "Brown", "Davis", "Miller"),
  city = c("Boston", "Boston", "Austin", "Austin", "Denver")
)
df_2 <- data.frame(
  name = c("Emma", "Liam", "Olivia", "Noah", "Ava"),
  surname = c("Smith", "Jones", "Brown", "Davis", "Miller"),
  city = c("Boston", "Boston", "Austin", "Austin", "Denver")
)

blocking_x <- matrix(
  c(1, 0, 0, 1, 1, 1, 2, 0, 0, 2),
  ncol = 2,
  byrow = TRUE
)
blocking_y <- blocking_x

result <- mec_blocking(
  A = df_1,
  B = df_2,
  variables = c("name", "surname", "city"),
  blocking_x = blocking_x,
  blocking_y = blocking_y,
  controls_blocking = list(
    representation = "custom_matrix",
    ann = "kd",
    distance = "euclidean",
    seed = 1
  ),
  true_matches = data.frame(a = 1:5, b = 1:5)
)
result
}
\author{
Adam Struzik
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/comparators.R
\name{jarowinkler_complement}
\alias{jarowinkler_complement}
\title{Jaro-Winkler Distance}
\usage{
jarowinkler_complement()
}
\value{
Returns a function taking two string arguments, \code{x} and \code{y},
and returning the Jaro-Winkler distance.
}
\description{
Creates a function that calculates the Jaro-Winkler distance between
two strings, defined as
\eqn{1 - \text{Jaro-Winkler similarity}}.
}
\author{
Adam Struzik
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{A_example}
\alias{A_example}
\title{A_example dataset}
\format{
A \code{data.frame} with 10 records. Each row represents one record, with the following columns:
\code{name}, \code{surname}, and \code{city}. Some records can be matched with records in the
\code{B_example} dataset.
}
\usage{
A_example
}
\description{
An example dataset containing artificial personal data.
}
\examples{
data("A_example")
A_example
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/comparison_vectors.R
\name{comparison_vectors}
\alias{comparison_vectors}
\title{Create Comparison Vectors for Record Linkage}
\usage{
comparison_vectors(
  A,
  B,
  variables,
  comparators = NULL,
  pairs = NULL,
  matches = NULL
)
}
\arguments{
\item{A}{A duplicate-free \code{data.frame} or \code{data.table}.}

\item{B}{A duplicate-free \code{data.frame} or \code{data.table}.}

\item{variables}{A character vector of key variables used to create comparison vectors.}

\item{comparators}{A named list of functions for comparing pairs of records.}

\item{pairs}{Optional. A \code{data.frame} or \code{data.table} with columns \code{a} and \code{b}
indicating pairs for which comparison vectors should be created.}

\item{matches}{Optional. A \code{data.frame} or \code{data.table} indicating known matches.}
}
\value{
Returns a list containing:\cr
\itemize{
\item{\code{Omega} -- a \code{data.table} with comparison vectors between records from both datasets,
including optional match information,}
\item{\code{variables} -- a character vector of key variables used for comparison,}
\item{\code{comparators} -- a list of functions used to compare pairs of records,}
\item{\code{match_prop} -- proportion of matches in the smaller dataset.}
}
}
\description{
Creates comparison vectors between records in two datasets based on specified variables
and comparison functions.
}
\details{
Consider two datasets: \eqn{A} and \eqn{B}.
For each pair of records \eqn{(a,b) \in \Omega},
the function creates a comparison vector
\eqn{\pmb{\gamma}_{ab} = (\gamma_{ab}^1,\gamma_{ab}^2,\ldots,\gamma_{ab}^K)'}
based on specified \eqn{K} variables and comparison functions.
}
\note{
Each comparison function must return another function,
which serves as the actual comparator.
}
\examples{
df_1 <- data.frame(
"name" = c("John", "Emily", "Mark", "Anna", "David"),
"surname" = c("Smith", "Johnson", "Taylor", "Williams", "Brown")
)
df_2 <- data.frame(
  "name" = c("Jon", "Emely", "Marc", "Michael"),
  "surname" = c("Smitth", "Jonson", "Tailor", "Henderson")
)
comparators <- list("name" = jarowinkler_complement(),
                    "surname" = jarowinkler_complement())
matches <- data.frame("a" = 1:3, "b" = 1:3)
result <- comparison_vectors(A = df_1, B = df_2, variables = c("name", "surname"),
                             comparators = comparators, matches = matches)
result
}
\author{
Adam Struzik
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised_learning.R
\name{train_rec_lin}
\alias{train_rec_lin}
\title{Train a Record Linkage Model}
\usage{
train_rec_lin(
  A,
  B,
  matches,
  variables,
  comparators = NULL,
  methods = NULL,
  prob_ratio = NULL,
  nonpar_hurdle = TRUE,
  controls_nleqslv = list(),
  controls_kliep = control_kliep(),
  verbose = FALSE
)
}
\arguments{
\item{A}{A duplicate-free \code{data.frame} or \code{data.table}.}

\item{B}{A duplicate-free \code{data.frame} or \code{data.table}.}

\item{matches}{A \code{data.frame} or \code{data.table} indicating known matches.}

\item{variables}{A character vector of key variables used to create comparison vectors.}

\item{comparators}{A named list of functions for comparing pairs of records.}

\item{methods}{A named list of methods used for estimation (\code{"binary"}, \code{"continuous_parametric"} or \code{"continuous_nonparametric"}).}

\item{prob_ratio}{Probability/density ratio type (\code{"1"} or \code{"2"}).}

\item{nonpar_hurdle}{Logical indicating whether to use a hurdle model or not
(used only if the \code{"continuous_nonparametric"} method has been chosen for at least one variable).}

\item{controls_nleqslv}{Controls passed to the \link[nleqslv:nleqslv]{nleqslv()} function (only if the \code{"continuous_parametric"} method has been chosen for at least one variable).}

\item{controls_kliep}{Controls passed to the \link[densityratio:kliep]{kliep()} function (only if the \code{"continuous_nonparametric"} method has been chosen for at least one variable).}

\item{verbose}{Logical indicating whether to print progress messages.}
}
\value{
Returns a list containing:\cr
\itemize{
\item{\code{b_vars} -- a character vector of variables used for the \code{"binary"} method (with the prefix \code{"gamma_"}),}
\item{\code{cpar_vars} -- a character vector of variables used for the \code{"continuous_parametric"} method (with the prefix \code{"gamma_"}),}
\item{\code{cnonpar_vars} -- a character vector of variables used for the \code{"continuous_nonparametric"} method (with the prefix \code{"gamma_"}),}
\item{\code{b_params} -- parameters estimated using the \code{"binary"} method,}
\item{\code{cpar_params} -- parameters estimated using the \code{"continuous_parametric"} method,}
\item{\code{cnonpar_params} -- probability of exact matching estimated using the \code{"continuous_nonparametric"} method,}
\item{\code{ratio_kliep} -- a result of the \link[densityratio:kliep]{kliep()} function,}
\item{\code{ratio_kliep_list} -- an object containing the results of the \link[densityratio:kliep]{kliep()} function,}
\item{\code{ml_model} -- here \code{NULL},}
\item{\code{pi_est} -- a prior probability of matching,}
\item{\code{match_prop} -- proportion of matches in the smaller dataset,}
\item{\code{variables} -- a character vector of key variables used for comparison,}
\item{\code{comparators} -- a list of functions used to compare pairs of records,}
\item{\code{methods} -- a list of methods used for estimation,}
\item{\code{"prob_ratio"} -- probability/density ratio type.}
}
}
\description{
Trains a supervised record linkage model using probability or density ratio estimation,
based on \href{https://www150.statcan.gc.ca/n1/pub/12-001-x/2022001/article/00007-eng.htm}{Lee et al. (2022)},
with several extensions.
}
\details{
Consider two datasets: \eqn{A} and \eqn{B}.
Let the bipartite comparison space \eqn{\Omega = A \times B} consist of
matches \eqn{M} and non-matches \eqn{U} between the records in files
\eqn{A} and \eqn{B}. For any pair of records \eqn{(a,b) \in \Omega},
let \eqn{\pmb{\gamma}_{ab} = (\gamma_{ab}^1,\gamma_{ab}^2,
\ldots,\gamma_{ab}^K)'} be the comparison vector between
a set of key variables. The original MEC algorithm uses the binary
comparison function to evaluate record pairs across two datasets.
However, this approach may be insufficient when handling datasets
with frequent errors across multiple variables.

We propose the use of continuous comparison functions to address
the limitations of binary comparison methods. We consider every
semi-metric, i.e., a function \eqn{d: A \times B \to \mathbb{R}},
satisfying the following conditions:\cr
\enumerate{
\item{\eqn{d(x,y) \geq 0},}
\item{\eqn{d(x,y) = 0} if and only if \eqn{x = y},}
\item{\eqn{d(x,y) = d(y,x)}.}
}
For example, we can use the Jaro-Winkler distance for character variables
(which is implemented in the \code{automatedRecLin} package as \code{\link[=jarowinkler_complement]{jarowinkler_complement()}})
or the Euclidean distance for numerical variables. The \code{automatedRecLin} package allows the use of
a different comparison function for each key variable (which should be specified
as a list in the \code{comparators} argument). The default function
for each key variable is \link[reclin2:cmp_identical]{cmp_identical()}
(the binary comparison function).

The \code{\link[=train_rec_lin]{train_rec_lin()}} function is used to train a record linkage model,
when \eqn{M} and \eqn{U} are known (which might later serve as a classifier
for pairs outside \eqn{\Omega}). It offers different approaches to estimate the
probability/density ratio between matches and non-matches, which should be
specified as a list in the \code{methods} argument. The method suitable for the binary
comparison function is \code{"binary"}, which is also the default method for each
variable.

For the continuous semi-metrics we suggest the usage
of \code{"continuous_parametric"} or \code{"continuous_nonparametric"}
method. The \code{"continuous_parametric"} method assumes that
\eqn{\gamma_{ab}^k|M} and \eqn{\gamma_{ab}^k|U} follow
hurdle Gamma distributions. The density function of a hurdle
Gamma distribution is characterized by three parameters
\eqn{p_0 \in (0,1)} and \eqn{\alpha, \beta > 0} as follows:
\deqn{
f(x;p_0,\alpha,\beta) = p_0^{\mathbb{I}(x = 0)}[(1 - p_0) v(x;\alpha,\beta)]^{\mathbb{I}(x > 0)},
}
where
\deqn{
v(x;\alpha,\beta) = \frac{\beta^{\alpha} x^{\alpha - 1} \exp(-\beta x)}
{\Gamma(\alpha)}
}
is the density function of a Gamma distribution
(for details see \href{https://ideas.repec.org/a/eee/csdana/v179y2023ics0167947322002365.html}{Vo et al. (2023)}).
The \code{"continuous_nonparametric"} method does not assume anything about
the distributions of the comparison vectors. It directly
estimates the density ratio between the matches and the non-matches, using
the Kullback-Leibler Importance Estimation Procedure (KLIEP).
For details see \href{https://link.springer.com/article/10.1007/s10463-008-0197-x}{Sugiyama et al. (2008)}.
}
\examples{
df_1 <- data.frame(
  "name" = c("James", "Emma", "William", "Olivia", "Thomas",
  "Sophie", "Harry", "Amelia", "George", "Isabella"),
  "surname" = c("Smith", "Johnson", "Brown", "Taylor", "Wilson",
  "Davis", "Clark", "Harris", "Lewis", "Walker")
)
 df_2 <- data.frame(
  "name" = c("James", "Ema", "Wimliam", "Olivia", "Charlotte",
  "Henry", "Lucy", "Edward", "Alice", "Jack"),
  "surname" = c("Smith", "Johnson", "Bron", "Tailor", "Moore",
  "Evans", "Hall", "Wright", "Green", "King")
)
comparators <- list("name" = jarowinkler_complement(),
                    "surname" = jarowinkler_complement())
matches <- data.frame("a" = 1:4, "b" = 1:4)
methods <- list("name" = "continuous_nonparametric",
                "surname" = "continuous_nonparametric")
model <- train_rec_lin(A = df_1, B = df_2, matches = matches,
                       variables = c("name", "surname"),
                       comparators = comparators,
                       methods = methods)
model
}
\references{
Lee, D., Zhang, L.-C. and Kim, J. K. (2022). Maximum entropy classification for record linkage.
Survey Methodology, Statistics Canada, Catalogue No. 12-001-X, Vol. 48, No. 1.

Vo, T. H., Chauvet, G., Happe, A., Oger, E., Paquelet, S., and Garès, V. (2023).
Extending the Fellegi-Sunter record linkage model for mixed-type data with application to the French national health data system.
Computational Statistics & Data Analysis, 179, 107656.

Sugiyama, M., Suzuki, T., Nakajima, S. et al. Direct importance estimation for covariate shift adaptation.
Ann Inst Stat Math 60, 699–746 (2008). \doi{10.1007/s10463-008-0197-x}
}
\author{
Adam Struzik
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.R
\name{predict.rec_lin_model}
\alias{predict.rec_lin_model}
\title{Predict Matches Based on a Given Record Linkage Model}
\usage{
\method{predict}{rec_lin_model}(
  object,
  newdata_A,
  newdata_B,
  duplicates_in_A = FALSE,
  set_construction = c("size", "flr", "mmr"),
  fixed_method = "Newton",
  target_rate = 0.03,
  tol = 0.005,
  max_iter = 50,
  data_type = c("data.frame", "data.table", "matrix"),
  true_matches = NULL,
  verbose = FALSE,
  ...
)
}
\arguments{
\item{object}{A \code{rec_lin_model} object from \code{\link[=train_rec_lin]{train_rec_lin()}} or \code{\link[=custom_rec_lin_model]{custom_rec_lin_model()}}.}

\item{newdata_A}{A duplicate-free \code{data.frame} or \code{data.table}.}

\item{newdata_B}{A duplicate-free \code{data.frame} or \code{data.table}.}

\item{duplicates_in_A}{Logical indicating whether to allow \code{A} to have duplicate records.}

\item{set_construction}{A method for constructing the predicted set of matches (\code{"size"}, \code{"flr"} or \code{"mmr"}).}

\item{fixed_method}{A method for solving fixed-point equations using the \link[FixedPoint:FixedPoint]{FixedPoint()} function.}

\item{target_rate}{A target false link rate (FLR) or missing match rate (MMR)
(used only if \code{set_construction == "flr"} or \code{set_construction == "mmr"}).}

\item{tol}{Error tolerance in the bisection procedure
(used only if \code{set_construction == "flr"} or \code{set_construction == "mmr"}).}

\item{max_iter}{A maximum number of iterations for the bisection procedure
(used only if \code{set_construction == "flr"} or \code{set_construction == "mmr"}).}

\item{data_type}{Data type for predictions with a custom ML model (\code{"data.frame"}, \code{"data.table"} or \code{"matrix"};
used only if \code{object} is from \code{\link[=custom_rec_lin_model]{custom_rec_lin_model()}}).}

\item{true_matches}{A \code{data.frame} or \code{data.table} indicating true matches.}

\item{verbose}{Logical indicating whether to print progress messages.}

\item{...}{Additional controls passed to \code{\link[=predict.rec_lin_model]{predict.rec_lin_model()}} for custom ML model
(used only if the \code{object} is from \code{\link[=custom_rec_lin_model]{custom_rec_lin_model()}}).}
}
\value{
Returns a list containing:\cr
\itemize{
\item{\code{M_est} -- a \code{data.table} with predicted matches,}
\item{\code{set_construction} -- a method for constructing the predicted set of matches,}
\item{\code{n_M_est} -- estimated classification set size,}
\item{\code{flr_est} -- estimated false link rate (FLR),}
\item{\code{mmr_est} -- estimated missing match rate (MMR),}
\item{\code{iter} -- the number of iterations in the bisection procedure,}
\item{\code{eval_metrics} -- standard metrics for quality assessment, if \code{true_matches} is provided,}
\item{\code{confusion} -- confusion matrix, if \code{true_matches} is provided.}
}
}
\description{
Predicts matches between records in two datasets based on a given record linkage model,
using the maximum entropy classification (MEC) algorithm
(see \href{https://www150.statcan.gc.ca/n1/pub/12-001-x/2022001/article/00007-eng.htm}{Lee et al. (2022)}).
}
\details{
The \code{\link[=predict.rec_lin_model]{predict.rec_lin_model()}} method estimates the probability/density ratio
between matches and non-matches for pairs in given
datasets, based on a model obtained using the
\code{\link[=train_rec_lin]{train_rec_lin()}} or \code{\link[=custom_rec_lin_model]{custom_rec_lin_model()}}.
Then, it estimates the number of matches and
returns the predicted matches, using the maximum
entropy classification (MEC) algorithm
(see \href{https://www150.statcan.gc.ca/n1/pub/12-001-x/2022001/article/00007-eng.htm}{Lee et al. (2022)}).

The \code{\link[=predict.rec_lin_model]{predict.rec_lin_model()}} method allows the construction of the predicted set
of matches using its estimated size or the bisection procedure,
described in \href{https://www150.statcan.gc.ca/n1/pub/12-001-x/2022001/article/00007-eng.htm}{Lee et al. (2022)},
based on a target false link rate (FLR)
or missing match rate (MMR). To use the second option, set \code{set_construction = "flr"}
or \code{set_construction = "mmr"} and
specify a target error rate using the \code{target_rate} argument.

By default, the function assumes that the datasets \code{newdata_A} and \code{newdata_B}
contain no duplicate records. This assumption
might be relaxed by allowing \code{newdata_A} to have duplicates. To do so,
set \code{duplicates_in_A = TRUE}.
}
\examples{
df_1 <- data.frame(
  "name" = c("James", "Emma", "William", "Olivia", "Thomas",
  "Sophie", "Harry", "Amelia", "George", "Isabella"),
  "surname" = c("Smith", "Johnson", "Brown", "Taylor", "Wilson",
  "Davis", "Clark", "Harris", "Lewis", "Walker")
)
 df_2 <- data.frame(
  "name" = c("James", "Ema", "Wimliam", "Olivia", "Charlotte",
  "Henry", "Lucy", "Edward", "Alice", "Jack"),
  "surname" = c("Smith", "Johnson", "Bron", "Tailor", "Moore",
  "Evans", "Hall", "Wright", "Green", "King")
)
comparators <- list("name" = jarowinkler_complement(),
                    "surname" = jarowinkler_complement())
matches <- data.frame("a" = 1:4, "b" = 1:4)
methods <- list("name" = "continuous_nonparametric",
                "surname" = "continuous_nonparametric")
model <- train_rec_lin(A = df_1, B = df_2, matches = matches,
                       variables = c("name", "surname"),
                       comparators = comparators,
                       methods = methods)

df_new_1 <- data.frame(
  "name" = c("John", "Emily", "Mark", "Anna", "David"),
  "surname" = c("Smith", "Johnson", "Taylor", "Williams", "Brown")
)
df_new_2 <- data.frame(
  "name" = c("John", "Emely", "Mark", "Michael"),
  "surname" = c("Smitth", "Johnson", "Tailor", "Henders")
)
predict(model, df_new_1, df_new_2)
}
\references{
Lee, D., Zhang, L.-C. and Kim, J. K. (2022). Maximum entropy classification for record linkage.
Survey Methodology, Statistics Canada, Catalogue No. 12-001-X, Vol. 48, No. 1.

Vo, T. H., Chauvet, G., Happe, A., Oger, E., Paquelet, S., and Garès, V. (2023).
Extending the Fellegi-Sunter record linkage model for mixed-type data with application to the French national health data system.
Computational Statistics & Data Analysis, 179, 107656.

Sugiyama, M., Suzuki, T., Nakajima, S. et al. Direct importance estimation for covariate shift adaptation.
Ann Inst Stat Math 60, 699–746 (2008). \doi{10.1007/s10463-008-0197-x}
}
\author{
Adam Struzik
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{census}
\alias{census}
\title{Fictional census data}
\format{
A \code{data.table} with 25343 records. Each row represents one record, with the following columns:
\itemize{
\item{\code{person_id} -- a unique number for each person, consisting of postcode, house number and person number,}
\item{\code{pername1} -- forename,}
\item{\code{pername2} -- surname,}
\item{\code{sex} -- gender (M/F),}
\item{\code{dob_day} -- day of birth,}
\item{\code{dob_mon} -- month of birth,}
\item{\code{dob_year} -- year of birth,}
\item{\code{hse_num} -- house number, a numeric label for each house within a street,}
\item{\code{enumcap} -- an address consisting of house number and street name,}
\item{\code{enumpc} -- postcode,}
\item{\code{str_nam} -- street name of person's household's street,}
\item{\code{cap_add} -- full address, consisting of house number, street name and postcode,}
\item{\code{census_id} -- person ID with "CENS" added in front.}
}
}
\usage{
census
}
\description{
This dataset was created by Paula McLeod, Dick Heasman and Ian Forbes, ONS,
for the ESSnet DI on-the-job training course, Southampton, 25-28 January 2011.
It contains fictional data representing some observations from a decennial Census.
}
\examples{

data("census")
head(census)

}
\references{
McLeod, P., Heasman, D., Forbes, I. (2011). Simulated data for the ESSnet DI on-the-job training course,
Southampton, 25-28 January 2011.
\url{https://wayback.archive-it.org/12090/20231221144450/https://cros-legacy.ec.europa.eu/content/job-training_en}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised_learning.R
\name{mec}
\alias{mec}
\title{Unsupervised Maximum Entropy Classifier for Record Linkage}
\usage{
mec(
  A,
  B,
  variables,
  comparators = NULL,
  methods = NULL,
  duplicates_in_A = FALSE,
  start_params = NULL,
  nonpar_hurdle = TRUE,
  set_construction = NULL,
  target_rate = 0.03,
  max_iter_bisection = 100,
  tol = 0.005,
  delta = 0.5,
  eps = 0.05,
  max_iter_em = 10,
  tol_em = 1,
  controls_nleqslv = list(),
  controls_kliep = control_kliep(),
  true_matches = NULL,
  verbose = FALSE
)
}
\arguments{
\item{A}{A duplicate-free \code{data.frame} or \code{data.table}.}

\item{B}{A duplicate-free \code{data.frame} or \code{data.table}.}

\item{variables}{A character vector of key variables used to create comparison vectors.}

\item{comparators}{A named list of functions for comparing pairs of records.}

\item{methods}{A named list of methods used for estimation (\code{"binary"}, \code{"continuous_parametric"}, \code{"continuous_nonparametric"} or \code{"hit_miss"}).}

\item{duplicates_in_A}{Logical indicating whether to allow \code{A} to have duplicate records.}

\item{start_params}{Start parameters for the \code{"binary"}, \code{"continuous_parametric"} and \code{"hit_miss"} methods.}

\item{nonpar_hurdle}{Logical indicating whether to use a hurdle model or not
(used only if the \code{"continuous_nonparametric"} method has been chosen for at least one variable).}

\item{set_construction}{A method for constructing the predicted set of matches (\code{"size"}, \code{"flr"} or \code{"mmr"}).}

\item{target_rate}{A target false link rate (FLR) or missing match rate
(MMR) (used only if \code{set_construction == "flr"} or \code{set_construction == "mmr"}).}

\item{max_iter_bisection}{A maximum number of iterations for the bisection procedure
(used only if \code{set_construction == "flr"} or \code{set_construction == "mmr"}).}

\item{tol}{Error tolerance in the bisection procedure
(used only if \code{set_construction == "flr"} or \code{set_construction == "mmr"}).}

\item{delta}{A numeric value specifying the tolerance for the change in the estimated number of matches between iterations.}

\item{eps}{A numeric value specifying the tolerance for the change in model parameters between iterations.}

\item{max_iter_em}{A maximum number of iterations for the EM algorithm
(used only if the \code{"hit_miss"} method has been chosen for at least one variable).}

\item{tol_em}{Error tolerance in the EM algorithm
(used only if the \code{"hit_miss"} method has been chosen for at least one variable).}

\item{controls_nleqslv}{Controls passed to the \link[nleqslv:nleqslv]{nleqslv()} function
(only if the \code{"continuous_parametric"} method has been chosen for at least one variable).}

\item{controls_kliep}{Controls passed to the \link[densityratio:kliep]{kliep()} function
(only if the \code{"continuous_nonparametric"} method has been chosen for at least one variable).}

\item{true_matches}{A \code{data.frame} or \code{data.table} indicating known matches.}

\item{verbose}{Logical indicating whether to print progress messages.}
}
\value{
Returns a list containing:\cr
\itemize{
\item{\code{M_est} -- a \code{data.table} with predicted matches,}
\item{\code{n_M_est} -- estimated classification set size,}
\item{\code{flr_est} -- estimated false link rate (FLR),}
\item{\code{mmr_est} -- estimated missing match rate (MMR),}
\item{\code{iter_bisection} -- the number of iterations in the bisection procedure,}
\item{\code{b_vars} -- a character vector of variables used for the \code{"binary"} method (with the prefix \code{"gamma_"}),}
\item{\code{cpar_vars} -- a character vector of variables used for the \code{"continuous_parametric"} method (with the prefix \code{"gamma_"}),}
\item{\code{cnonpar_vars} -- a character vector of variables used for the \code{"continuous_nonparametric"} method (with the prefix \code{"gamma_"}),}
\item{\code{hm_vars} -- a character vector of variables used for the \code{"hit_miss"} method (with the prefix \code{"gamma_"}),}
\item{\code{b_params} -- parameters estimated using the \code{"binary"} method,}
\item{\code{cpar_params} -- parameters estimated using the \code{"continuous_parametric"} method,}
\item{\code{hm_params} -- parameters estimated using the \code{"hit_miss"} method,}
\item{\code{ratio_kliep} -- a result of the \link[densityratio:kliep]{kliep()} function,}
\item{\code{variables} -- a character vector of key variables used for comparison,}
\item{\code{set_construction} -- a method for constructing the predicted set of matches,}
\item{\code{eval_metrics} -- standard metrics for quality assessment (if \code{true_matches} is provided),}
\item{\code{confusion} -- confusion matrix (if \code{true_matches} is provided).}
}
}
\description{
Implements several extensions to the maximum entropy classification (MEC) algorithm for record linkage
(see \href{https://www150.statcan.gc.ca/n1/pub/12-001-x/2022001/article/00007-eng.htm}{Lee et al. (2022)}),
iteratively estimating probability/density ratios to classify record pairs into matches and non-matches
based on comparison vectors.
}
\details{
Consider two datasets without duplicates: \eqn{A} and \eqn{B}.
Let the bipartite comparison space \eqn{\Omega = A \times B} consist of
matches \eqn{M} and non-matches \eqn{U} between the records in files
\eqn{A} and \eqn{B}. For any pair of records \eqn{(a,b) \in \Omega},
let \eqn{\pmb{\gamma}_{ab} = (\gamma_{ab}^1,\gamma_{ab}^2,
\ldots,\gamma_{ab}^K)'} be the comparison vector between
a set of key variables. The original MEC algorithm uses the binary
comparison function to evaluate record pairs across two datasets.
However, this approach may be insufficient when handling datasets
with frequent errors across multiple variables.

We propose the use of continuous comparison functions to address
the limitations of binary comparison methods. We consider every
semi-metric, i.e., a function \eqn{d: A \times B \to \mathbb{R}},
satisfying the following conditions:\cr
\enumerate{
\item{\eqn{d(x,y) \geq 0},}
\item{\eqn{d(x,y) = 0} if and only if \eqn{x = y},}
\item{\eqn{d(x,y) = d(y,x)}.}
}
For example, we can use the Jaro-Winkler distance for character variables
(which is implemented in the \code{automatedRecLin} package as \code{\link[=jarowinkler_complement]{jarowinkler_complement()}})
or the Euclidean distance for numerical variables. The \code{automatedRecLin} package allows the use of
a different comparison function for each key variable (which should be specified
as a list in the \code{comparators} argument). The default function
for each key variable is \link[reclin2:cmp_identical]{cmp_identical()}
(the binary comparison function).

The \code{\link[=mec]{mec()}} function offers different approaches to estimate the
probability/density ratio between matches and non-matches,
which should be specified as a list in the \code{methods} argument.
The available methods suitable for the binary comparison function
are \code{"binary"} and \code{"hit_miss"}. Both assume that \eqn{\gamma_{ab}^k|M}
and \eqn{\gamma_{ab}^k|U} follow Bernoulli distributions.
\code{"binary"} and \code{"hit_miss"} both estimate the parameters for the matches iteratively,
but \code{"binary"} estimates the parameters for the non-matches
only at the start, while \code{"hit_miss"} does
so iteratively using a hit-miss model (for details see
\href{https://www150.statcan.gc.ca/n1/pub/12-001-x/2022001/article/00007-eng.htm}{Lee et al. (2022)}).
\code{"binary"} is the default method for each variable.

For the continuous semi-metrics we suggest the usage
of \code{"continuous_parametric"} or \code{"continuous_nonparametric"}
method. The \code{"continuous_parametric"} method assumes that
\eqn{\gamma_{ab}^k|M} and \eqn{\gamma_{ab}^k|U} follow
hurdle Gamma distributions. The density function of a hurdle
Gamma distribution is characterized by three parameters
\eqn{p_0 \in (0,1)} and \eqn{\alpha, \beta > 0} as follows:
\deqn{
f(x;p_0,\alpha,\beta) = p_0^{\mathbb{I}(x = 0)}[(1 - p_0) v(x;\alpha,\beta)]^{\mathbb{I}(x > 0)},
}
where
\deqn{
v(x;\alpha,\beta) = \frac{\beta^{\alpha} x^{\alpha - 1} \exp(-\beta x)}
{\Gamma(\alpha)}
}
is the density function of a Gamma distribution
(for details see \href{https://ideas.repec.org/a/eee/csdana/v179y2023ics0167947322002365.html}{Vo et al. (2023)}).
At the beginning, the algorithm estimates the parameters for the non-matches
and then does it iteratively for the matches.
The \code{"continuous_nonparametric"} method does not assume anything about
the distributions of the comparison vectors. It iteratively directly
estimates the density ratio between the matches and the non-matches, using
the Kullback-Leibler Importance Estimation Procedure (KLIEP).
For details see \href{https://link.springer.com/article/10.1007/s10463-008-0197-x}{Sugiyama et al. (2008)}.

The \code{\link[=mec]{mec()}} function allows the construction of the predicted set
of matches using its estimated size or the bisection procedure,
described in \href{https://www150.statcan.gc.ca/n1/pub/12-001-x/2022001/article/00007-eng.htm}{Lee et al. (2022)},
based on a target false link rate (FLR)
or missing match rate (MMR). To use the second option, set \code{set_construction = "flr"}
or \code{set_construction = "mmr"} and
specify a target error rate using the \code{target_rate} argument.

The assumption that \eqn{A} and \eqn{B} contain no duplicate records
might be relaxed by allowing \eqn{A} to have duplicates. To do so,
set \code{duplicates_in_A = TRUE}.
}
\examples{
df_1 <- data.frame(
  name = c("Emma", "Liam", "Olivia", "Noah", "Ava",
           "Ethan", "Sophia", "Mason", "Isabella", "James"),
  surname = c("Smith", "Johnson", "Williams", "Brown", "Jones",
              "Garcia", "Miller", "Davis", "Rodriguez", "Wilson"),
  city = c("New York", "Los Angeles", "Chicago", "Houston", "Phoenix",
           "Philadelphia", "San Antonio", "San Diego", "Dallas", "San Jose")
)

df_2 <- data.frame(
  name = c(
    "Emma", "Liam", "Olivia", "Noah",
    "Ava", "Ehtan", "Sopia", "Mson",
    "Charlotte", "Benjamin", "Amelia", "Lucas"
  ),
  surname = c(
     "Smith", "Johnson", "Williams", "Brown",
    "Jnes", "Garca", "Miler", "Dvis",
    "Martinez", "Lee", "Hernandez", "Clark"
  ),
  city = c(
    "New York", "Los Angeles", "Chicago", "Houston",
    "Phonix", "Philadelpia", "San Antnio", "San Dieg",
    "Seattle", "Miami", "Boston", "Denver"
  )
)
true_matches <- data.frame(
  "a" = 1:8,
  "b" = 1:8
)

variables <- c("name", "surname", "city")
comparators <- list(
  "name" = jarowinkler_complement(),
  "surname" = jarowinkler_complement(),
  "city" = jarowinkler_complement()
)
methods <- list(
  "name" = "continuous_parametric",
  "surname" = "continuous_parametric",
  "city" = "continuous_parametric"
)

set.seed(1)
result <- mec(A = df_1, B = df_2,
              variables = variables,
              comparators = comparators,
              methods = methods,
              true_matches = true_matches)
result
}
\references{
Lee, D., Zhang, L.-C. and Kim, J. K. (2022). Maximum entropy classification for record linkage.
Survey Methodology, Statistics Canada, Catalogue No. 12-001-X, Vol. 48, No. 1.

Vo, T. H., Chauvet, G., Happe, A., Oger, E., Paquelet, S., and Garès, V. (2023).
Extending the Fellegi-Sunter record linkage model for mixed-type data with application to the French national health data system.
Computational Statistics & Data Analysis, 179, 107656.

Sugiyama, M., Suzuki, T., Nakajima, S. et al. Direct importance estimation for covariate shift adaptation.
Ann Inst Stat Math 60, 699–746 (2008). \doi{10.1007/s10463-008-0197-x}
}
\author{
Adam Struzik
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised_learning.R
\name{custom_rec_lin_model}
\alias{custom_rec_lin_model}
\title{Create a Custom Record Linkage Model}
\usage{
custom_rec_lin_model(ml_model, vectors)
}
\arguments{
\item{ml_model}{A trained ML model that predicts the probability of a match based on comparison vectors.}

\item{vectors}{An object of class \code{comparison_vectors} (a result of \code{\link[=comparison_vectors]{comparison_vectors()}}), used for training the \code{ml_model}.}
}
\value{
Returns a list containing:\cr
\itemize{
\item{\code{b_vars} -- here \code{NULL},}
\item{\code{cpar_vars} -- here \code{NULL},}
\item{\code{cnonpar_vars} -- here \code{NULL},}
\item{\code{b_params} -- here \code{NULL},}
\item{\code{cpar_params} -- here \code{NULL},}
\item{\code{cnonpar_params} -- here \code{NULL},}
\item{\code{ratio_kliep} -- here \code{NULL},}
\item{\code{ratio_kliep_list} -- here \code{NULL},}
\item{\code{ml_model} -- ML model used for creating the record linkage model,}
\item{\code{pi_est} -- a prior probability of matching,}
\item{\code{match_prop} -- proportion of matches in the smaller dataset,}
\item{\code{variables} -- a character vector of key variables used for comparison,}
\item{\code{comparators} -- a list of functions used to compare pairs of records,}
\item{\code{methods} -- here \code{NULL},}
\item{\code{prob_ratio} -- here \code{"2"}.}
}
}
\description{
Creates a supervised record linkage model using a custom machine learning (ML) classifier.
}
\details{
The \code{\link[=custom_rec_lin_model]{custom_rec_lin_model()}} function creates a custom record linkage model,
based on known matches and non-matches (which might later serve as a classifier
for pairs outside training data). The procedure of creating a custom model
based on training data is as follows.
\enumerate{
\item{Use \code{\link[=comparison_vectors]{comparison_vectors()}} to compare pairs of records.}
\item{Train a machine learning classifier using the \code{Omega} element
of the output of \code{\link[=comparison_vectors]{comparison_vectors()}}. The classifier should
predict the probability of matching based on a given vector.}
\item{Use \code{\link[=custom_rec_lin_model]{custom_rec_lin_model()}} with
appropriate arguments.}
}
}
\examples{
if (requireNamespace("xgboost", quietly = TRUE)) {
  df_1 <- data.frame(
    "name" = c("James", "Emma", "William", "Olivia", "Thomas",
    "Sophie", "Harry", "Amelia", "George", "Isabella"),
    "surname" = c("Smith", "Johnson", "Brown", "Taylor", "Wilson",
    "Davis", "Clark", "Harris", "Lewis", "Walker")
  )
  df_2 <- data.frame(
    "name" = c("James", "Ema", "Wimliam", "Olivia", "Charlotte",
    "Henry", "Lucy", "Edward", "Alice", "Jack"),
    "surname" = c("Smith", "Johnson", "Bron", "Tailor", "Moore",
    "Evans", "Hall", "Wright", "Green", "King")
  )
  comparators <- list("name" = jarowinkler_complement(),
                      "surname" = jarowinkler_complement())
  matches <- data.frame("a" = 1:4, "b" = 1:4)
  vectors <- comparison_vectors(A = df_1, B = df_2, variables = c("name", "surname"),
                               comparators = comparators, matches = matches)
  model_xgb <- xgboost::xgboost(x = as.matrix(vectors$Omega[, c("gamma_name", "gamma_surname")]),
                       y = factor(vectors$Omega$match),
                       objective = "binary:logistic", eval_metric = "logloss",
                       nrounds = 100, verbosity = 0, nthread = 1)
  custom_xgb_model <- custom_rec_lin_model(model_xgb, vectors)
  custom_xgb_model
}
}
\author{
Adam Struzik
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/controls.R
\name{control_kliep}
\alias{control_kliep}
\title{Controls for the \link[densityratio:kliep]{kliep()} Function}
\usage{
control_kliep(scale = NULL, progressbar = FALSE, nfold = 2, ...)
}
\arguments{
\item{scale}{\code{"numerator"}, \code{"denominator"} or \code{NULL}, indicating whether to standardize each numeric variable according
to the numerator means and standard deviations, the denominator means and standard deviations,
or apply no standardization at all.}

\item{progressbar}{Logical indicating whether or not to display a progress bar.}

\item{nfold}{Number of cross-validation folds used in order to calculate the optimal sigma value (default is 2-fold cross-validation).}

\item{...}{Additional arguments.}
}
\value{
Returns a list with parameters.
}
\description{
Controls for the \link[densityratio:kliep]{kliep()} function used in the package.
}
\author{
Adam Struzik
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{B_example}
\alias{B_example}
\title{B_example dataset}
\format{
A \code{data.frame} with 12 records. Each row represents one record, with the following columns:
\code{name}, \code{surname}, and \code{city}. Some records can be matched with records in the
\code{A_example} dataset.
}
\usage{
B_example
}
\description{
An example dataset containing artificial personal data.
}
\examples{
data("B_example")
B_example
}
\keyword{datasets}
